package xic;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import polyglot.util.OptimalCodeWriter;
import xic.Configuration.Architecture;
import xic.Configuration.AssemblySyntax;
import xic.Configuration.Target;
import xic.analysis.AnticipatedExpressions;
import xic.analysis.AvailableCopies;
import xic.analysis.AvailableExpressions;
import xic.analysis.BoundingIntervals;
import xic.analysis.CFGGenerator;
import xic.analysis.CFGNode;
import xic.analysis.CodeOrdering;
import xic.analysis.ConstPropagation;
import xic.analysis.CopyPropagation;
import xic.analysis.DataFlowAnalysis;
import xic.analysis.DeadCodeElimination;
import xic.analysis.Delayedness;
import xic.analysis.DotGenerator;
import xic.analysis.EarliestExpressions;
import xic.analysis.EssentialVariables;
import xic.analysis.IntervalAnalysis;
import xic.analysis.Isolated;
import xic.analysis.Latestness;
import xic.analysis.LiveVariables;
import xic.analysis.LoopStructure;
import xic.analysis.PartialRedundancy;
import xic.analysis.PropagatedConstants;
import xic.analysis.PropagatedExpressions;
import xic.analysis.ReachingDefinitions;
import xic.analysis.Substitutable;
import xic.analysis.WebGenerator.CoalesceCriteria;
import xic.ast.ClassType;
import xic.ast.CompUnit;
import xic.ast.ErrorReporter;
import xic.ast.TypeNode;
import xic.codegen.AsmDirective;
import xic.codegen.GraphColoringRegAlloc;
import xic.codegen.InstrList;
import xic.codegen.PeepholeOptimizer;
import xic.codegen.RegAlloc;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.ir.IRCompUnit;
import xic.ir.IRFuncDecl;
import xic.ir.IRPrinter;
import xic.ir.IRSimulator;
import xic.parser.ParserImpl;
import xic.parser.ParserImpl.Language;
import xic.parser.TypedASTPrinter;
import xic.visitors.IRConstantFolder;
import xic.visitors.IRGenerator;
import xic.visitors.IRLowering;
import xic.visitors.TypeChecker;

public class Driver {
	public static void main(String args[]) {
		System.exit(run(args));
	}

	/**
	 * Instead of calling <code>System.exit</code>, this function returns an 
	 * <code>int</code> to indicate success/failure (to ease construction of test cases)
	 */
	public static int run(String[] args) {
		boolean runSim = false, runRA = true, displayStats = false;
		boolean dumpWebs = false, dumpLoops = false, dumpObjs = false;
		String inputFile = null, outputFile = null;
		String dumpAST = "", dumpCFG = "", dumpIR = "";
		List<String> simArgs = new ArrayList<String>();
		List<String> dumpDFAs = new ArrayList<String>();
		List<String> searchPaths = new ArrayList<String>();
		Set<String> boostedTiles = new HashSet<String>();
		Set<String> validOptimizations = new HashSet<String>();
		Statistics stats = new Statistics();

		/* Simplistic command line parser */
		validOptimizations.add("cf");
		validOptimizations.add("dce");
		validOptimizations.add("reg");
		validOptimizations.add("rc");
		validOptimizations.add("rc2");
		validOptimizations.add("pre");
		validOptimizations.add("copy");
		validOptimizations.add("cp");
		validOptimizations.add("peel");
		validOptimizations.add("peep");
		
		searchPaths.add(".");

		Target target = Configuration.detectTarget();
		Architecture architecture = Architecture.X86_64;
		AssemblySyntax syntax = AssemblySyntax.ATT_SYNTAX;

		String argsStr = "java -jar xic.jar ";
		for (int i=0; i<args.length; ++i)
			argsStr += args[i] + " ";

		Set<String> optimizations = new HashSet<String>(validOptimizations);
		boolean hasPlusOParameter = false;
		for (int i=0; i<args.length; ++i) {
			String arg = args[i];
			if (i + 1 == args.length
					&& (arg.equals("--dump_dfa") || arg.equals("--dump_ir")
							|| arg.equals("--dump_cfg")
							|| arg.equals("--dump_ast")
							|| arg.equals("--boost"))) {
				System.out.println("Error: Missing an argument to '" + arg + "'!");
				return -1;
			}

			if (arg.equals("--arch")) {
				String format = args[i+1];
				if (format.equals("i386")) {
					architecture = Architecture.I386;
				} else if (format.equals("x86_64")) {
					architecture = Architecture.X86_64;
				} else {
					System.out.println("Error: architecture parameter must be "
							+"one of 'i386', or 'x86_64'");
					return -1;
				}
				
				++i;
			} else if(arg.equals("--target")) {
				String format = args[i+1];
				if (format.equals("darwin")) {
					target = Target.DARWIN;
				} else if (format.equals("linux")) {
					target = Target.LINUX;
				} else if (format.equals("windows")) {
					target = Target.WINDOWS;
				} else {
					System.out.println("Error: target parameter must be "
							+"one of 'linux', 'windows', or 'darwin'");
					return -1;
				}
				
				++i;
			} else if(arg.equals("--asmformat")) {
				String format = args[i+1];
				if (format.equals("att")) {
					syntax = AssemblySyntax.ATT_SYNTAX;
				} else if (format.equals("intel")) {
					syntax = AssemblySyntax.INTEL_SYNTAX;
				} else {
					System.out.println("Error: asmformat parameter must be "
							+"one of 'att', or 'intel'");
					return -1;
				}
				
				++i;
			} else if(arg.equals("-I")) {
				searchPaths.add(args[i+1]);
				++i;
			} else if (arg.equals("--dump_ast")) {
				dumpAST = args[i+1];
				if (!dumpAST.equals("initial") && !dumpAST.equals("final")) {
					System.out.println("Error: dump_ast parameter must be one "+
							"of 'initial' or 'final'");
					return -1;
				}
				++i;
			} else if (arg.equals("--dump_ir")) {
				dumpIR= args[i+1];
				if (!dumpIR.equals("initial") && !dumpIR.equals("final")
						&& !dumpIR.equals("lowered")) {
					System.out.println("Error: dump_ir parameter must be one "+
							"of 'initial', 'lowered' or 'final'");
					return -1;
				}
				++i;
			} else if(arg.equals("--dump_cfg")) {
				dumpCFG = args[i+1];
				if (!dumpCFG.equals("initial") && !dumpCFG.equals("final") && 
					 !dumpCFG.equals("asm")) {
					System.out.println("Error: dump_cfg parameter must be "
							+"one of 'initial', 'final' or 'asm'");
					return -1;
				}
				++i;
			} else if (arg.equals("--dump_dfa")) {
				String dumpDFA = args[i+1];
				if (!dumpDFA.equals("AnticipatedExpressions") && !dumpDFA.equals("AvailableExpressions")
						&& !dumpDFA.equals("Delayedness") && !dumpDFA.equals("EarliestExpressions") 
						&& !dumpDFA.equals("EssentialVariables") && !dumpDFA.equals("Isolated") 
						&& !dumpDFA.equals("LiveVariablesAsm") && !dumpDFA.equals("ReachingDefinitions")
						&& !dumpDFA.equals("Optimal") && !dumpDFA.equals("Redundant")
						&& !dumpDFA.equals("AvailableCopies") && !dumpDFA.equals("PropagatedConstants")
						&& !dumpDFA.equals("BoundingIntervals") && !dumpDFA.equals("Latestness")
						&& !dumpDFA.equals("Eval") && !dumpDFA.equals("LiveVariablesIR")
						&& !dumpDFA.equals("SubstitutablePRE") && !dumpDFA.equals("PropagatedExpressions")
						&& !dumpDFA.equals("LiveVariablesPRE") && !dumpDFA.equals("PropagatedExpressionsPRE")) {
					System.out.println("Error: Invalid parameter to dump_dfa:" + dumpDFA);
					return -1;
				}
				dumpDFAs.add(dumpDFA);
				++i;
			} else if (arg.equals("--dump_webs")) {
				dumpWebs = true;
			} else if (arg.equals("--dump_objs")) {
				dumpObjs = true;
			} else if (arg.equals("--dump_loops")) {
				dumpLoops = true;
			} else if (arg.startsWith("+O")) {
				if (!hasPlusOParameter) {
					hasPlusOParameter = true;
					optimizations.clear();
				}
				String opt = arg.substring(2);
				if (opt.length() == 0) {
					System.out.println("+O: need to specify an optimizations!");
					return -1;
				}
				if (!validOptimizations.contains(opt)) {
					System.out.println("Unknown optimization '"+opt+"'");
					return -1;
				}
				optimizations.add(opt);
			} else if (arg.startsWith("-O")) {
				String opt = arg.substring(2);
				if (opt.length() == 0) {
					optimizations.clear();
				} else {					
					if (!validOptimizations.contains(opt)) {
						System.out.println("Unknown optimization '"+opt+"'");
						return -1;
					}
					optimizations.remove(opt);
				}
			} else if (arg.equals("--sim")) {
				runSim = true;
			} else if (arg.equals("--no_ra")) {
				runRA = false;
			} else if (arg.equals("--annotate")) {
				GraphColoringRegAlloc.VERBOSE = true;
			} else if (arg.equals("-v")) {
				displayStats = true;
			} else if (arg.equals("-o")) {
				outputFile = args[i+1];
				++i;
			} else if (arg.equals("--boost")) {
				boostedTiles.add(args[i+1]);
				++i;
			} else {
				if (inputFile == null) {
					inputFile = arg;
				} else {
					simArgs.add(arg);
				}
			}
		}

		if (inputFile == null) {
			System.out.println("Iota/09 + Xi reference compiler -- based on the CS4120 Iota/09 compiler by the");
			System.out.println("group of Owen Arden, Catalin Dumitru, Wenzel Jakob, and Danfeng Zhang");
			System.out.println();
			System.out.println("Ported to 64bit and extended to additionally handle Xi by Wenzel Jakob");
			System.out.println();
			System.out.println("Syntax: java -jar xic.jar [options] <Iota09/Xi input file> [sim args]\n");
			System.out.println("   -v                 Display statistics and timing information\n");
			System.out.println("   -o <file>          Write the generated assembly to <file>\n");
			System.out.println("   -I <path>          Add a directory to the interface file search path\n");
			System.out.println("   +O<opt>            Enable optimization <opt>, which can be one of");
			System.out.println("                        'dce'   : Dead code elimination");
			System.out.println("                        'rc'    : Register coalescing");
			System.out.println("                        'rc2'   : More drastic register coalescing");
			System.out.println("                        'pre'   : Partial redundancy elimination");
			System.out.println("                        'copy'  : Copy propagation");
			System.out.println("                        'cp'    : Constant propagation");
			System.out.println("                        'peel'  : Loop peeling");
			System.out.println("                        'peep'  : Peephole optimizations");
			System.out.println("                        'subst' : Forward expression substitution\n");
			System.out.println("   -O<opt>            Disable optimization <opt>\n");
			System.out.println("   -O                 Disable all optimizations\n");
			System.out.println("   --arch <fmt>       Generate assembly using the one of the following styles:");
			System.out.println("                        'i386'   : 32-bit instructions, cdecl calling conventions");
			System.out.println("                        'x86_64' : AMD64 instructions, OS-dependent conventions\n");
			System.out.println("   --target <trg>     Override the target OS (for calling conventions etc.)");
			System.out.println("                         Must be one of 'linux', 'windows', or 'darwin'\n");
			System.out.println("   --asmformat <fmt>  Generate assembly using the one of the following styles:");
			System.out.println("                        'att'   : AT&T syntax (default on OSX)");
			System.out.println("                        'intel' : Intel syntax (default everywhere else)\n");
			System.out.println("   --dump_ir <phase>  Output a dump of the IR representation. Phase is one of");
			System.out.println("                        'initial' : IR before lowering");
			System.out.println("                        'lowered' : IR after lowering");
			System.out.println("                        'final'   : IR after all optimizations\n");
			System.out.println("   --dump_cfg <phase> Output a dump of the control flow graph. Phase is one of");
			System.out.println("                        'initial' : Canonical IR CFG before any optimizations");
			System.out.println("                        'final'   : Canonical IR CFG after all optimizations");
			System.out.println("                        'asm'     : Abstract assembly control flow graph\n");
			System.out.println("   --dump_ast <phase> Output a dump of the typed AST. Phase is one of");
			System.out.println("                        'initial' or 'final' (both produce the same result)\n");
			System.out.println("   --dump_webs        Generate GraphViz files illustrating the webs used");
			System.out.println("                      by the register allocator. Creates one file for each");
			System.out.println("                      function and register allocator pass.\n");
			System.out.println("   --dump_loops       Generate GraphViz files illustrating the loop structure");
			System.out.println("                      of the program. Creates one file for each function.\n");
			System.out.println("   --dump_dfa <name>  Generate GraphViz files illustrating the result of a data");
			System.out.println("                      flow analysis. <name> must be one of:");
			System.out.println("                        'AnticipatedExpressions', 'AvailableExpressions',");
			System.out.println("                        'Delayedness', 'EarliestExpressions', 'Isolated',");
			System.out.println("                        'EssentialVariables', 'Optimal', 'Redundant',");
			System.out.println("                        'Latestness', 'LiveVariablesIR', 'LiveVariablesAsm',");
			System.out.println("                        'AvailableCopies', 'PropagatedConstants'");
			System.out.println("                        'BoundingIntervals', 'Eval', 'ReachingDefinitions'.\n");
			System.out.println("   --dump_objs        Dump the object and dispatch table layout for all classes.\n");
			System.out.println("   --annotate         Generate annotated assembly output\n");
			System.out.println("   --no_ra            Don't run the register allocator, emit abstract assembly\n");
			System.out.println("   --sim              Run the IR-level simulator on the output and call ");
			System.out.println("                         'main(args:int[][])' with the supplied parameters\n");
			System.out.println("   --boost <class>    Boost the priority of a slice matcher. May be");
			System.out.println("                      specified multiple times for different classes.\n");
			return 1;
		}

		if (optimizations.contains("rc") && !optimizations.contains("reg"))
			System.out.println("Warning: not performing register coalescing since the "
							+ "graph-coloring allocator is inactive.");

		stats.log("general", "Active optimizations: " + optimizations);
		
		Language lang;

		if (inputFile.endsWith(".xi")) {
			lang = Language.XI;
		} else if (inputFile.endsWith(".ix") || inputFile.endsWith(".i9")) {
			lang = Language.IOTA_09;
		} else {
			System.out.println("The name '" + inputFile + "' does not specify "
					+ "an Iota/09 or Xi source code file!");
			return -1;
		}
		
		if (outputFile == null)
			outputFile = inputFile.substring(0, inputFile.length()-3) + ".s";

		File f = new File(inputFile);
		if (!f.exists() || !f.canRead()) {
			System.out.println("Unable to read from file '" + f.getName() + "'");
			return -1;
		}

		try {
			Configuration.setArchitecture(architecture);
			Configuration.setTarget(target);
			Configuration.setSyntax(syntax);

			stats.startTimer("parse");
			ErrorSink errorSink = new ErrorSink();
			ParserImpl p = new ParserImpl(lang, 
					new BufferedReader(new FileReader(f)), inputFile);
			CompUnit compUnit = (CompUnit) p.parse(errorSink);
			stats.stopTimer("parse");

			OutputStreamWriter writer = new OutputStreamWriter(System.out);
			OptimalCodeWriter ocw = new OptimalCodeWriter(writer, 80);

			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(errorSink);
				return -1;
			}

			/* Dump the AST if requested */
			if (dumpAST.equals("initial"))
				new TypedASTPrinter(ocw).print(compUnit);

			stats.startTimer("typeCheck");

			TypeChecker typeChecker = new TypeChecker(lang, errorSink, searchPaths);
			CompUnit typeCheckedUnit = typeChecker.run(compUnit);
	
			if (errorSink.hasErrors()) {
				ErrorReporter.printErrors(errorSink);
				return -1;
			}
			stats.stopTimer("typeCheck");

			/* Dump the AST if requested */
			if (dumpAST.equals("final"))
				new TypedASTPrinter(ocw).print(typeCheckedUnit);

			/* Dump information about objects if requested */
			if (dumpObjs) {
				for (TypeNode t : typeCheckedUnit.globalStmtBlock()
						.getSymbolTable().values()) {
					if (t instanceof ClassType)
						System.out.println(((ClassType) t).dump());
				}
			}

			stats.startTimer("irGen");
			IRGenerator generator = new IRGenerator(typeChecker.getNameGenerator(), optimizations.contains("peel"),
					lang);
			IRCompUnit irCompUnit = generator.translate(typeCheckedUnit, typeChecker.getImports());
			stats.stopTimer("irGen");
			
			Map<String,Boolean> isProcedureHash = new HashMap<String,Boolean>();
			for(IRFuncDecl func : irCompUnit.functions().values()) {
				isProcedureHash.put(func.name(), func.isProcedure());
			}
			if (optimizations.contains("peel")) {
				if (generator.getNumPeeled() > 0)
					stats.log("irGen", "Total # of loops peeled: " + generator.getNumPeeled());
			}

			if (dumpIR.equals("initial"))
				new IRPrinter(ocw).print(irCompUnit);

			if (optimizations.contains("cf")) {
				stats.startTimer("constFold");
				IRConstantFolder constantFolder = new IRConstantFolder(errorSink);
				irCompUnit = (IRCompUnit) constantFolder.visit(irCompUnit);

				if (errorSink.hasErrors()) {
					ErrorReporter.printErrors(errorSink);
					return -1;
				}
				stats.stopTimer("constFold");
			}

			stats.startTimer("lower");
			IRLowering loweringVisitor = new IRLowering(typeChecker.getNameGenerator());
			irCompUnit = (IRCompUnit) loweringVisitor.visit(irCompUnit);
			stats.stopTimer("lower");

			Map<String, CFGNode> cfgs = CFGGenerator.createCFG(irCompUnit);

			if (dumpIR.equals("lowered"))
				new IRPrinter(ocw).print(irCompUnit);

			if (dumpCFG.equals("initial"))
				System.out.println(DotGenerator.toDot(cfgs));
			
			/* Dump the loop structure if requested */
			if (dumpLoops)
				LoopStructure.dumpLoops(cfgs);

				
			/* Dump data flow analyses if requested (IR-level) */
			Map<String, CFGNode> dfaCfgs = CFGGenerator.createCFG(irCompUnit);
			if (dumpDFAs.contains("BoundingIntervals")) {
				/* Constant propagation */
				if (optimizations.contains("cp")) {
					stats.startTimer("cp");
					/* Two passes (the first may remove control flow, which
					   provides additional opportunities in the second pass) */
					ConstPropagation.transform(stats, 1, cfgs);
					ConstPropagation.transform(stats, 2, cfgs);
					stats.stopTimer("cp");
				}

				/* Copy propagation */
				if (optimizations.contains("copy")) {
					stats.startTimer("copy");
					CopyPropagation.transform(stats, 1, cfgs);
					stats.stopTimer("copy");
				}
				dfaCfgs = cfgs;
			} 
			if(dumpDFAs.contains("SubstitutablePRE")) {
				/* Constant propagation */
				if (optimizations.contains("cp")) {
					stats.startTimer("cp");
					/* Two passes (the first may remove control flow, which
					   provides additional opportunities in the second pass) */
					ConstPropagation.transform(stats, 1, cfgs);
					ConstPropagation.transform(stats, 2, cfgs);
					stats.stopTimer("cp");
				}

				/* Copy propagation */
				if (optimizations.contains("copy")) {
					stats.startTimer("copy");
					CopyPropagation.transform(stats, 1, cfgs);
					stats.stopTimer("copy");
				}	

				/* Dead code removal */
				if (optimizations.contains("dce")) {
					stats.startTimer("dce");
					DeadCodeElimination.transform(stats, 1, cfgs);
					stats.stopTimer("dce");
				}
				/* Partial redundancy elimination */
				if (optimizations.contains("pre")) {
					irCompUnit = CodeOrdering.orderCode(irCompUnit.name(), cfgs,
							irCompUnit.extraSequence(), isProcedureHash);
					stats.startTimer("pre");
					PartialRedundancy pre = new PartialRedundancy(stats, typeChecker.getNameGenerator());
					pre.run(cfgs);
					stats.stopTimer("pre");

					/* Copy propagation */
					if (optimizations.contains("copy")) {
						stats.startTimer("copy");
						CopyPropagation.transform(stats, 2, cfgs);
						stats.stopTimer("copy");
					}

					/* Dead code removal */
					if (optimizations.contains("dce")) {
						stats.startTimer("dce");
						DeadCodeElimination.transform(stats, 2, cfgs);
						stats.stopTimer("dce");
					}
				}
				dfaCfgs = cfgs;
			}

			for (String name : irCompUnit.functions().keySet()) {

				DotGenerator dg = new DotGenerator();
				List<DataFlowAnalysis> dfas = new ArrayList<DataFlowAnalysis>();
				CFGNode rootNode = dfaCfgs.get(name);
				Set<CFGNode> nodes = rootNode.flatten();
				try {			
					for(String dumpDFA : dumpDFAs) {
						DataFlowAnalysis dfa=null;

						if (dumpDFA.equals("AvailableExpressions")) {
							dfa = new AvailableExpressions(nodes);
						} else if (dumpDFA.equals("AnticipatedExpressions")) {
							dfa = new AnticipatedExpressions(nodes);
						} else if (dumpDFA.equals("PropagatedConstants")) {
							dfa = new PropagatedConstants(nodes);
						} else if (dumpDFA.equals("BoundingIntervals")) {
							AvailableCopies ac = new AvailableCopies(nodes);
							ac.solve();
							dfa = new BoundingIntervals(nodes);
						} else if (dumpDFA.equals("EarliestExpressions")) {
							AnticipatedExpressions ante = new AnticipatedExpressions(nodes);
							ante.solve();
							dfa = new EarliestExpressions(ante, nodes);
						} else if (dumpDFA.equals("Delayedness")) {
							AnticipatedExpressions ante = new AnticipatedExpressions(nodes);
							ante.solve();
							EarliestExpressions early = new EarliestExpressions(ante, nodes);
							early.solve();
							dfa = new Delayedness(nodes, rootNode, ante, early);
						} else if (dumpDFA.equals("EssentialVariables")) {
							dfa = new EssentialVariables(nodes);
						} else if (dumpDFA.equals("AvailableCopies")) {
							dfa = new AvailableCopies(nodes);
						} else if (dumpDFA.equals("Isolated")) {
							AnticipatedExpressions ante = new AnticipatedExpressions(nodes);
							ante.solve();
							EarliestExpressions early = new EarliestExpressions(ante, nodes);
							early.solve();
							Delayedness del = new Delayedness(nodes, rootNode, ante, early);
							del.solve();
							Latestness last = new Latestness(nodes, ante, del);
							dfa = new Isolated(nodes, rootNode, ante, last);
						} else if (dumpDFA.equals("LiveVariablesIR")) {
							dfa = new LiveVariables(nodes);
						} else if (dumpDFA.equals("SubstitutablePRE") || dumpDFA.equals("LiveVariablesPRE")
								|| dumpDFA.equals("PropagatedExpressionsPRE")) {
							
							LiveVariables live = new LiveVariables(nodes);
							if(dumpDFA.equals("LiveVariablesPRE"))
								dfa = live;
							else {
								PropagatedExpressions prop = new PropagatedExpressions(nodes);
								/*prop.solve();
								for (CFGNode node : nodes) {
									String in_head="";
									if(dg.getIn().containsKey(node))
										in_head = dg.getIn().get(node) + "\n";
									PEFlowInfo peinfo = (PEFlowInfo)prop.getFlowInfo(node);
									dg.getIn().put(node, in_head + "Kill:" + String.valueOf(peinfo.kill));
								}
								*/
								if(dumpDFA.equals("PropagatedExpressionsPRE"))
									dfa=prop;
								else {
									live.solve();
									prop.solve();

									dfa = new Substitutable(nodes, rootNode,prop,live,5);
			
								}
							}
						} else if (dumpDFA.equals("PropagatedExpressions")) {
							dfa = new PropagatedExpressions(nodes);
						} else if (dumpDFA.equals("Optimal") || dumpDFA.equals("Redundant") 
								|| dumpDFA.equals("Latestness")) {
							PartialRedundancy pre = new PartialRedundancy(stats, typeChecker.getNameGenerator());
							pre.analyze(name, rootNode);
							
							for (CFGNode node : nodes) {
								String in_head="";
								if(dg.getIn().containsKey(node))
									in_head = dg.getIn().get(node) + "\n";
								
								if (dumpDFA.equals("Latestness"))
									dg.getIn().put(node, in_head + "Latest:" + String.valueOf(pre.latestness(node)));
								else if (dumpDFA.equals("Optimal"))
									dg.getIn().put(node, in_head + "Optimal:" + String.valueOf(pre.opt(node)));
								else
									dg.getIn().put(node, in_head + "Redundant:" + String.valueOf(pre.redn(node)));
							}
						} else if (dumpDFA.equals("Eval")) {
							AnticipatedExpressions ante = new AnticipatedExpressions(nodes);
							ante.solve();	
							for (CFGNode node : nodes) {
								String in_head="";
								if(dg.getIn().containsKey(node))
									in_head = dg.getIn().get(node) + "\n";
								
								dg.getIn().put(node, in_head+ "Eval:"+String.valueOf(ante.eval(node)));
							}
						} else {
							break;
						}

						if (dfa != null) {
							dfa.solve();
							dfas.add(dfa);
						}

					}
					if (dumpDFAs.size() > 0) {
						dg.setMultiDataFlowAnalysis(dfas, nodes);
						PrintWriter pw = new PrintWriter(new FileWriter("dfa_" + name + ".dot"));
						pw.print(dg.generate(nodes));
						pw.close();
					}
				} catch (IOException e) {
					throw new InternalCompilerException(
							"Could not generate debug output", e);
				}
			}

			/* Constant propagation */
			if (optimizations.contains("cp")) {
				stats.startTimer("cp");
				/* Two passes (the first may remove control flow, which
				   provides additional opportunities in the second pass) */
				ConstPropagation.transform(stats, 1, cfgs);
				ConstPropagation.transform(stats, 2, cfgs);
				stats.stopTimer("cp");
			}
			
			/* Copy propagation */
			if (optimizations.contains("copy")) {
				stats.startTimer("copy");
				CopyPropagation.transform(stats, 1, cfgs);
				stats.stopTimer("copy");
			}	

			/* Dead code removal */
			if (optimizations.contains("dce")) {
				stats.startTimer("dce");
				DeadCodeElimination.transform(stats, 1, cfgs);
				stats.stopTimer("dce");
			}

			/* Interval analysis */
			if (optimizations.contains("intr")) {
				stats.startTimer("intr");
				IntervalAnalysis.transform(stats, 1, cfgs);
				stats.stopTimer("intr");
			}

			/* Partial redundancy elimination */
			if (optimizations.contains("pre")) {
				irCompUnit = CodeOrdering.orderCode(irCompUnit.name(), cfgs,
						irCompUnit.extraSequence(), isProcedureHash);
				stats.startTimer("pre");
				PartialRedundancy pre = new PartialRedundancy(stats, typeChecker.getNameGenerator());
				pre.run(cfgs);
				stats.stopTimer("pre");

				/* Copy propagation */
				if (optimizations.contains("copy")) {
					stats.startTimer("copy");
					CopyPropagation.transform(stats, 2, cfgs);
					stats.stopTimer("copy");
				}

				/* Dead code removal */
				if (optimizations.contains("dce")) {
					stats.startTimer("dce");
					DeadCodeElimination.transform(stats, 2, cfgs);
					stats.stopTimer("dce");
				}
			}
			
			/* Forward Substitution for register pressure relief */
			/*if (optimizations.contains("subst")) {
				stats.startTimer("subst");
				LiveRangeSubstitution.transform(stats, 1, cfgs, 5);
				stats.stopTimer("subst");
			}*/

			/* Dead code removal */
			if (optimizations.contains("dce")) {
				stats.startTimer("dce");
				DeadCodeElimination.transform(stats, 3, cfgs);
				stats.stopTimer("dce");
			}
			
			if (dumpCFG.equals("final"))
				System.out.println(DotGenerator.toDot(cfgs));


			irCompUnit = CodeOrdering.orderCode(irCompUnit.name(), cfgs,
					irCompUnit.extraSequence(), isProcedureHash);
			if (dumpIR.equals("final"))
				new IRPrinter(ocw).print(irCompUnit);

			/* Run the IR-level simulator if requested */
			if (runSim) {
				IRSimulator sim = new IRSimulator(irCompUnit);
				int ws = Configuration.WORD_SIZE;

				/* Set up the command-line arguments */
				long listPtr = sim.malloc((simArgs.size()+1)*ws);
				sim.store(listPtr, simArgs.size());

				for (int i=0; i<simArgs.size(); ++i) {
					String s = simArgs.get(i);
					long stringPtr = sim.malloc((s.length()+1)*ws);
					sim.store(stringPtr, s.length());
					for (int j=0; j<s.length(); ++j)
						sim.store(stringPtr+(j+1)*ws, s.charAt(j));
					sim.store(listPtr + (i+1)*ws, stringPtr+ws);
				}

				/* Call the main function */
				sim.call("_Imain_paai", new long[] { listPtr + ws });
				return 0;
			}

			InstrList output = new InstrList();
			
			CoalesceCriteria cCriteria = CoalesceCriteria.COALESCE_NONE;
			if (optimizations.contains("rc"))
				cCriteria = CoalesceCriteria.COALESCE_BRIGGS;
			if (optimizations.contains("rc2"))
				cCriteria = CoalesceCriteria.COALESCE_BRIGGS_AND_GEORGE;

			Tiler tiler = new Tiler(boostedTiles);

			output.add(new AsmDirective("# Compiled using \"xic\" by Owen Arden, Catalin Dumitru, Wenzel Jakob, and Danfeng Zhang"));
			output.add(new AsmDirective("# Command line: " + argsStr.substring(0, argsStr.length()-1)));
			output.add(new AsmDirective(""));
			
			output.add(new AsmDirective(".file \"" + inputFile + "\""));
			if (syntax == AssemblySyntax.INTEL_SYNTAX)
				output.add(new AsmDirective(".intel_syntax noprefix"));
			else
				output.add(new AsmDirective(".att_syntax prefix"));
			output.add(new AsmDirective(".text"));
			Map<String, CFGNode> asmCFG = new HashMap<String, CFGNode>();

			RegAlloc ra = new GraphColoringRegAlloc(stats, dumpWebs, cCriteria, typeChecker.getNameGenerator());
			for (String funcName : irCompUnit.functions().keySet()) {
				IRFuncDecl funcDecl = irCompUnit.functions().get(funcName);
				stats.startTimer("tile");
				Tile tile = tiler.tile(funcDecl);
				stats.stopTimer("tile");

				InstrList il = new InstrList();
				stats.startTimer("translate");
				tile.translate(il, typeChecker.getNameGenerator());
				stats.stopTimer("translate");				
				
				Set<CFGNode> cfg = CFGGenerator.createCFG(funcName, il).flatten();
				List<DataFlowAnalysis> dfas = new ArrayList<DataFlowAnalysis>();

				for(String dumpDFA : dumpDFAs) {
					DataFlowAnalysis dfa;
		
					/* Dump data flow analyses if requested (pseudo assembly-level) */
					if (dumpDFA.equals("LiveVariablesAsm") || dumpDFA.equals("ReachingDefinitions")) {
						
						if (dumpDFA.equals("LiveVariablesAsm"))
							dfa = new LiveVariables(cfg);
						else
							dfa = new ReachingDefinitions(cfg);
						dfa.solve();
						dfas.add(dfa);
					}
				}
				if(dfas.size()>0) {
					DotGenerator dg = new DotGenerator();
					dg.setMultiDataFlowAnalysis(dfas, cfg);
	
					try {
						PrintWriter pw = new PrintWriter(new FileWriter("dfa_" + funcName + ".dot"));
						pw.print(dg.generate(cfg));
						pw.close();
					} catch (IOException e) {
						throw new InternalCompilerException(
								"Could not generate debug output", e);
					}
				}
				if (dumpCFG.equals("asm")) 
					asmCFG.put(funcName, CFGGenerator.createCFG(funcName, il));

				if (runRA) {
					InstrList ia = ra.regAlloc(funcName, funcDecl, il, true);

					/* Peephole optimizations */
					if (optimizations.contains("peep")) {
						stats.startTimer("peep");
						ia = PeepholeOptimizer.transform(stats, funcName, ia);
						stats.stopTimer("peep");
					}
					output.addAll(ia);
				} else {
					output.addAll(il);
				}
			}

			if (irCompUnit.hasExtraSequence()) {
				String funcName = "extra_sequence";
				output.add(new AsmDirective(""));
				stats.startTimer("tile");
				Tile tile = tiler.tile(irCompUnit.extraSequence());
				stats.stopTimer("tile");

				InstrList il = new InstrList();
				stats.startTimer("translate");
				tile.translate(il, typeChecker.getNameGenerator());
				stats.stopTimer("translate");
	
				if (dumpCFG.equals("asm"))
					asmCFG.put(funcName, CFGGenerator.createCFG(funcName, il));

				if (runRA) {
					output.addAll(ra.regAlloc(funcName, null, il, false));
				} else {
					output.addAll(il);
				}
			}

			if (dumpCFG.equals("asm"))				
				System.out.println(DotGenerator.toDot(asmCFG));

			PrintWriter pw = new PrintWriter(new FileWriter(outputFile));  
			pw.print(output.toString());
			pw.close();

			if (displayStats)
				System.out.println(stats.toString());

			return 0;
		} catch (IOException e) {
			System.out.println("I/O error:");
			e.printStackTrace(System.out);
			return -1;
		}
	}
}
