package xic.analysis;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.RegAlloc;
import xic.ir.IRCJump;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRExp;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMove;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;


public class AvailableCopies extends DataFlowAnalysis {
	public static class CPPair {
		String v1;
		String v2;

		public CPPair(String v1, String v2) {
			this.v1 = v1;
			this.v2 = v2;
		}

		public boolean equals(Object obj) {
			if (obj instanceof CPPair) {
				CPPair v = (CPPair) obj;
				return (v1.equals(v.v1) && v2.equals(v.v2));
			}
			else return false;
		}

		public int hashCode() {
			return 17 + v1.hashCode() + v2.hashCode() * 37;
		}
		
		public String toString() {
			return v1 + "->" + v2;
		}
	};

	/// Value class used during copy propagation analysis
	public static class CPValue implements Value {
		/* List of variable pairs */
		private Set<CPPair> pairs;

		public CPValue() {
			pairs = new HashSet<CPPair>();
		}
		
		public CPValue(Set<CPPair> pairs) {
			this.pairs = new HashSet<CPPair>(pairs);
		}

		public Value copy() {
			return new CPValue(pairs);
		}

		public void meet(Value v) {
			pairs.retainAll(((CPValue) v).pairs);
		}
		
		public String toString() {
			return pairs.toString();
		}

		public int hashCode() {
			return pairs.hashCode();
		}
		
		public Set<CPPair> pairs() {
			return pairs;
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof CPValue))
				return false;
			return this.pairs.equals(((CPValue) obj).pairs);
		}

		public void setNullMeet() {
			pairs.clear();
		}
	};

	/// Dataflow-related information associated with each node
	public static class CPFlowInfo extends FlowInfo {
		public Set<CPPair> copy;
		public Set<String> kill;

		public CPFlowInfo(Object o) {
			CFGNode node = (CFGNode)o;
			// for kill, actually don't need to store every possible pair, only need one variable
			copy = new HashSet<CPPair>();
			kill = new HashSet<String>();
			entry = new CPValue();
			exit = new CPValue();
			computeCopyKill(node, copy, kill);
		}
	};

	public AvailableCopies(Set<CFGNode> nodes) {
		super(nodes);

		dir = Direction.FORWARD;
		Set<CPPair> allCopyPairs = new HashSet<CPPair>(); 
		for (CFGNode node : nodes)
			computeCopyKill(node, allCopyPairs, null);
		top = new CPValue(allCopyPairs);
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new CPFlowInfo(node);
	}

	public CPValue in(CFGNode node) {
		CPFlowInfo info = (CPFlowInfo) flowInfo.get(node);
		return (CPValue) info.entry;
	}
	
	public CPValue out(CFGNode node) {
		CPFlowInfo info = (CPFlowInfo) flowInfo.get(node);
		return (CPValue) info.exit;
	}

	private static void computeCopyKill(CFGNode node, Set<CPPair> copy, Set<String> kill) {
		if (node.value() instanceof IRStmt) {
			IRStmt irStmt = (IRStmt) node.value();
			if (irStmt instanceof IRMove) {
				IRMove move = (IRMove) irStmt;
				if (move.target() instanceof IRTemp) {
					String targetName = ((IRTemp) move.target()).name();

					// x = y
					if (move.expr() instanceof IRTemp) {
						String srcName = ((IRTemp) move.expr()).name();
						/* Don't propagate assignments from registers */
						if (RegAlloc.isAbstractRegisterName(srcName))
							return;
						copy.add(new CPPair(targetName, srcName));
						if (kill != null)
							kill.add(targetName);
					} else {
						// x = e
						if (kill != null)
							kill.add(targetName);
					}
				}
			} else if(irStmt instanceof IRDataSeq
						|| irStmt instanceof IRJump
						|| irStmt instanceof IRLabel	
						|| irStmt instanceof IRSeq
						|| irStmt instanceof IRDecl
						|| irStmt instanceof IRExp
						|| irStmt instanceof IRCJump) {
				// don't do anything
			} else {
				throw new InternalCompilerException("Unknown IRStmt: "+irStmt);
			}
		}
	}

	// transform function: out(n) = gen(n) cup (in(n) - kill (n))
	protected void F(FlowInfo f) {
		CPFlowInfo fi = (CPFlowInfo) f;
		Set<CPPair> entry = ((CPValue) fi.entry).pairs();
		Set<CPPair> exit = ((CPValue) fi.exit).pairs();
		
		exit.clear();
		exit.addAll(entry);
		Iterator<CPPair> it = exit.iterator();
		while (it.hasNext()) {
			CPPair pair = it.next();
			if (fi.kill.contains(pair.v1) || fi.kill.contains(pair.v2))
				it.remove();
		}
		exit.addAll(fi.copy);
	}
}
