package xic.analysis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.RegAlloc;
import xic.codegen.matchers.AddrModeHelper;
import xic.ir.IRBinOp;
import xic.ir.IRCJump;
import xic.ir.IRCall;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.StructHashSet;


public class AvailableExpressions extends DataFlowAnalysis {
	StructHashSet<IRNode> allExprs; 

	public static int countOccurrences(IRExpr sub, IRNode node) {
		int count=0;
		HashSet<IRExpr> exprs= new HashSet<IRExpr>();

		if(node instanceof IRStmt) {
			IRStmt irStmt = (IRStmt)node;
			if(irStmt instanceof IRExp) {
				IRExp exp = (IRExp) irStmt;
				exprs.add(exp.expr());
			} else if(irStmt instanceof IRMove) {	
				IRMove move = (IRMove) irStmt;
				if (move.target() instanceof IRMem) {
					IRMem mem = (IRMem) move.target();
					exprs.add(mem.expr());
				}
				exprs.add(move.expr());
			/* Include conditional expressions */
			} else if(irStmt instanceof IRCJump) {
				IRCJump cjump = (IRCJump) irStmt;
				exprs.add(cjump.expr());
				
			} else if(irStmt instanceof IRDataSeq
						|| irStmt instanceof IRJump
						|| irStmt instanceof IRLabel	
						|| irStmt instanceof IRSeq) {
			} else
				throw new InternalCompilerException("Unknown IRStmt: "+irStmt);
		} else {
			exprs.add((IRExpr)node);
		}
		for(IRExpr expr : exprs) {
			if(sub.equalsTree(expr))
				count+=1;
			else if(expr instanceof IRBinOp) {
			
				IRBinOp binOp = (IRBinOp)expr;
				count+=countOccurrences(sub, binOp.left());
				count+=countOccurrences(sub, binOp.right());
			
			} else if(expr instanceof IRCall) {
				
				IRCall call = (IRCall)expr;
				for(IRExpr e : call.params())
					count+=countOccurrences(sub, e);
				
				count+=countOccurrences(sub, call.target());
			} else if(expr instanceof IRMem) {
				IRMem mem = (IRMem)expr;
				count+=countOccurrences(sub, mem.expr());
			}
		}
		return count;
	}
	public static StructHashSet<IRNode> expressions(IRExpr expr) {
		StructHashSet<IRNode> exprSet = new StructHashSet<IRNode>();
		if(AddrModeHelper.hasZeroCostAddressingMode(expr))
			return exprSet;
		if(expr instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp)expr;
			exprSet.add(expr);
			exprSet.addAll(expressions(binOp.left()));
			exprSet.addAll(expressions(binOp.right()));
		} else if(expr instanceof IRCall) {
			IRCall call = (IRCall)expr;
			for(IRExpr e : call.params())
				exprSet.addAll(expressions(e));
			exprSet.addAll(expressions(call.target()));
		} else if(expr instanceof IRMem) {
			IRMem mem = (IRMem)expr;
			exprSet.add(expr);
			exprSet.addAll(expressions(mem.expr()));
		} 
		return exprSet;
	}

	public static String victim(IRExpr target) {	
		if (target instanceof IRTemp) {
			return ((IRTemp) target).name();
		} else if (target instanceof IRMem) {
			return RegAlloc.alias((IRMem) target);
		}
		return null;
	}

	public AvailableExpressions(Set<CFGNode> nodes) {
		super(nodes);
		initFlowInfo();
		dir = Direction.FORWARD;		
		allExprs = new StructHashSet<IRNode>();
		for(CFGNode node: nodes) {
			AvailExprInfo aefi = (AvailExprInfo) flowInfo.get(node);
			allExprs.addAll(aefi.eval);			
		}
		top = new ExprValue(allExprs);


	}
	protected void removeKill(List<String> kill, StructHashSet<IRNode> exprs) {
		/* remove all expressions that depend on kill */
		StructHashSet<IRNode> killset = computeKill(kill, exprs);
		exprs.removeAll(killset);
	}

	protected  StructHashSet<IRNode> computeKill(List<String> kills, StructHashSet<IRNode> exprs) {
		/* remove all expressions that depend on kill */
		StructHashSet<IRNode> killset = new StructHashSet<IRNode>();
		HashSet<String> names = new HashSet<String>();
		for(IRNode expr : exprs) {	
			names.clear();
			expr.findReadTemps(names);
			
			for (String kill : kills) {
				if (names.contains(kill)) {
					killset.add(expr);
					break;
				}
			}
		}
		return killset;
	}

	protected void F(FlowInfo f) {
		AvailExprInfo fi = (AvailExprInfo) f;
		StructHashSet<IRNode> entry = ((ExprValue) f.entry).exprs;
		StructHashSet<IRNode> exit = ((ExprValue) f.exit).exprs;
		
		exit.clear();
		exit.addAll(entry);		
		removeKill(fi.kill, exit);
		exit.addAll(fi.eval);
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new AvailExprInfo(node);
	}
	
	public StructHashSet<IRNode> in(CFGNode node) {
		AvailExprInfo info = (AvailExprInfo)flowInfo.get(node);
		return ((ExprValue) info.entry).exprs;
	}

	public StructHashSet<IRNode> out(CFGNode node) {
		AvailExprInfo info = (AvailExprInfo)flowInfo.get(node);
		return ((ExprValue) info.exit).exprs;
	}

	public StructHashSet<IRNode> eval(CFGNode node) {
		return ((AvailExprInfo) flowInfo.get(node)).eval;
	}
	
	public List<String> kill(CFGNode node) {
		AvailExprInfo info = (AvailExprInfo)flowInfo.get(node);
		return info.kill;
	}
	
	public static class ExprValue implements Value {
		StructHashSet<IRNode> exprs;

		public ExprValue() {
			exprs = new StructHashSet<IRNode>();
		}

		public ExprValue(StructHashSet<IRNode> exprs) {
			this.exprs = new StructHashSet<IRNode>(exprs);
		}

		public Value copy() {
			return new ExprValue(exprs);
		}

		/* meet is the intersection of the available expressions */
		public void meet(Value v) {
			exprs.retainAll(((ExprValue) v).exprs);
		}

		public int hashCode() {
			return exprs.hashCode();
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof ExprValue))
				return false;
			return exprs.equals(((ExprValue) obj).exprs);
		}
		public String toString(){
			String str = "<";
			for(IRNode expr : exprs)
				str+=" ["+ expr.toString()+"]\n";
			str+=">";
			return str; 
		}

		public void setNullMeet() {
			exprs.clear();
		}
	}

	// Dataflow-related information associated with each node
	public class AvailExprInfo extends FlowInfo {
		StructHashSet<IRNode> eval;
		List<String> kill;
		CFGNode node;

		public AvailExprInfo(Object o) {
			super();
			node = (CFGNode) o;
			entry = new ExprValue();
			exit = new ExprValue();
			kill = new ArrayList<String>();

			if (node.value() instanceof String){
				eval = new StructHashSet<IRNode>();
			} else {
				IRStmt irStmt = (IRStmt) node.value();

				if (irStmt instanceof IRMove) {	
					IRMove move = (IRMove) irStmt;
					eval = expressions(move.expr());
					if (move.target() instanceof IRMem)
						eval.addAll(expressions(((IRMem) move.target()).expr()));
					if (move.expr() instanceof IRCall) {
						kill.add(RegAlloc.CALL_ARG_SCRATCH);
						kill.add(RegAlloc.CALL_RV_SCRATCH);
						kill.add(RegAlloc.MEM_NAME);
					}
					kill.add(victim(move.target()));
					/* Include conditional expressions */
				} else if(irStmt instanceof IRCJump) {
					IRCJump cjump = (IRCJump) irStmt;
					IRExpr expr = cjump.expr();
					eval = expressions(expr);
				} else if(irStmt instanceof IRDataSeq
							|| irStmt instanceof IRDecl
							|| irStmt instanceof IRJump
							|| irStmt instanceof IRLabel	
							|| irStmt instanceof IRSeq) {
					eval = new StructHashSet<IRNode>();
				} else
					throw new InternalCompilerException("Unknown IRStmt: "+irStmt);				
			}
		}
	};

}
