package xic.analysis;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import xic.Configuration;
import xic.InternalCompilerException;
import xic.ir.IRBinOp;
import xic.ir.IRBinOp.OpType;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRDataSeq;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMem;
import xic.ir.IRMem.MemType;
import xic.ir.IRMove;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;


/**
 * Performs a bounding interval analysis similar to the one described in PS4.
 * The underlying numerical representation also handles interval endpoints of
 * the form "length(array_name)+offset".
 */
public class BoundingIntervals extends DataFlowAnalysis {
	private static final int MAX_RANGE = 50;

	private enum TriBool { // Tri-valued boolean type used below
		NO, YES, MAYBE
	}

	/**
	 * Underlying partially ordered data type: either refers to a number or the
	 * (unknown) length of a named array plus an offset.
	 */
	public static class NumberType {
		private static final NumberType P_INF = new NumberType(Long.MAX_VALUE);
		private static final NumberType N_INF = new NumberType(Long.MIN_VALUE);
		private static final NumberType ZERO = new NumberType(0);
		private static final NumberType ONE = new NumberType(1);

		private long value;
		private String name;

		public NumberType(String name) {
			this.name = name;
			this.value = 0;
		}

		public NumberType(String name, long value) {
			this.name = name;
			this.value = value;
		}

		public NumberType(long value) {
			this.name = null;
			this.value = value;
		}

		public boolean isLength() {
			return name != null;
		}

		public boolean isNumber() {
			return name == null;
		}

		public boolean isPosInf() {
			return isNumber() && value == P_INF.value;
		}

		public boolean isNegInf() {
			return isNumber() && value == N_INF.value;
		}

		public String name() {
			return name;
		}

		public long value() {
			return value;
		}

		public boolean equals(Object obj) {
			if (!(obj instanceof NumberType))
				return false;
			NumberType other = (NumberType) obj;
			if (value != other.value)
				return false;
			if (isLength() && !other.isLength() || !isLength()
					&& other.isLength())
				return false;
			if (isLength() && !name.equals(other.name))
				return false;
			return true;
		}

		public TriBool ge(NumberType other) {
			if (isNumber() && other.isNumber()) {
				return value >= other.value() ? TriBool.YES : TriBool.NO;
			} else if (isLength()) {
				if (other.isPosInf())
					return TriBool.NO; // length >= +inf == false
				else if (other.isNegInf())
					return TriBool.YES; // length >= -inf == true
				else if (other.isNumber() && other.value() - value <= 0)
					return TriBool.YES; // length(a)+offset >= value when
										// value-offset<=0
				else if (isLength() && other.isLength()
						&& name.equals(other.name))
					/* length(a)+offset1 >= length(a)+offset 2? */
					return value() >= other.value() ? TriBool.YES : TriBool.NO;
			} else if (isNumber() && other.isLength()) {
				if (isPosInf())
					return TriBool.YES; // +inf >= length== true
				else if (isNegInf())
					return TriBool.NO; // -inf >= length == false
			}
			return TriBool.MAYBE;
		}

		public TriBool gt(NumberType other) {
			if (isNumber() && other.isNumber()) {
				return value > other.value() ? TriBool.YES : TriBool.NO;
			} else if (isLength()) {
				if (other.isPosInf())
					return TriBool.NO; // length > +inf == false
				else if (other.isNegInf())
					return TriBool.YES; // length > -inf == true
				else if (other.isNumber() && other.value() - value < 0)
					return TriBool.YES; // length(a)+offset >= value when
										// value-offset<=0
				else if (isLength() && other.isLength()
						&& name.equals(other.name))
					/* length(a)+offset1 >= length(a)+offset 2? */
					return value() > other.value() ? TriBool.YES : TriBool.NO;
			} else if (isNumber() && other.isLength()) {
				if (isPosInf())
					return TriBool.YES; // +inf > length== true
				else if (isNegInf())
					return TriBool.NO; // -inf > length == false
			}
			return TriBool.MAYBE;
		}

		public TriBool le(NumberType other) {
			return negate(gt(other));
		}

		private TriBool negate(TriBool value) {
			if (value == TriBool.YES)
				return TriBool.NO;
			else if (value == TriBool.NO)
				return TriBool.YES;
			else
				return TriBool.MAYBE;
		}

		public TriBool lt(NumberType other) {
			return negate(ge(other));
		}

		public static NumberType min(NumberType n1, NumberType n2) {
			if (n1 == null || n2 == null)
				return null;
			else if (n1.ge(n2) == TriBool.YES)
				return n2;
			else if (n2.ge(n1) == TriBool.YES)
				return n1;
			else
				return null; /* Too little information to compare */
		}

		public static NumberType max(NumberType n1, NumberType n2) {
			if (n1 == null || n2 == null)
				return null;
			else if (n1.ge(n2) == TriBool.YES)
				return n1;
			else if (n2.ge(n1) == TriBool.YES)
				return n2;
			else
				return null; /* Too little information to compare */
		}

		public static NumberType add(NumberType n1, NumberType n2) {
			if (n1 == null || n2 == null)
				return null;
			else if (n1.isNegInf() && n2.isPosInf() || n1.isPosInf()
					&& n2.isNegInf())
				return null; // inf - inf not valid
			if (n1.isPosInf() || n2.isPosInf())
				return NumberType.P_INF;
			else if (n1.isNegInf() || n2.isNegInf())
				return NumberType.N_INF;
			else if (n1.isLength() && n2.isLength())
				return null; // cannot add two lengths
			else if (n1.isLength())
				return new NumberType(n1.name(), n1.value() + n2.value());
			else if (n2.isLength())
				return new NumberType(n2.name(), n1.value() + n2.value());
			else
				return new NumberType(n1.value() + n2.value());
		}

		public static NumberType sub(NumberType n1, NumberType n2) {
			if (n1 == null || n2 == null)
				return null;
			else if (n1.isPosInf() && n2.isPosInf())
				return null; // inf - inf not allowd
			else if (n1.isNegInf() && n2.isNegInf())
				return null; // inf - inf not allowed
			else if (n1.isPosInf() || n2.isNegInf())
				return P_INF;
			else if (n1.isNegInf() || n2.isPosInf())
				return N_INF;
			else if (n1.isLength() && n2.isLength()) {
				if (n1.name.equals(n2.name))
					return new NumberType(n1.value - n2.value);
				else
					return null;
			} else if (n1.isLength()) {
				return new NumberType(n1.name, n1.value - n2.value);
			} else if (n2.isLength()) {
				return null;
			} else {
				return new NumberType(n1.value - n2.value);
			}
		}

		public static NumberType mul(NumberType n1, NumberType n2) {
			if (n1 == null || n2 == null)
				return null;
			else if (n1.equals(ZERO) || n2.equals(ZERO))
				return ZERO;
			else if (n1.isLength() || n2.isLength())
				return null; // cannot use lengths in multiplications
			else if (n1.isPosInf() && n2.value() > 0)
				return P_INF;
			else if (n1.isPosInf() && n2.value() < 0)
				return N_INF;
			else if (n1.isNegInf() && n2.value() > 0)
				return N_INF;
			else if (n1.isNegInf() && n2.value() < 0)
				return P_INF;
			else if (n2.isPosInf() && n1.value() > 0)
				return P_INF;
			else if (n2.isPosInf() && n1.value() < 0)
				return N_INF;
			else if (n2.isNegInf() && n1.value() > 0)
				return N_INF;
			else if (n2.isNegInf() && n1.value() < 0)
				return P_INF;
			else
				return new NumberType(n1.value() * n2.value());
		}

		public int hashCode() {
			int result = 17;
			if (isNumber())
				result = result * 37 + (int) value;
			else
				result = result * 37 + name.hashCode();
			return result;
		}

		public boolean containsLength(String tempName) {
			return (!isNumber() && tempName.equals(name));
		}

		private NumberType clamp() {
			if (value < -MAX_RANGE)
				// for both arrays and numbers
				return NumberType.N_INF;
			else if (value > MAX_RANGE)
				// for both arrays and numbers
				return NumberType.P_INF;
			else
				return this;
		}

		public String toString() {
			if (isNumber()) {
				if (value == NumberType.N_INF.value)
					return "-inf";
				else if (value == NumberType.P_INF.value)
					return "+inf";
				else
					return String.valueOf(value);
			} else if (value < 0)
				return "length(" + name + ")" + value;
			else if (value > 0)
				return "length(" + name + ")+" + value;
			else
				return "length(" + name + ")";
		}
	};

	public static class Interval {
		private NumberType l, r;

		public Interval() {
			l = NumberType.P_INF;
			r = NumberType.N_INF;
		}

		public Interval(NumberType lv, NumberType rv) {
			set(lv, rv);
		}

		public Interval(Interval interval) {
			l = interval.l;
			r = interval.r;
		}

		public Interval(NumberType value) {
			l = r = value;
		}

		public Interval clamp() {
			return new Interval(l.clamp(), r.clamp());
		}

		public void set(NumberType lv, NumberType rv) {
			l = lv;
			r = rv;
		}

		public boolean isTop() {
			return l.isNumber() && r.isNumber() && l.value() > r.value();
		}

		public boolean isBottom() {
			return (l.equals(NumberType.N_INF) && r.equals(NumberType.P_INF));
		}

		public boolean isSingleElement() {
			return l.equals(r);
		}

		public NumberType l() {
			return l;
		}

		public NumberType r() {
			return r;
		}

		// set the greatest lower bound of two values
		public void setGLB(Interval v1, Interval v2) {
			l = NumberType.min(v1.l, v2.l);
			r = NumberType.max(v1.r, v2.r);
			if (l == null)
				l = NumberType.N_INF;
			if (r == null)
				r = NumberType.P_INF;
		}

		public String toString() {
			if (isTop())
				return "top";
			return "<" + l + ", " + r + ">";
		}

		public boolean equals(Object other) {
			if (!(other instanceof Interval))
				return false;
			Interval io = (Interval) other;
			return io.l.equals(l) && io.r.equals(r);
		}

		public int hashCode() {
			int result = 17;
			result = result * 37 + l.hashCode();
			result = result * 37 + r.hashCode();
			return result;
		}

		public void intersect(NumberType lOther, NumberType rOther) {
			// System.out.print("Intersect("+this +", " + new Interval(lOther,
			// rOther) +")=");
			NumberType newL = NumberType.max(lOther, l);
			NumberType newR = NumberType.min(rOther, r);

			if (newL == null)
				newL = lOther;
			if (newR == null)
				newR = rOther;

			set(newL, newR);
			// System.out.println(this);
		}

		// Conservative interval when treated as a 32bit unsigned integer
		public Interval unsignedInterval() {
			if (l.isNumber() && l.value() < 0)
				return new Interval(NumberType.ZERO, NumberType.P_INF);
			else
				return this;
		}

		public boolean containsLength(String tempName) {
			return l.containsLength(tempName) || r.containsLength(tempName);
		}

		public void removeLength(String tempName) {
			if (l.containsLength(tempName))
				l = NumberType.N_INF;
			if (r.containsLength(tempName))
				r = NumberType.P_INF;
		}
	}

	public static class BIValue implements Value {
		private Map<String, Interval> intervals;
		private Map<String, Interval> arrayLengths;

		public BIValue() {
			intervals = new TreeMap<String, Interval>();
			arrayLengths = new TreeMap<String, Interval>();
		}

		public Map<String, Interval> intervals() {
			return intervals;
		}

		public Map<String, Interval> arrayLengths() {
			return arrayLengths;
		}

		public void meet(Value v) {
			Map<String, Interval> otherIntervals = ((BIValue) v).intervals;
			Map<String, Interval> otherArrayLengths = ((BIValue) v).arrayLengths;

			for (String key : otherIntervals.keySet()) {
				Interval v1 = otherIntervals.get(key);
				Interval v2 = intervals.get(key);
				if (v2 == null)
					intervals.put(key, v1);
				else
					v2.setGLB(v1, v2);
			}

			for (String key : otherArrayLengths.keySet()) {
				Interval v1 = otherArrayLengths.get(key);
				Interval v2 = arrayLengths.get(key);

				if (v2 == null)
					arrayLengths.put(key, v1);
				else
					v2.setGLB(v1, v2);
			}
		}

		public Value copy() {
			BIValue r = new BIValue();
			for (String key : intervals.keySet())
				r.intervals.put(key, new Interval(intervals.get(key)));
			for (String key : arrayLengths.keySet())
				r.arrayLengths.put(key, new Interval(arrayLengths.get(key)));
			return r;
		}

		public String toString() {
			StringBuffer buf = new StringBuffer("V[");
			boolean first = true;
			for (String name : intervals.keySet()) {
				Interval i = intervals.get(name);
				if (!i.isBottom()) {
					if (!first)
						buf.append(", ");
					buf.append(name + "=" + i);
					first = false;
				}
			}
			buf.append("], L[");
			first = true;
			for (String name : arrayLengths.keySet()) {
				Interval i = arrayLengths.get(name);
				if (!i.isBottom()) {
					if (!first)
						buf.append(", ");
					buf.append(name + "=" + i);
					first = false;
				}
			}
			return buf.append(']').toString();
		}

		public boolean equals(Object obj) {
			if (!(obj instanceof BIValue))
				return false;
			return intervals.equals(((BIValue) obj).intervals)
					&& arrayLengths.equals(((BIValue) obj).arrayLengths);
		}

		public int hashCode() {
			return intervals.hashCode() + 37 * arrayLengths.hashCode();
		}

		public void setNullMeet() {
			intervals.clear();
//			intervals.put(IRGenerator.FP_NAME,
//					new Interval(NumberType.N_INF, NumberType.P_INF));
			arrayLengths.clear();
		}
	};

	public BoundingIntervals(Set<CFGNode> nodes) {
		super(nodes);

		dir = Direction.FORWARD;
		top = new BIValue();
	}

	public class BIFlowInfo extends FlowInfo {
		private String tempName;
		private IRExpr expr = null;
		private boolean kill = false;
		private Set<String> usedTemps = new HashSet<String>();
		private CFGNode node;

		public BIFlowInfo(Object o) {
			CFGNode node = (CFGNode) o;
			this.entry = new BIValue();
			this.exit = new BIValue();
			this.node = node;

			if (node.value() instanceof IRStmt) {
				IRStmt irStmt = (IRStmt) node.value();
				if (irStmt instanceof IRMove) {
					IRMove move = (IRMove) irStmt;
					if (move.target() instanceof IRTemp) {
						tempName = ((IRTemp) move.target()).name();
						if (move.expr().containsCalls()
								|| move.expr().containsMemReads()) {
							kill = true;
						} else {
							expr = move.expr();
							move.expr().findReadTemps(usedTemps);
						}
					}
				} else if (irStmt instanceof IRDataSeq
						|| irStmt instanceof IRJump
						|| irStmt instanceof IRLabel || irStmt instanceof IRSeq
						|| irStmt instanceof IRExp || irStmt instanceof IRCJump) {
					// don't do anything
				} else
					throw new InternalCompilerException("Unknown IRStmt: "
							+ irStmt);
			}
		}
	};

	public BIValue in(CFGNode node) {
		BIFlowInfo info = (BIFlowInfo) flowInfo.get(node);
		return (BIValue) info.entry;
	}

	public BIValue out(CFGNode node) {
		BIFlowInfo info = (BIFlowInfo) flowInfo.get(node);
		return (BIValue) info.exit;
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new BIFlowInfo(node);
	}

	protected void F(FlowInfo f) {
		BIFlowInfo fi = (BIFlowInfo) f;
		BIValue entry = null;
		CFGNode node = fi.node;

		if (node != null && node.inEdges().size() > 0) {
			for (CFGNode np : node.inEdges()) {
				BIValue inValue = (BIValue) flowInfo.get(np).exit;

				/*
				 * If this node is the consequent/alternative of an IRCJump and
				 * the condition expression is usable, introduce this
				 * information using an interval
				 */
				if (np.value() instanceof IRCJump) {
					IRCJump cJump = (IRCJump) np.value();
					IRExpr expr = cJump.expr();
					boolean trueBranch;

					if (cJump.trueLabel() == null
							&& !(node.value() instanceof IRLabel || node
									.value() instanceof String)
							|| cJump.trueLabel() != null
							&& node.value() instanceof IRLabel
							&& ((IRLabel) node.value()).name().equals(
									cJump.trueLabel())) {
						trueBranch = true;
					} else {
						trueBranch = false;
					}

					inValue = (BIValue) inValue.copy();
					extractInvariants(expr, inValue.intervals(), trueBranch);
				}

				if (entry == null) {
					entry = (BIValue) inValue.copy();
				} else {
					// System.out.print("MEET(\n\t" + entry + ",\n\t" + inValue
					// + ")=\n\t");
					entry.meet(inValue);
					// System.out.println(entry.toString());
				}
			}
		} else {
			entry = (BIValue) top.copy();
			entry.setNullMeet();
		}
		fi.entry = entry;

		Map<String, Interval> exit = ((BIValue) fi.exit).intervals;
		Map<String, Interval> exitLengths = ((BIValue) fi.exit).arrayLengths;
		exit.clear();
		exit.putAll(entry.intervals);
		exitLengths.putAll(entry.arrayLengths);

		/* Does this node assign an array length? */
		if (node != null && node.value() instanceof IRMove
				&& ((IRMove) node.value()).target() instanceof IRMem) {
			IRMove move = (IRMove) node.value();
			IRMem mem = (IRMem) move.target();

			if (mem.expr() instanceof IRBinOp
					&& mem.memType() == MemType.IMMUTABLE
					&& ((IRBinOp) mem.expr()).opType() == IRBinOp.OpType.SUB
					&& ((IRBinOp) mem.expr()).left() instanceof IRTemp
					&& ((IRBinOp) mem.expr()).right() instanceof IRConst
					&& ((IRConst) ((IRBinOp) mem.expr()).right()).value() == Configuration.WORD_SIZE) {
				String tempName = ((IRTemp) ((IRBinOp) mem.expr()).left())
						.name();
				exitLengths.put(tempName, evalUnclamped(move.expr(),
						entry.intervals, exitLengths));
			}
		}

		/*
		 * Is this a temp->temp move? Forward array lengths if the RHS-expr is
		 * an array
		 */
		if (node != null && node.value() instanceof IRMove
				&& ((IRMove) node.value()).expr() instanceof IRTemp
				&& ((IRMove) node.value()).target() instanceof IRTemp) {
			String from = ((IRTemp) ((IRMove) node.value()).expr()).name();
			String to = ((IRTemp) ((IRMove) node.value()).target()).name();
			if (exitLengths.containsKey(from))
				exitLengths.put(to, exitLengths.get(from));
		}

		if (fi.tempName != null) {
			if (fi.kill) {
				exit.put(fi.tempName, new Interval(NumberType.N_INF,
						NumberType.P_INF));
				exitLengths.remove(fi.tempName);
				for (String varName : new HashSet<String>(exit.keySet())) {
					Interval i = exit.get(varName);
					if (i.containsLength(fi.tempName)) {
						i = new Interval(i);
						i.removeLength(fi.tempName);
						exit.put(varName, i);
					}
				}
			} else {
				exit.put(fi.tempName, eval(fi.expr, entry.intervals, null));
			}
		}
	}

	private void extractInvariants(IRExpr expr,
			Map<String, Interval> intervals, boolean inTrueBranch) {
		if (expr instanceof IRTemp) {
			intervals.put(((IRTemp) expr).name(), new Interval(
					inTrueBranch ? NumberType.ONE : NumberType.ZERO));
		} else if (expr instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) expr;

			/* Recursively extract invariants from boolean expressions */
			switch (((IRBinOp) expr).opType()) {
			case AND:
				if (inTrueBranch) {
					extractInvariants(binOp.left(), intervals, true);
					extractInvariants(binOp.right(), intervals, true);
				}
				break;
			case OR:
				if (!inTrueBranch) {
					extractInvariants(binOp.left(), intervals, false);
					extractInvariants(binOp.right(), intervals, false);
				}
				break;
			case XOR:
				if (binOp.left() instanceof IRConst
						&& ((IRConst) binOp.left()).value() == 1)
					extractInvariants(binOp.right(), intervals, !inTrueBranch);
				if (binOp.right() instanceof IRConst
						&& ((IRConst) binOp.right()).value() == 1)
					extractInvariants(binOp.left(), intervals, !inTrueBranch);
				break;
			default:
			}

			if (binOp.left() instanceof IRTemp && binOp.isComparison()) {
				String tempName = ((IRTemp) binOp.left()).name();
				Interval left = intervals.get(tempName);
				Interval right = evalUnclamped(binOp.right(), intervals, null);

				if (left == null || left.isTop()) {
					return;
				}
				Interval evaluated = evalUnclamped(binOp, intervals, null);

				if (binOp.isComparison() && evaluated != null
						&& evaluated.isSingleElement()) {
					if (inTrueBranch && evaluated.l().value() == 0
							|| !inTrueBranch && evaluated.l().value() == 1) {
						/* This will never happen - remove the interval */
						intervals.remove(tempName);
						return;
					}
				}

				switch (binOp.opType()) {
				case LT:
					if (inTrueBranch)
						left.intersect(NumberType.N_INF, NumberType.sub(right.l(), NumberType.ONE));
					else
						left.intersect(right.l(), NumberType.P_INF);
					break;
				case ULT:
					right = right.unsignedInterval();
					if (inTrueBranch)
						left.intersect(NumberType.ZERO, NumberType.sub(right.l(), NumberType.ONE));
					break;
				case GT:
					if (inTrueBranch)
						left.intersect(NumberType.sub(right.r(), NumberType.ONE),
								NumberType.P_INF);
					else
						left.intersect(NumberType.N_INF, right.r());
					break;
				case LE:
					if (inTrueBranch)
						left.intersect(NumberType.N_INF, right.l());
					else
						left.intersect(NumberType.add(right.l(), NumberType.ONE),
								NumberType.P_INF);
					break;
				case GE:
					if (inTrueBranch)
						left.intersect(right.r(), NumberType.P_INF);
					else
						left.intersect(NumberType.N_INF, NumberType.sub(right.r(), NumberType.ONE));
					break;
				case EQ:
					if (inTrueBranch)
						left.intersect(right.l(), right.r());
					break;
				case NEQ:
					if (!inTrueBranch)
						left.intersect(right.l(), right.r());
					break;
				}
			}
		}
	}

	public static Interval eval(IRExpr expr, Map<String, Interval> intervals, Map<String, Interval> lengthExprs) {
		return evalUnclamped(expr, intervals, lengthExprs).clamp();
	}

	/**
	 * Evaluate an expression given intervals for all TEMPS that it uses
	 */
	public static Interval evalUnclamped(IRExpr expr,
			Map<String, Interval> intervals,
			Map<String, Interval> lengthExprs) {
		if (expr instanceof IRConst) {
			return new Interval(new NumberType(((IRConst) expr).value()));
		} else if (expr instanceof IRTemp) {
			String tempName = ((IRTemp) expr).name();
			Interval interval = intervals.get(tempName);
			if (interval == null)
				return new Interval(); // top
			return new Interval(interval);
		} else if (expr instanceof IRMem
				&& ((IRMem) expr).expr() instanceof IRBinOp
				&& ((IRMem) expr).memType() == MemType.IMMUTABLE
				&& ((IRBinOp) ((IRMem) expr).expr()).opType() == IRBinOp.OpType.SUB
				&& ((IRBinOp) ((IRMem) expr).expr()).left() instanceof IRTemp
				&& ((IRBinOp) ((IRMem) expr).expr()).right() instanceof IRConst
				&& ((IRConst) ((IRBinOp) ((IRMem) expr).expr()).right())
						.value() == Configuration.WORD_SIZE) {
			
			/* Resolve to the original array name */
			String name = ((IRTemp) ((IRBinOp) ((IRMem) expr).expr()).left()).name();

			if (lengthExprs != null && lengthExprs.containsKey(name))
				return new Interval(lengthExprs.get(name));
			
			/* This evaluates the length of an array */
			return new Interval(new NumberType(name));
		} else if (expr instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) expr;
			Interval left = evalUnclamped(binOp.left(), intervals, lengthExprs);
			Interval right = evalUnclamped(binOp.right(), intervals, lengthExprs);
			if (left == null || right == null || left.isTop() || right.isTop())
				return new Interval();
			NumberType l = null;
			NumberType r = null;
			OpType opType = binOp.opType();

			// System.out.print("eval(" + expr + ", " + left + ", " + right +
			// ")=");

			if (opType == OpType.UGE || opType == OpType.ULE
					|| opType == OpType.UGT || opType == OpType.ULT) {
				left = left.unsignedInterval();
				right = right.unsignedInterval();
			}

			switch (opType) {
			case ADD:
				l = NumberType.add(left.l(), right.l());
				r = NumberType.add(left.r(), right.r());
				break;
			case SUB:
				l = NumberType.sub(left.l(), right.r());
				r = NumberType.sub(left.r(), right.l());
				break;
			case MUL:
				l = NumberType.min(NumberType.min(NumberType.mul(left.l(),
						right.l()), NumberType.mul(left.l(), right.r())),
						NumberType.min(NumberType.mul(left.r(), right.l()),
								NumberType.mul(left.r(), right.r())));
				r = NumberType.max(NumberType.max(NumberType.mul(left.l(),
						right.l()), NumberType.mul(left.l(), right.r())),
						NumberType.max(NumberType.mul(left.r(), right.l()),
								NumberType.mul(left.r(), right.r())));
				break;
			case AR_LSHIFT:
				if (binOp.right() instanceof IRConst) {
					NumberType value = new NumberType(1 << ((IRConst) binOp
							.right()).value());
					l = NumberType.mul(left.l(), value);
					r = NumberType.mul(left.r(), value);
				}
				break;
			case EQ:
				if (left.isSingleElement() && right.isSingleElement()
						&& left.l().equals(right.l()))
					l = r = NumberType.ONE;
				else if (left.r().lt(right.l()) == TriBool.YES
						|| right.r().lt(left.l()) == TriBool.YES)
					l = r = NumberType.ZERO;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			case NEQ:
				if (left.isSingleElement() && right.isSingleElement()
						&& left.l().equals(right.l()))
					l = r = NumberType.ZERO;
				else if (left.r().lt(right.l()) == TriBool.YES
						|| right.r().lt(left.l()) == TriBool.YES)
					l = r = NumberType.ONE;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			case LT:
			case ULT:
				if (left.r().lt(right.l()) == TriBool.YES)
					l = r = NumberType.ONE;
				else if (left.l().ge(right.r()) == TriBool.YES)
					l = r = NumberType.ZERO;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			case GT:
			case UGT:
				if (right.r().lt(left.l()) == TriBool.YES)
					l = r = NumberType.ONE;
				else if (right.l().ge(left.r()) == TriBool.YES)
					l = r = NumberType.ZERO;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			case LE:
			case ULE:
				if (left.r().le(right.l()) == TriBool.YES)
					l = r = NumberType.ONE;
				else if (left.l().gt(right.r()) == TriBool.YES)
					l = r = NumberType.ZERO;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			case GE:
			case UGE:
				if (right.r().le(left.l()) == TriBool.YES)
					l = r = NumberType.ONE;
				else if (right.l().gt(left.r()) == TriBool.YES)
					l = r = NumberType.ZERO;
				else {
					l = NumberType.ZERO;
					r = NumberType.ONE;
				}
				break;
			}
			if (l == null)
				l = NumberType.N_INF;
			if (r == null)
				r = NumberType.P_INF;
			// System.out.println(new Interval(l, r));
			return new Interval(l, r);
		}

		/* Don't know ... */
		return new Interval(NumberType.N_INF, NumberType.P_INF);
	}
}
