package xic.analysis;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.AbstractInstr;
import xic.codegen.AsmDirective;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.LabelInstr;
import xic.ir.IRCJump;
import xic.ir.IRCompUnit;
import xic.ir.IRDataSeq;
import xic.ir.IRFuncDecl;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRName;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.visitors.IRGenerator;


/**
 * Control flow graph generator capable of producing both IR-level and
 * assembly-level CFGs.
 */
public class CFGGenerator {
	public static Map<String, CFGNode> createCFG(IRCompUnit compUnit) {
		Map<String, CFGNode> labelMap = new HashMap<String, CFGNode>();
		Map<String, CFGNode> internalMap = new HashMap<String, CFGNode>();
		Map<String, IRFuncDecl> functions = compUnit.functions();

		for (String funcName : functions.keySet()) {
			IRFuncDecl func = functions.get(funcName);
			labelMap.put(funcName, createCFG(funcName, internalMap,
					(IRSeq) func.body()));
		}

		if (internalMap.containsKey(IRGenerator.CALL_ABORT))
			internalMap.get(IRGenerator.CALL_ABORT).setValue(IRGenerator.CALL_ABORT);

		return labelMap;
	}

	public static CFGNode createCFG(String name, InstrList il) {
		Map<String, CFGNode> labelMap = new HashMap<String, CFGNode>();
		CFGNode result, cur, next = null;
		result = cur = new CFGNode("entry [" + name + "]");

		Iterator<AbstractInstr> it = il.iterator();

		while (true) {
			next = new CFGNode();

			if (it.hasNext())
				next.setValue(it.next());
			else
				next.setValue("exit");

			if (next.value() instanceof AsmDirective)
				continue;

			if (next.value() instanceof LabelInstr) {
				LabelInstr instr = ((LabelInstr) next.value());
				if (labelMap.containsKey(instr.label())) {
					next = labelMap.get(instr.label());
					if (next.value() == null)
						next.setValue(instr);
				} else {
					labelMap.put(instr.label(), next);
				}
			}

			if (cur.value() instanceof Instr) {
				Instr instr = (Instr) cur.value();
				if (instr.isConditionalJump() || instr.isUnconditionalJump()) {
					String targetName = instr.getJumpTarget();
					CFGNode target;

					if (!labelMap.containsKey(targetName)) {
						target = new CFGNode();
						labelMap.put(targetName, target);
					} else {
						target = labelMap.get(targetName);
					}

					cur.addEdgeTo(target);

					if (!instr.isUnconditionalJump()) {
						if (!next.inEdges().contains(cur))
							cur.addEdgeTo(next);
					}
				} else {
					cur.addEdgeTo(next);
				}
			} else {
				cur.addEdgeTo(next);
			}
			cur = next;

			if (next.value().equals("exit"))
				break;
		}

		if (labelMap.containsKey(IRGenerator.CALL_ABORT))
			labelMap.get(IRGenerator.CALL_ABORT).setValue(IRGenerator.CALL_ABORT);

		removeOrphans(result);

		return result;
	}

	private static CFGNode createCFG(String funcName,
			Map<String, CFGNode> labelMap, IRSeq seq) {
		CFGNode result, cur, next = null;

		result = cur = new CFGNode("entry [" + funcName + "]");
		Iterator<IRStmt> it = seq.stmts().iterator();
		labelMap.clear();

		while (true) {
			next = new CFGNode();

			if (it.hasNext())
				next.setValue(it.next());
			else
				next.setValue("exit");

			if (next.value() instanceof IRDataSeq)
				continue;

			if (next.value() instanceof IRLabel) {
				IRLabel label = ((IRLabel) next.value());
				if (labelMap.containsKey(label.name())) {
					next = labelMap.get(label.name());
					if (next.value() == null)
						next.setValue(label);
				} else {
					/* haven't seen this label before */
					labelMap.put(label.name(), next);
				}
			}

			if (cur.value() instanceof IRCJump) {
				IRCJump cjump = (IRCJump) cur.value();
				String[] targets = new String[] { cjump.trueLabel(),
						cjump.falseLabel() };
				boolean fallthrough = true;

				for (String targetName : targets) {
					CFGNode target = null;

					if (targetName == null) {
						fallthrough = true;
						continue;
					}

					if (!labelMap.containsKey(targetName)) {
						target = new CFGNode();
						labelMap.put(targetName, target);
					} else {
						target = labelMap.get(targetName);
					}

					cur.addEdgeTo(target);
				}

				if (fallthrough) {
					if (!cur.outEdges().contains(next))
						cur.addEdgeTo(next);
				}
			} else if (cur.value() instanceof IRJump) {
				IRJump jump = (IRJump) cur.value();
				CFGNode target;
				if (jump.target() instanceof IRName) {
					IRName name = (IRName) jump.target();
					if (!labelMap.containsKey(name.name())) {
						target = new CFGNode();
						labelMap.put(name.name(), target);
					} else {
						target = labelMap.get(name.name());
					}
				} else {
					throw new InternalCompilerException(
							"Computed jump expressions not supported.");
				}

				/* Jumps only jump to target */
				cur.addEdgeTo(target);
			} else {
				/* Everything else flows into next stmt */
				cur.addEdgeTo(next);
			}
			cur = next;

			if (next.value().equals("exit"))
				break;
		}

		if (labelMap.containsKey(IRGenerator.CALL_ABORT))
			labelMap.get(IRGenerator.CALL_ABORT).setValue(IRGenerator.CALL_ABORT);

		removeOrphans(result);
		return result;
	}

	public static void removeOrphans(CFGNode head) {
		Set<CFGNode> set = head.flatten();
		for (CFGNode node : set) {
			for (CFGNode n : new HashSet<CFGNode>(node.inEdges())) {
				if (!set.contains(n)) 
					node.removeInEdge(n);
			}
		}
	}
}
