package xic.analysis;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ir.IRCJump;
import xic.ir.IRCompUnit;
import xic.ir.IRFuncDecl;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRName;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.visitors.IRGenerator;


/**
 * This class generates a trace of one possible execution sequence of the code,
 * and generates a new IRCompUnit on the way. Unreachable code elimination is 
 * a natural consequence of never adding a node that is inaccessible from 
 * the method entry.
 */
public class CodeOrdering {
	public static IRCompUnit orderCode(String name, Map<String, CFGNode> cfgs,
			IRSeq extraSeq, Map<String,Boolean> procs) {
		IRCompUnit compUnit = new IRCompUnit(name);
		compUnit.setExtraSequence(extraSeq);
		
		int index = 0;
		for (String funcName : cfgs.keySet()) {
			CFGNode head = cleanupPass(cfgs.get(funcName));
			IRSeq seq = ordering(head, index++);
			compUnit.appendFunc(new IRFuncDecl(funcName, seq, procs.get(funcName)));
		}

		return compUnit;
	}

	private static CFGNode cleanupPass(CFGNode head) {
		/* Filter pass: remove useless label<->jump pairs 
		   created from splitting critical edges 
		 */ 
		for (CFGNode n : head.flatten()) {
			if (!(n.value() instanceof IRLabel))
				continue;
			if (n.inEdges().size() != 1 || n.outEdges().size() != 1)
				continue;
			if (!(n.outEdges().get(0).value() instanceof IRJump))
				continue;
			CFGNode pred = n.inEdges().get(0);
			CFGNode uselessJump = n.outEdges().get(0);
			CFGNode succ = uselessJump.outEdges().get(0);

			String uselessLabelName = ((IRLabel) n.value()).name();
			String targetLabelName = (succ.value() instanceof String) ?
					(String) succ.value() : ((IRLabel) succ.value()).name();

			pred.removeEdgeTo(n);
			uselessJump.removeEdgeTo(succ);
			if (!pred.outEdges().contains(succ))
				pred.addEdgeTo(succ);

			if (pred.value() instanceof IRCJump) {
				IRCJump cJump = (IRCJump) pred.value();
				if (uselessLabelName.equals(cJump.trueLabel()))
					pred.setValue(new IRCJump(cJump.expr(), targetLabelName,
							cJump.falseLabel()));
				else if (uselessLabelName.equals(cJump.falseLabel()))
					pred.setValue(new IRCJump(cJump.expr(), cJump.trueLabel(),
							targetLabelName));
				else
					throw new InternalCompilerException("Internal error during cleanup pass (1)");
			} else if (pred.value() instanceof IRJump) {
				String jumpLabel = ((IRName) ((IRJump) pred.value()).target()).name();
				if (uselessLabelName.equals(jumpLabel))
					pred.setValue(new IRJump(new IRName(targetLabelName)));
				else
					throw new InternalCompilerException("Internal error during cleanup pass (2)");
			}
		}

		return head;
	}

	private static IRSeq ordering(CFGNode head, int index) {
		HashSet<String> usedLabels = new HashSet<String>();
		ArrayList<CFGNode> trace = new ArrayList<CFGNode>();
		IRSeq seq = new IRSeq();
		List<Object> valueList = new ArrayList<Object>();

		// we do a depth-first traverse of the CFG, so the unreachable nodes
		// will not appear in this set
		Set<CFGNode> unmarkedNodes = head.flatten();

		// start the ordering from head
		CFGNode next = head;
		while (!unmarkedNodes.isEmpty()) {
			unmarkedNodes.remove(next);

			if (next.value().equals(IRGenerator.CALL_ABORT)) {
				next = selectUnmarked(unmarkedNodes);
				continue;
			}
			
			trace.add(next);
			if (next.outEdges().size() > 1) {
				// this is a CJUMP, need to choose one way to jump and one way to fall through
				if (!(next.value() instanceof IRCJump))
					throw new InternalCompilerException("Expected a CJump");
				next = selectCJump(next.outEdges(), (IRCJump) next.value());

				if (unmarkedNodes.contains(next))
					continue;
				else
					next = selectUnmarked(unmarkedNodes);
			} else if (next.outEdges().size() == 1) {
				boolean isJump = next.value() instanceof IRJump;
				CFGNode node = next;
				// stmt inside a block, goto next stmt
				next = next.outEdges().get(0);
				if (unmarkedNodes.contains(next)) {
					continue;
				} else {
					// the next stmt is moved, add a jump to that block
					if (!(next.value() instanceof IRLabel) && !(next.value() instanceof String))
						throw new InternalCompilerException(
								"Expected a Label following '" + node + "', got: " + next);
					String name = next.value() instanceof String ? (String) next.value()
							: ((IRLabel) next.value()).name();
					if (!isJump)
						trace.add(new CFGNode(new IRJump(new IRName(name))));
					next = selectUnmarked(unmarkedNodes);
				}
			} else {
				// need an end label for reordered stmts that lies at the end of the sequence
				if (!unmarkedNodes.isEmpty()) {
					CFGNode end = new CFGNode();
					end.setValue(new IRJump(new IRName(".E"+index)));
					trace.add(end);
				}
				// also pick a node from the unmarked set to start
				next = selectUnmarked(unmarkedNodes);
			}
		}
		trace.add(new CFGNode(new IRLabel(".E"+index)));

		// now we can eliminate unnecessary jumps in our trace
		for (int i = 0; i < trace.size(); i++) {
			CFGNode node = trace.get(i);
			if (node.value() instanceof IRJump) {
				IRJump jump = (IRJump) node.value();
				// if the target is next stmt, we can eliminate this jump
				if (i+1 < trace.size()) {
					CFGNode lookahead = trace.get(i + 1);
					if (lookahead.value() instanceof String
							|| lookahead.value() instanceof IRLabel) {
						String labelName = lookahead.value() instanceof String ?
								(String) lookahead.value()
								: ((IRLabel) lookahead.value()).name();
						// else the target is next stmt, skip it
						if (labelName.equals(((IRName) jump.target()).name()))
							continue;
					}
				}
				// otherwise, add the original IRNode
				valueList.add((IRStmt) node.value());
				usedLabels.add(((IRName) jump.target()).name());
			} else if (node.value() instanceof IRCJump) {
				// if the false branch is next stmt, transform it into the form
				// from
				// CJUMP(E, T, F) T: F:
				// to
				// CJUMP(E, T, null) ... T: 
				IRCJump jump = (IRCJump) node.value();
				if (i<trace.size() - 1) {
					CFGNode lookahead = trace.get(i + 1);
					if (lookahead.value() instanceof IRLabel) {
						String labelName = ((IRLabel) lookahead.value()).name();
						if (labelName.equals(jump.falseLabel())) {
							valueList.add(new IRCJump(jump.expr(), jump.trueLabel(), null));
							usedLabels.add(jump.trueLabel());
							continue;
						} else if (labelName.equals(jump.trueLabel())) {
							valueList.add(new IRCJump(jump.expr().negate(), jump.falseLabel(), null));
							usedLabels.add(jump.falseLabel());
							continue;
						}
					}
				}

				// otherwise, translate
				// CJUMP(E, T, F) T: F:
				// to
				// CJUMP(E, T, null) JUMP(F) T: F: 
				if (jump.falseLabel() == null) {
					valueList.add(jump);
					usedLabels.add(jump.trueLabel());
				} else if (jump.trueLabel() == null) {
					IRCJump cJump = new IRCJump(jump.expr().negate(), jump.falseLabel(), null);
					valueList.add(cJump);
					usedLabels.add(jump.falseLabel());
				} else {
					valueList.add(new IRCJump(jump.expr(), jump.trueLabel(), null));
					valueList.add(new IRJump(new IRName(jump.falseLabel())));
					usedLabels.add(jump.trueLabel());
					usedLabels.add(jump.falseLabel());
				}
				
			} else {
				if (node.value() instanceof IRStmt)
					valueList.add((IRStmt) node.value());
			}
		}

		// finally, eliminate the useless labels and cjumps
		for (Object value : valueList) {
			if ((!(value instanceof IRLabel)
					|| usedLabels.contains(((IRLabel) value).name()))
				&& (!(value instanceof IRCJump)
						|| ((IRCJump) value).trueLabel() != null
						|| ((IRCJump) value).falseLabel() != null)) {
				seq.appendStmt((IRStmt) value);
			}
		}

		return seq;
	}

	private static CFGNode selectUnmarked(Collection<CFGNode> nodes) {
		// pick up the nodes with smallest in-edges
		CFGNode result = null;
		int smallest = Integer.MAX_VALUE;

		for (CFGNode node : nodes) {
			if (((node.value() instanceof IRLabel) || node.value().equals(IRGenerator.CALL_ABORT))
					&& node.inEdges().size() < smallest) {
				result = node;
				smallest = node.inEdges().size();
			}
		}

		return result;
	}

	private static CFGNode selectCJump(Collection<CFGNode> nodes, IRCJump jump) {
		// Speculate that the positive branch is generally taken (this might be in a loop)
		for (CFGNode node : nodes) {
			if (node.value() instanceof IRLabel) {
				String label = ((IRLabel) node.value()).name();
				if (label.equals(jump.trueLabel()))
					return node;
			} else if (node.value() instanceof String) {
				String label = (String) node.value();
				if (label.equals(jump.trueLabel()))
					return node;
				else if (jump.trueLabel() == null && !label.equals(jump.falseLabel()))
					return node;
			} else {
				if (jump.trueLabel() == null)
					return node;
			}
		}
		throw new InternalCompilerException("Should never get here!");
	}
}
