package xic.analysis;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ir.IRBinOp;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMove;
import xic.ir.IRName;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;


public class PropagatedConstants extends DataFlowAnalysis {
	private static final long TOP = Long.MAX_VALUE;
	private static final long BOTTOM = Long.MIN_VALUE;
	private static final long IS_NAME = Long.MIN_VALUE+1;

	public static class Constant {
		private long value;
		private String name;

		// create the top value by default
		public Constant() {
			value = TOP;
		}
		
		public Constant(long val) {
			value = val;
		}

		public Constant(String name) {
			this.name = name;
			value = IS_NAME;
		}

		public Constant(Constant val) {
			name = val.name;
			value = val.value;
		}

		public boolean isTop () {
			return value == TOP;
		}

		public boolean isBottom() {
			return value == BOTTOM;
		}

		public boolean isName() {
			return value == IS_NAME;
		}

		public boolean equals (Object val) {
			if (!(val instanceof Constant))
				return false;
			Constant c2 = (Constant) val;
			if ((isName() != c2.isName()) || (isName() && !name.equals(c2.name)))
				return false;
			return value == c2.value;
		}
		
		public int hashCode() {
			int result = 17;
			result = 37 * result + (int) value;
			if (isName())
				result = 37 * result + name.hashCode();
			return result;
		}

		public long value() {
			return value;
		}

		public String name() {
			return name;
		}

		// set the greatest lower bound of two values
		public void setGLB(Constant v1, Constant v2) {
			if (v1.equals(v2)) {
				value = v1.value;
				name = v1.name;
			} else if (v1.isBottom() || v2.isBottom()) {
				value = BOTTOM;
				name = null;
			} else if (v1.isTop() || v2.isTop()) {
				value = v1.value() > v2.value() ? 
						v2.value() : v1.value();

				if (isName()) {
					if (v1.name != null)
						name = v1.name;
					else if (v2.name != null)
						name = v2.name;
					else
						throw new InternalCompilerException("Could not find name!");
				} else {
					name = null;
				}
			} else {
				value = BOTTOM;
				name = null;
			}
		}

		public String toString() {
			if (isTop())
				return "top";
			if (isBottom())
				return "bottom";
			if (isName())
				return "name[" + name + "]";
			return String.valueOf(value);
		}
	}

	/// Value class used during copy propagation analysis
	public static class ConstValue implements Value {
		/* List of variable pairs */
		private Map<String, Constant> values;

		public ConstValue() {
			values = new HashMap<String, Constant>();
		}

		public Map<String, Constant> values() {
			return values;
		}

		public Value copy() {
			ConstValue r = new ConstValue();
			for (String key:values.keySet())
				r.values.put(key, new Constant(values.get(key)));
			return r;
		}

		public void meet(Value v) {
			Map<String, Constant> otherValues = ((ConstValue)v).values;

			for (String key : otherValues.keySet()) {
				Constant v1 = otherValues.get(key);
				Constant v2 = values.get(key);

				if (v2 == null)
					values.put(key, v1);
				else
					v2.setGLB(v1, v2);
			}
		}

		public String toString() {
			StringBuffer buf = new StringBuffer("[");
			Iterator<String> it = values.keySet().iterator();
			while (it.hasNext()) {
				String key = it.next();
				buf.append(key+"="+ values.get(key));
				if (it.hasNext())
					buf.append(", ");
			}
			return buf.append(']').toString(); 
		}

		public boolean equals(Object obj) {
			if (!(obj instanceof ConstValue))
				return false;
			return values.equals(((ConstValue) obj).values);
		}

		public int hashCode() {
			return values.hashCode();
		}

		public void setNullMeet() {
			return;
		}
	};
	
	/// Dataflow-related information associated with each node
	public static class ConstFlowInfo extends FlowInfo {
		private boolean isConst = false, isName = false;
		private long value = 0;
		private String tempName = null, nameExpr = null;
		private IRExpr expr = null;
		private boolean kill = false;
		private Set<String> usedTemps = new HashSet<String>();

		public ConstFlowInfo(Object o) {
			CFGNode node = (CFGNode) o;
			this.entry = new ConstValue();
			this.exit = new ConstValue();

			if(node.value() instanceof IRStmt) {
				IRStmt irStmt = (IRStmt)node.value();
				if (irStmt instanceof IRMove) {
					IRMove move = (IRMove) irStmt;

					if (move.target() instanceof IRTemp) {
						tempName = ((IRTemp)move.target()).name();
						// x = const
						if (move.expr() instanceof IRConst) {
							isConst = true;
							value = ((IRConst) move.expr()).value();
						} else if (move.expr() instanceof IRName) {
							isName = true;
							nameExpr = ((IRName) move.expr()).name();
						} else if (move.expr().containsCalls() || move.expr().containsMemReads()) {
							kill = true;
						} else {
							expr = move.expr();
							move.expr().findReadTemps(usedTemps);
						}
					}
				} else if (irStmt instanceof IRDecl) {
					tempName = ((IRDecl) irStmt).name();
					kill = true;
				} else if (irStmt instanceof IRDataSeq
							|| irStmt instanceof IRJump
							|| irStmt instanceof IRLabel	
							|| irStmt instanceof IRSeq
							|| irStmt instanceof IRExp
							|| irStmt instanceof IRCJump) {
					// don't do anything
				} else
					throw new InternalCompilerException("Unknown IRStmt: "+irStmt);
			}
		}
	};

	public PropagatedConstants(Set<CFGNode> nodes) {
		super(nodes);
		
		dir = Direction.FORWARD;
		top = new ConstValue();
	}

	public ConstValue in(CFGNode node) {
		ConstFlowInfo info = (ConstFlowInfo) flowInfo.get(node);
		return (ConstValue) info.entry;
	}
	
	public ConstValue out(CFGNode node) {
		ConstFlowInfo info = (ConstFlowInfo) flowInfo.get(node);
		return (ConstValue) info.exit;
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new ConstFlowInfo(node);
	}

	protected void F(FlowInfo f) {
		ConstFlowInfo fi = (ConstFlowInfo) f;
		Map<String, Constant> entry = ((ConstValue) fi.entry).values;
		Map<String, Constant> exit = ((ConstValue) fi.exit).values;
		exit.clear();
		exit.putAll(entry);
					
		if (fi.tempName != null) {
			boolean containsTop = false;
			boolean containsBottom = false;

			if (fi.isConst) {
				exit.put(fi.tempName, new Constant(fi.value));
			}  else if (fi.isName) {
				exit.put(fi.tempName, new Constant(fi.nameExpr));
			} else if (fi.kill) {
				exit.put(fi.tempName, new Constant(BOTTOM));
			}  else {
				/// TOP: don't know yet,  BOTTOM: conservative non-constant 
				for (String temp : fi.usedTemps) {
					Constant v = entry.get(temp);
					if (v == null || v.isTop()) {
						containsTop = true;
						continue;
					}
					if (v.isBottom())
						containsBottom = true;
				}

				if (containsTop && fi.expr != null) {
					exit.put(fi.tempName, new Constant(TOP));
				} else if (containsBottom && fi.expr != null) {
					exit.put(fi.tempName, new Constant(BOTTOM));
				} else if (fi.expr != null) {
					long value = eval(fi.expr, entry);
					if (value != BOTTOM)
						exit.put(fi.tempName, new Constant(value));
				}
			}
		}	
	}

	// take an expression and a map from variables to value, eval the expression if possible
	private long eval(IRExpr expr, Map<String, Constant> map) {
		if (expr instanceof IRConst) {
			return ((IRConst) expr).value();
		} else if (expr instanceof IRTemp) {
			return map.get(((IRTemp) expr).name()).value();
		} else if (expr instanceof IRBinOp) {
			IRBinOp bop = (IRBinOp) expr;
			long v1 = eval(bop.left(), map);
			long v2 = eval(bop.right(), map);

			if (v1 != BOTTOM && v2 != BOTTOM)
				return IRBinOp.eval(bop.opType(), v1, v2);

			else return BOTTOM;
		}
		return BOTTOM;
	}
}
