package xic.analysis;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.Configuration;
import xic.Configuration.Architecture;
import xic.InternalCompilerException;
import xic.Statistics;
import xic.analysis.LiveVariables.LVFlowInfo;
import xic.codegen.Instr;
import xic.codegen.Instr.Register;


/**
 * This class analyzes a single function in pseudo-assembly form and produces
 * a list of all detected webs (following section 16.3 in Muchnick). A web
 * inference graph is also generated.
 */
public class WebGenerator {
	/* Should the register allocator be biased to prefer certain registers in some cases? */
	private static final boolean ENABLE_BIAS = true; 

	/* Possible register coalescing criteria (see below) */
	public enum CoalesceCriteria {
		COALESCE_NONE,
		COALESCE_BRIGGS,
		COALESCE_BRIGGS_AND_GEORGE
	};

	public final static class Web implements Comparable<Web> {
		private boolean spilled;
		private Set<String> varNames;
		private Set<CFGNode> defs;
		private Set<CFGNode> uses;
		private Set<Web> neighbors;
		private Map<CFGNode, Integer> loopDepth;
		private Set<String> registerPreferences;
		private int color;
		private long id = 0;
		private static long id_counter = 0;
		private static Object lock = new Object();

		/// Cache some information for move-related temps
		private boolean isMoveRelated = true;
		private boolean isStrictlyMoveRelated = true;
		private Set<String> moveSources = null;

		public Web(String varName, CFGNode def) {
			spilled = false;
			varNames = new HashSet<String>();
			defs = new HashSet<CFGNode>();
			uses = new HashSet<CFGNode>();
			neighbors = new HashSet<Web>();
			moveSources = new HashSet<String>();
			registerPreferences = new HashSet<String>();
			if (def != null)
				defs.add(def);
			varNames.add(varName);
			color = GraphColoringUtils.UNINITIALIZED_COLOR;
			update();

			synchronized (lock) {
				id = id_counter++;
			}
		}
		
		public int color() {
			return color;
		}

		public void setColor(int color) {
			this.color = color;
		}

		public Set<Web> neighbors() {
			return neighbors;
		}

		public boolean isSpilled() {
			return spilled;
		}

		public void setSpilled(boolean spilled) {
			this.spilled = spilled;
		}

		public Set<CFGNode> uses() {
			return uses;
		}

		public Set<CFGNode> defs() {
			return defs;
		}

		public Set<String> varNames() {
			return varNames;
		}
		
		public Set<String> registerPreferences() {
			return registerPreferences;
		}
		
		public void addRegisterPreference(String registerPreference) {
			registerPreferences.add(registerPreference);
		}


		public boolean hasRegisterPreferences() {
			return registerPreferences.size() > 0;
		}

		public void merge(Web w) {
			if (isNeighborOf(w) || w.isNeighborOf(this))
				throw new InternalCompilerException("Invalid merge operation!");
			varNames.addAll(w.varNames);
			registerPreferences.addAll(w.registerPreferences);
			uses.addAll(w.uses);
			defs.addAll(w.defs);
			for (Web wNeighbor : w.neighbors) {
				if (!wNeighbor.isNeighborOf(w))
					throw new InternalCompilerException("Error in graph structure");
				wNeighbor.neighbors.remove(w);
				wNeighbor.neighbors.add(this);
			}
			neighbors.addAll(w.neighbors);
			update();
		}

		public void addUse(CFGNode node) {
			uses.add(node);
		}

		public void addNeighbor(Web web) {
			if (web != this) {
				neighbors.add(web);
				web.neighbors.add(this);
			}
		}

		public int degree() {
			int result = neighbors.size();
			for (Web n : neighbors)
				if (n.color() == GraphColoringUtils.IGNORE_COLOR)
					--result;
			return result;
		}

		public int compareTo(Web o) {
			if (degree() > o.degree())
				return 1;
			else if (degree() < o.degree())
				return -1;
			else
				return 0;
		}

		public String toString() {
			if (defs.size() == 0 && varNames.size() == 1)
				return varNames.iterator().next();
			String result = "\nWeb[\n" +
			"  varName = " + stringSetToString(varNames) + ",\n" +
			"  moveRelated = " + isMoveRelated + ",\n" +
			"  moveSources = " + moveSources + ",\n" +
			"  defs = {" + nodeSetToString(defs) + "},\n" +
			"  uses = {" + nodeSetToString(uses) + "},\n";
			if (color >= 0 && color < Configuration.AVAILABLE_REGISTERS.length)
				result += "  register = " + Configuration.AVAILABLE_REGISTERS[color] + ",\n";
			else if (color == Configuration.AVAILABLE_REGISTERS.length)
				result += "  register = " + Configuration.IP_REGISTER + ",\n";
			else if (color != GraphColoringUtils.UNINITIALIZED_COLOR)
				throw new InternalCompilerException("Unknown color " + color);
			if (loopDepth != null)
				result += "  spillCost = " + spillCost() + ",\n";
			return result + "  registerPreferences = {"
					+ stringSetToString(registerPreferences) + "},\n"
					+ "  spilled = " + spilled + "\n" + "]";
		}

		private static String stringSetToString(Set<String> names) {
			Iterator<String> it = names.iterator();
			StringBuffer buf = new StringBuffer();
			if (names.size() > 1)
				buf.append("{");

			while (it.hasNext()) {
				buf.append('\"');
				buf.append(it.next());
				buf.append('\"');
				if (it.hasNext())
					buf.append(", ");
			}

			if (names.size() > 1)
				buf.append("}");

			return buf.toString();	
		}

		private String nodeSetToString(Set<CFGNode> nodes) {
			StringBuffer sb = new StringBuffer();
			int width = 0;
			Iterator<CFGNode> it = nodes.iterator();
			while (it.hasNext()) {
				CFGNode n = it.next();
				String nodeStr = n.toString();
				if (loopDepth != null)
					nodeStr += ":" + loopDepth.get(n);
				width += nodeStr.length();
				sb.append(nodeStr);
				if (it.hasNext()) {
					sb.append(", ");
					if (width > 20)
						sb.append("\n    ");
				}
			}
			return sb.toString();
		}

		public boolean isColored() {
			return color != GraphColoringUtils.UNINITIALIZED_COLOR;
		}

		public String varNameString() {
			if (varNames.size() == 1)
				return varNames.iterator().next();
			return stringSetToString(varNames);
		}

		private void update() {
			/* Precompute information for isMoveRelated() and getMoveTargets() */
			moveSources.clear();
			isMoveRelated = false;
			isStrictlyMoveRelated = true;
			
			for (CFGNode def: defs) {
				if (!(def.value() instanceof Instr))
					continue;
				Instr i = (Instr) def.value();
				
				if (!i.isMove()) {
					isStrictlyMoveRelated = false;
					continue;
				}

				if (i.getOp1().getType() == i.getOp2().getType() &&
						i.getOp1().getType() == Instr.Op.Type.REG&&
						!varNames.contains(i.getOp2().getBase())) {
					isMoveRelated = true;
					moveSources.add(i.getOp2().getBase());
				} else {
					isStrictlyMoveRelated = false;
				}
			}
		}

		public Set<String> getMoveSources() {
			if (!isMoveRelated())
				throw new InternalCompilerException("Not supported!");
			return moveSources;
		}

		/// Does this web include definitions of the form 'mov t1, t2'? 
		public boolean isMoveRelated() {
			return isMoveRelated;
		}
	
		/// Is this web created only from definitions of the form 'mov t1, t2'? 
		public boolean isStrictlyMoveRelated() {
			return isStrictlyMoveRelated;
		}

		public int hashCode() {
			// Want reference equality without non-determinism
			// introduced by Object.hashCode()
			return (int) id;
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof Web))
				return false;
			return ((Web) obj).id == id;
		}

		public boolean isNeighborOf(Web w) {
			return neighbors.contains(w);
		}

		public int spillCost() {
			if (varNameString().contains("spilled") || loopDepth == null)
				return Integer.MAX_VALUE;

			double cost = 0;
			for (CFGNode def : defs)
				cost += Math.pow(10, loopDepth.get(def));
			for (CFGNode use : uses)
				cost += Math.pow(10, loopDepth.get(use));
			if (neighbors.size() > 0)
				cost /= neighbors().size();

			return (int) cost;
		}

		public void setLoopDepth(CFGNode n, int depth) {
			if (loopDepth == null)
				loopDepth = new HashMap<CFGNode, Integer>();
			loopDepth.put(n, depth);
		}

	}

	public static Set<Web> run(String funcName, Statistics stats, int pass,
			Set<CFGNode> nodes, CoalesceCriteria coalesceCriteria) {
		stats.startTimer("regAlloc/webs");

		String regA = new Register(Register.Type.A).toString();
		String regD = new Register(Register.Type.D).toString();
		String regSI = new Register(Register.Type.SI).toString();
		String regDI = new Register(Register.Type.DI).toString();
		String regBP = new Register(Register.Type.BP).toString();
		
		/* 1. Perform a live variable analysis */
		LiveVariables lv = new LiveVariables(nodes);
		lv.solve();

		/* 2. Perform a reaching definitions analysis */
		ReachingDefinitions rd = new ReachingDefinitions(nodes);
		rd.solve();

		/* 3. Initially build a web for each definition */
		List<Web> webs = new ArrayList<Web>();
		Map<Web, List<Integer>> reverseMapping = new HashMap<Web, List<Integer>>();
		Map<String, Web> registerWebs = new HashMap<String, Web>();

		for (CFGNode def : rd.getDefinitions()) {
			Web web = new Web(rd.getVarName(def), def);
			List<Integer> list = new ArrayList<Integer>();
			list.add(webs.size());
			webs.add(web);
			reverseMapping.put(web, list);

			Set<String> names = web.varNames;

			if (names.size() == 1) {
				String name = names.iterator().next();

				if (!Configuration.REG_NAMES.contains(name))
					continue;
				
				if (registerWebs.containsKey(name)) {
					web.setColor(registerWebs.get(name).color());
					union(webs, reverseMapping, registerWebs.get(name), web);
					continue;
				}

				registerWebs.put(name, web);
				boolean found = false;
				for (int i=0; i<Configuration.AVAILABLE_REGISTERS.length; ++i) {
					String regName = Configuration.AVAILABLE_REGISTERS[i].toString();
					if (regName.equals(name)) {
						found = true;
						web.setColor(i);
					}
				}
				if (Configuration.IP_REGISTER.equals(name)) {
					web.setColor(Configuration.AVAILABLE_REGISTERS.length);
					found = true;
				}
				if (!found)
					throw new InternalCompilerException(
							"Unable to resolve a declaration to a register index!");
			}
		}

		/* 4. Merge chains */
		int defCount = rd.getDefinitions().size();
		for (CFGNode node : nodes) {
			Set<String> liveVariables = lv.getLiveVariablesBefore(node);
			BitSet reachingDefinitions = rd.getReachingDefinitionsBefore(node);
			
			if (node.value() instanceof Instr) {
				Instr i = (Instr) node.value();
				int defIdx = rd.getDefinitionIndex(node);

				if (i.isDeclaration())
					continue;

				/* If this is a non-move instruction, which writes to a TEMP */
				if (defIdx >= 0 && !(i.isLea() || i.isMove() || i.isXORSelf() || i.isCall())) {
					/* Merge the chains associated with the target temporary
					   This is necessary since an instruction such as
					        add t1, 5
					   can't use two separate registers for the use and definition
					*/
					BitSet bitSet = new BitSet(defCount);
					Web web = webs.get(defIdx);

					for (int j: rd.getDefinitionsForVariable(i.getOp1().getBase()))
						bitSet.set(j);
					
					bitSet.and(reachingDefinitions);

					for (int j=0; j<defCount; ++j) {
						if (!bitSet.get(j))
							continue;
						union(webs, reverseMapping, web, webs.get(j));
					}
				} else if (defIdx >= 0 && i.isXORSelf()) {
					/* Instructions of the form 'xor t1, t1' are special because
					 * they "use" a temporary, which may not have been defined yet */
					webs.get(defIdx).addUse(node);
				}

				/* Bias the register allocator to put a temp into a certain register, 
				   e.g. 'RV' into eax after seeing the instruction 'mov eax, RV' */
				if (ENABLE_BIAS) {
					if (i.isMoveTempToReg()) {
						/* mov reg, temp */
						BitSet bitSet = new BitSet(defCount);
						List<Integer> rdefs = rd.getDefinitionsForVariable(i.getOp2().getBase());
						if (rdefs == null)
							continue;

						for (int j: rdefs)
							bitSet.set(j);
						bitSet.and(reachingDefinitions);
	
						for (int j=0; j<defCount; ++j) {
							if (!bitSet.get(j))
								continue;
							webs.get(j).addRegisterPreference(i.getOp1().getBase());
						}
					} else if (i.isMoveRegToTemp()) {
						/* mov temp, reg */
						Web web = webs.get(rd.getDefinitionIndex(node));
						web.addRegisterPreference(i.getOp2().getBase());
					} else if (i.isDivOrMod()) {
						int defIdx2 = rd.getDefinitionIndex(node);
						if (defIdx2 > 0) {
							Web web = webs.get(defIdx2);
							web.addRegisterPreference(regA);
						}
					}
				}
			}

			LVFlowInfo lvf = (LVFlowInfo) lv.getFlowInfo(node);

			/* Merge intersecting chains with a common use */
			analyze(webs, reverseMapping, rd, node, lvf.use,
					liveVariables, reachingDefinitions);
		}

		Set<Web> uniqueWebs = new HashSet<Web>(webs);
		List<Integer> defList = new ArrayList<Integer>();
		List<String> coalescedBriggs = new ArrayList<String>();
		List<String> coalescedGeorge = new ArrayList<String>();

		/* Add all available registers */
		for (int i=0; i<Configuration.AVAILABLE_REGISTERS.length; ++i) {
			String regName = Configuration.AVAILABLE_REGISTERS[i].toString();
			if (registerWebs.containsKey(regName))
				continue;
			Web web = new Web(regName, null);
			registerWebs.put(regName, web);
			web.setColor(i);
		}

		/* 6. Handle special cases: call sites and divisions/modulo operations */
		for (CFGNode n : nodes) {
			if (!(n.value() instanceof Instr))
				continue;

			Instr i = (Instr) n.value();
			if (i.isCall() || i.isDivOrMod() || 
					(i.is8BitOperation() && i.getOp1().getType() == Instr.Op.Type.REG)) {
				BitSet reachingDefs = rd.getReachingDefinitionsBefore(n);

				if (i.is8BitOperation() && i.getOp1().getType() == Instr.Op.Type.REG
					&& Configuration.ARCH == Architecture.I386) {
					/* x86: Can't use registers that do not exist as 8-bit versions */
					List<Integer> defs = rd.getDefinitionsForVariable(i.getOp1().getBase());
					for (int j: defs) {
						if (!reachingDefs.get(j))
							continue;
						Web web = webs.get(j);
						web.addNeighbor(registerWebs.get(regSI));
						web.addNeighbor(registerWebs.get(regDI));
						web.addNeighbor(registerWebs.get(regBP));
					}
				}

				if (i.isCall()) {
					/* Look at live vars *after* the call (the instruction might
					 * be "call temp", and we can avoid adding an interference edge
					 * to 'temp' in this case) */
					Set<String> liveVars = lv.getLiveVariablesAfter(n);
					reachingDefs = rd.getReachingDefinitionsAfter(n);

					for (String var : liveVars) {
						List<Integer> defs = rd.getDefinitionsForVariable(var);
						for (int j: defs) {
							if (reachingDefs.get(j)) {
								Web web = webs.get(j);
	
								for (String reg : Configuration.CALLER_SAVE_REGS)
									web.addNeighbor(registerWebs.get(reg));
							}
						}
					}
				} else if (i.isDivOrMod()) {
					Set<String> liveVars = lv.getLiveVariablesBefore(n);
					for (String var : liveVars) {
						List<Integer> defs = rd.getDefinitionsForVariable(var);
						for (int j: defs) {
							if (reachingDefs.get(j)) {
								Web web = webs.get(j);

								if (!web.defs().contains(n)) {
									web.addNeighbor(registerWebs.get(regA));
									web.addNeighbor(registerWebs.get(regD));
								}
							}
						}
					}
				}
			}
		}
		
		/* 5. Build the interference graph - uses the improved
	       interference detection criterion explained in 16.3.4 */
		for (Web w1: uniqueWebs) {
			defList.clear();
			for (CFGNode def : w1.defs())
				defList.add(rd.getDefinitionIndex(def));
	
			for (Web w2: uniqueWebs) {				
				if (w1 == w2 || w1.isNeighborOf(w2))
					continue;

				/* Be smarter about moves */
				if (w2.isStrictlyMoveRelated()) {
					boolean possibleInterference = false;

					for (String s : w2.getMoveSources()) {
						if (!w1.varNames().contains(s))
							possibleInterference = true;
					}

					if (!possibleInterference)
						continue;
				}

				/* Is w1 live at any of the definitions in w2? */
				for (CFGNode def: w2.defs()) {
					Set<String> liveVars = lv.getLiveVariablesAfter(def);
	
					/* Definitely no interference if none of the variables associated 
					   with w1 are live here. */
					boolean possibleInterference = false;
					for (String name : w1.varNames()) {
						if (liveVars.contains(name)) {
							possibleInterference = true;
							break;
						}
					}
					if (!possibleInterference)
						continue;
	
					/* It is live - check if any of the definitions 
					   of w1 reach here */
					BitSet reachingDefs = rd.getReachingDefinitionsBefore(def);
					boolean interference = false;
	
					for (int i : defList) {
						if (reachingDefs.get(i)) {
							interference = true;
							break;
						}
					}
					
					if (interference) {
						w1.addNeighbor(w2);
						break;
					}
				}
			}			
		}

		/* 6. Register coalescing */
		if (coalesceCriteria != CoalesceCriteria.COALESCE_NONE) {
			for (Web w1: uniqueWebs) {
				if (!w1.isMoveRelated())
					continue;
				if (!reverseMapping.containsKey(w1) || registerWebs.containsValue(w1))
					continue;
				
				for (Web w2 : uniqueWebs) {
					if (w1 == w2 || w1.isNeighborOf(w2)
							|| !reverseMapping.containsKey(w2)
							|| registerWebs.containsValue(w2))
						continue;

					/* Detect assignments of the form  t_i <- t_j
					 * and rewrite the uses of t_i to use t_j instead
					 */
					boolean coalescingCandidate = true;
					for (String s : w1.getMoveSources()) {
						if (!w2.varNames().contains(s))
							coalescingCandidate = false;
					}

					if (!coalescingCandidate)
						continue;

					/* Only do this if a definition of t_j actually reaches the use */
					boolean proceed = false;
					for (CFGNode def : w1.defs()) {
						Instr instr = (Instr) def.value();
						BitSet rDefs = rd.getReachingDefinitionsAfter(def);
						if (!instr.isMove() || !w1.getMoveSources().contains(instr.getOp2().getBase()))
							continue;
						String varName = instr.getOp2().getBase();
						List<Integer> allDefs = rd.getDefinitionsForVariable(varName);
						
						for (int i: allDefs) {
							if (rDefs.get(i)) {
								proceed = true;
								break;
							}
						}
						if (proceed)
							break;
					}
					
					if (proceed) {
						Set<Web> neighborSet = new HashSet<Web>(w1.neighbors());
						neighborSet.addAll(w2.neighbors());
						/* Conservative criterion (Briggs):
						   Coalesce if combined # of neighbors < # of registers */
						if (neighborSet.size() < Configuration.AVAILABLE_REGISTERS.length) {
							coalescedBriggs.add("<" + w1.varNameString() + ", "
									+ w2.varNameString() + ">");
							union(webs, reverseMapping, w2, w1);
							break;
						} else if (coalesceCriteria == CoalesceCriteria.COALESCE_BRIGGS_AND_GEORGE) {
							/* Iterated coalescing criterion (George):
							   Can coalesce if for each neighbor T of w2
							     - T interferes with w1 OR
							     - deg(t) < # of registers
							   In this case, the resulting node will have the
							   same degree as w1. */
							boolean mayCoalesce = true;
							for (Web w : w2.neighbors()) {
								if (!(w.neighbors().size() < Configuration.AVAILABLE_REGISTERS.length || w
										.isNeighborOf(w1))) {
									mayCoalesce = false;
									break;
								}
							}

							if (!mayCoalesce) {
								mayCoalesce = true;
								/* Try the other direction */
								for (Web w : w1.neighbors()) {
									if (!(w.neighbors().size() < Configuration.AVAILABLE_REGISTERS.length
										|| w.isNeighborOf(w2))) {
										mayCoalesce = false;
										break;
									}
								}
							}

							if (mayCoalesce) {
								coalescedGeorge.add("<" + w1.varNameString()
										+ ", " + w2.varNameString() + ">");
								union(webs, reverseMapping, w2, w1);
								break;
							}
						}
					}
				}
			}
		}

		if (coalescedBriggs.size() > 0)
			stats.log("regAlloc/webs", "'" + funcName
					+ "': Briggs criterion: coalesced " + coalescedBriggs);
		if (coalescedGeorge.size() > 0)
			stats.log("regAlloc/webs", "'" + funcName
					+ "': George criterion: coalesced " + coalescedGeorge);
		if (coalescedBriggs.size() > 0)
			stats.log("regAlloc/webs", "Pass "+pass+ ": '" + funcName
					+ "': Briggs criterion: coalesced " + coalescedBriggs.size() + " variables");
		if (coalescedGeorge.size() > 0)
			stats.log("regAlloc/webs", "Pass "+pass+": '" + funcName
					+ "': George criterion: coalesced " + coalescedGeorge.size() + " variables");
		uniqueWebs = new HashSet<Web>(webs);
		uniqueWebs.addAll(registerWebs.values());
		stats.stopTimer("regAlloc/webs");
		return uniqueWebs;
	}

	private static void analyze(List<Web> webs,
			Map<Web, List<Integer>> reverseMapping, ReachingDefinitions rdAnalysis,
			CFGNode node, Set<String> uses, Set<String> liveVars,
			BitSet reachingDefinitions) {
		int defCount = rdAnalysis.getDefinitions().size();
		BitSet bitSet = new BitSet(defCount);

		for (String varName : liveVars) {
			bitSet.clear();
			List<Integer> defsForVar = rdAnalysis.getDefinitionsForVariable(varName);
			if (defsForVar == null)
				throw new InternalCompilerException(
						"WebGenerator: Referenced variable '" + varName
								+ "' which has no definition!");
			for (int i : defsForVar)
				bitSet.set(i);

			/* After the following line, bitSet contains a list
			 * of all intersecting du-chains associated with a live variable
			 * at the current program point.
			 */
			bitSet.and(reachingDefinitions);

			/* Merge intersecting chains */
			Web web = null;
			for (int i=0; i<defCount; ++i) {
				if (!bitSet.get(i))
					continue;
				if (web == null) {
					web = webs.get(i);
					if (uses.contains(varName))
						web.addUse(node);
				} else {
					union(webs, reverseMapping, web, webs.get(i));
				}
			}
		}
	}

	private static void union(List<Web> webs,
			Map<Web, List<Integer>> reverseMapping, Web web1, Web web2) {
		if (web1 == web2)
			return;
		List<Integer> indices1 = reverseMapping.get(web1);
		List<Integer> indices2 = reverseMapping.get(web2);

		for (int i : indices2)
			webs.set(i, web1);
		indices1.addAll(indices2);
		reverseMapping.remove(web2);
		web1.merge(web2);
	}
}
