package xic.ast;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.ErrorSink;
import xic.ir.NameMangler;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.xi.CompilationException;
import edu.cornell.cs.cs4120.xi.Position;

public class ClassType extends TypeNode {
	private String name = null;
	private String parentName = null;
	private ClassType parent = null;
	private Map<String, TypeNode> symbolTable;
	private Map<String, Integer> dispatchTable;
	private Map<String, List<Integer>> fieldTable;
	private List<String> order;
	private boolean imported, finalized;

	public ClassType(Position pos, String name, String parentName, boolean imported) {
		super(pos);
		this.name = name;
		this.imported = imported;
		this.parentName = parentName;
		this.finalized = false;
		dispatchTable = new HashMap<String, Integer>();
		fieldTable = new HashMap<String, List<Integer>>();
		symbolTable = new HashMap<String, TypeNode>();
	}

	public ClassType(Position pos, String name, String parentName, boolean imported,
			ClassType parent, Map<String, TypeNode> symbolTable,
			Map<String, Integer> dispatchTable,
			Map<String, List<Integer>> fieldTable,
			List<String> order, boolean finalized) {
		super(pos);
		this.name = name;
		this.parentName = parentName;
		this.parent = parent;
		this.symbolTable = symbolTable;
		this.dispatchTable = dispatchTable;
		this.finalized = finalized;
		this.fieldTable = fieldTable;
		this.imported = imported;
		this.order = order;
	}
	
	/**
	 * Look up a method type by name
	 */
	public FuncType lookupMethod(String name) {
		if (symbolTable.containsKey(name) &&
				symbolTable.get(name) instanceof FuncType)
			return (FuncType) symbolTable.get(name);
		else if (parent != null)
			return parent.lookupMethod(name);
		else
			return null;
	}

	/**
	 * Look up an attribute type by name
	 */
	public TypeNode lookupAttribute(String name) {
		if (symbolTable.containsKey(name) &&
				!(symbolTable.get(name) instanceof FuncType))
			return symbolTable.get(name);
		else
			return null;
	}

	/**
	 * Find out, which class implements a particular function
	 */
	public ClassType findImplementingClass(String name) {
		if (symbolTable.containsKey(name) &&
				symbolTable.get(name) instanceof FuncType)
			return this;
		else if (parent != null)
			return parent.findImplementingClass(name);
		else
			return null;
	}

	public boolean isDerived() {
		return parentName != null;
	}
	
	public boolean isImported() {
		return imported;
	}

	public String name() {
		return name;
	}
	
	public ClassType parent() {
		return parent;
	}
	
	public String parentName() {
		return parentName;
	}
	
	public int inheritanceDepth() {
		if (parent == null)
			return 0;
		else
			return 1 + parent.inheritanceDepth();
	}

	public ClassType setParent(ClassType parent) {
		if (parent == this.parent)
			return this;
		ClassType ct = (ClassType) copy();
		ct.parent = parent;
		return ct;
	}
	

	public boolean isSubtypeOf(TypeNode node) {
		if (node instanceof UnitType) {
			return true;
		} else if (node instanceof PrimitiveType) {
			return false;
		} else if (node instanceof ClassType) {
			if (name.equals(((ClassType) node).name()))
				return true;
			else if (parent != null)
				return parent.isSubtypeOf(node);
			else
				return false;
		} else if (node instanceof TupleType) {
			TupleType tnode = (TupleType) node;
			if (tnode.elementTypes().size() == 1)
				return isSubtypeOf(tnode.elementTypes().get(0));
			else
				return false;
		} else if (node instanceof ArrayType) {
			ArrayType tnode = (ArrayType) node;
			return isSubtypeOf(tnode.elementType());
		} else if (node instanceof UnitType) {
			return true;
		} else {
			return false;
		}
	}

	public Map<String, TypeNode> getSymbolTable() {
		return symbolTable;
	}

	public Map<String, Integer> getDispatchTable() {
		return dispatchTable;
	}

	public Map<String, List<Integer>> getFieldTable() {
		return fieldTable;
	}
	
	public void setFinalized(boolean finalized) {
		this.finalized = finalized;
	}

	/**
	 * Size of all local attributes (neglecting imported ones)
	 */
	public int getSizeSelf() {
		int size = 0;
		for (List<Integer> l : fieldTable.values())
			size += l.size();
		return size;
	}
	
	public List<String> getOrder() {
		return order;
	}

	public ClassType setSymbolTable(Map<String, TypeNode> symbolTable) {
		if (symbolTable == this.symbolTable)
			return this;
		ClassType ct = (ClassType) copy();
		ct.symbolTable = symbolTable;
		return ct;
	}

	public ClassType setDispatchTable(Map<String, Integer> dispatchTable) {
		if (dispatchTable == this.dispatchTable)
			return this;
		ClassType ct = (ClassType) copy();
		ct.dispatchTable = dispatchTable;
		return ct;
	}

	public ClassType setFieldTable(Map<String, List<Integer>> fieldTable) {
		if (fieldTable == this.fieldTable)
			return this;
		ClassType ct = (ClassType) copy();
		ct.fieldTable = fieldTable;
		return ct;
	}

	public ClassType setOrder(List<String> order) {
		if (order == this.order)
			return this;
		ClassType ct = (ClassType) copy();
		ct.order = order;
		return ct;
	}

	public String dump() {
		/* Reverse the dispatch & field tables */
		int maxIndexDt = 0, maxIndexFields = 0;
		Map<Integer, String> revDt = new HashMap<Integer, String>();
		for (String name : dispatchTable.keySet()) {
			int dtIndex = dispatchTable.get(name);
			revDt.put(dtIndex, name);
			if (dtIndex > maxIndexDt)
				maxIndexDt = dtIndex;
		}
		Map<Integer, String> revFields = new HashMap<Integer, String>();
		for (String name : fieldTable.keySet()) {
			List<Integer> fieldIndices = fieldTable.get(name);
			int offset = 0;
			for (int i : fieldIndices) {
				if (fieldIndices.size() == 1)
					revFields.put(i, name);
				else
					revFields.put(i, name + "." + (offset++));
				if (i > maxIndexFields)
					maxIndexFields = i;
			}
		}

		StringBuffer buf = new StringBuffer();
		String name = this.name;
		while (name.length()+2  <=16)
			name = " " + name + " ";
		while (name.length()+1  <=16)
			name =  name + " ";
		buf.append("================="+name+"=================\n");
		if (parent != null)
			buf.append(" Parent class : " + parent + "\n");
		buf.append(" Dispatch table:\n");
		for (int index=0; index<=maxIndexDt; ++index) {
			buf.append("    " + index);
			if (revDt.containsKey(index)) {
				String funcName = revDt.get(index);
				buf.append(" : " + funcName);
				if (symbolTable.containsKey(funcName)) {
					for (int i=30-funcName.length(); i>=0; --i)
						buf.append(' ');
					buf.append("    [impl]");
				}
				buf.append("\n");
			} else {
				buf.append(" : <unused>\n");
			}
		}

		buf.append("\n Local object layout:\n");
		for (int index=1; index<=maxIndexFields; ++index) {
			buf.append("    " + index);
			if (revFields.containsKey(index)) {
				String funcName = revFields.get(index);
				buf.append(" : " + funcName);
				buf.append("\n");
			} else {
				buf.append(" : <unknown>\n");
			}
		}
		buf.append("==================================================\n");
		return buf.toString();
	}

	public String label() {
		if (parent != null)
			return "<class " + name + " extends " + parent.name() + ">";
		else
			return "<class " + name + ">";
	}


	public String getClassSizeIdentifier() {
		return "_I_size_" + NameMangler.translateName(name);
	}

	public String getDispatchTableIdentifier() {
		return "_I_vt_" + NameMangler.translateName(name);
	}

	public String getStaticInitializationIdentifier() {
		return "_I_init_" + NameMangler.translateName(name);
	}

	public String toString() {
		return label();
	}

	public ASTNode visitChildren(NodeVisitor v) {
		return this;
	}

	public boolean isIdenticalType(TypeNode node) {
		if (!(node instanceof ClassType))
			return false;
		return ((ClassType) node).name.equals(name);
	}

	public Copyable copy() {
		return new ClassType(position, name, parentName, imported, parent,
				symbolTable == null ? null : 
				new HashMap<String, TypeNode>(symbolTable),
				dispatchTable == null ? null : 
				new HashMap<String, Integer>(dispatchTable),
				fieldTable == null ? null : 
				new HashMap<String, List<Integer>>(fieldTable),
				order == null ? null :
				new ArrayList<String>(order),
				finalized
			);
	}

	public int size() {
		return 1;
	}

	public boolean isFinalized() {
		return finalized;
	}

	public void updateParent(ClassType parent) {
		this.parent = parent;
	}
	
	public void match(TypeNode node, ErrorSink sink) {
	
		if (node instanceof ClassType) {
			ClassType ct = (ClassType) node;		
			
			if ( !this.name.equals(ct.name)) {
				sink.addError(new CompilationException("Unmatched class name, excepted: "
						+this.name+", got: "+ct.name, node.position()));
				return;
			}
			if ((parent==null&&ct.parent!=null)||
					(parent!=null&&ct.parent==null)) {
				sink.addError(new CompilationException("Unmatched class parent, excepted: "
						+(this.parentName==null?"null":this.parentName)+
						", got: "+(ct.parentName==null?"null":ct.parentName), node.position()));
				return;
			}
			if (parent != null && ct.parent != null)
				parent.match(ct.parent, sink);
			Set<String> keys = this.symbolTable.keySet();
			
			Map<String,TypeNode> implFuns = new HashMap<String,TypeNode>();
			// get rid of the fields
			for (String key:ct.symbolTable.keySet()) {
				if (ct.symbolTable.get(key) instanceof FuncType) {
					implFuns.put(key, ct.symbolTable.get(key));
				}
			}
			
			if (keys.size() != implFuns.keySet().size())
				sink.addError(new CompilationException(
						"Unmatched fields, excepted: " + keys.size() + ", got: "
								+ implFuns.keySet().size(), node.position()));
			else {
				
				for (String key:keys) {
					if (!implFuns.containsKey(key)) {
						sink.addError(new CompilationException(
								"Field "+key+" is not implementated", node
										.position()));
						return;
					}
					else
						this.symbolTable.get(key).match(implFuns.get(key),sink);
				}
				// check order
				int i=0,j=0;
				for (;i<keys.size();i++,j++) {
					while (!(ct.symbolTable.get(ct.order.get(j)) instanceof FuncType))
						j++;
					if (!this.order.get(i).equals(ct.order.get(j))) {
						sink.addError(new CompilationException(
								"Misordered field: " + this.order.get(i), node.position()));
						return;
					}
				}
			}

		}
		else
			sink.addError(new CompilationException("Unmatched type, excepted: "
					+this.label()+", got: "+node.label(), node.position()));
	}
}
