package xic.ast;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import xic.ErrorSink;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.xi.CompilationException;
import edu.cornell.cs.cs4120.xi.Position;
import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

/**
 * Describes a compound types in the form of a tuple
 */
public class TupleType extends TypeNode {
	private List<TypeNode> elementTypes;

	/**
	 * Create an empty tuple type
	 */
	public TupleType() {
		super(null);
		this.elementTypes = new ArrayList<TypeNode>();
	}

	/**
	 * Create an empty tuple type
	 */
	public TupleType(Position pos) {
		super(pos);
		this.elementTypes = new ArrayList<TypeNode>();
	}

	public TupleType(Position pos, List<TypeNode> elementTypes) {
		super(pos);
		this.elementTypes = elementTypes;
	}

	/**
	 * Create a tuple type with one entry
	 */
	public TupleType(TypeNode node) {
		this();
		appendType(node);
	}

	/**
	 * Copy constructor, makes a shallow copy while
	 * providing new position information.
	 */
	public TupleType(Position pos, TupleType type) {
		super(pos);
		this.elementTypes = type.elementTypes();
	}


	public void appendType(TypeNode type) {
		elementTypes.add(type);
		extendPositionTo(type.position());
	}
	
	public List<TypeNode> elementTypes() {
		return elementTypes;
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	boolean modified = false;

    	ArrayList<TypeNode> results = new ArrayList<TypeNode>();
    	for (TypeNode typeNode : elementTypes) {
    		TypeNode newType = (TypeNode) v.visit(this, typeNode);
    		if (typeNode != newType)
    			modified = true;
    		results.add(newType);
    	}

		if (modified)
			return new TupleType(position, results);

		return this;
	}

	public Iterable<VisualizableTreeNode> children() {
		ArrayList<VisualizableTreeNode> result 
			= new ArrayList<VisualizableTreeNode>(elementTypes.size());
		result.addAll(elementTypes);
		return result;
	}

	public String label() {
		return "tupleType";
	}
	
	public boolean isSubtypeOf(TypeNode node) {
		if(node instanceof TupleType) {
			/* A tuple is a subtype of another tuple if
			 * the tuples are the same size and each element
			 * of the first tuple is a subtype of the corresponding
			 * element in the second tuple.
			 */
			TupleType tnode = (TupleType)node;
			
			if(tnode.elementTypes().size() == 1) 
				return isSubtypeOf(tnode.elementTypes().get(0));
			else
				if(tnode.elementTypes().size() != elementTypes.size())
					return false;		
						
			Iterator<TypeNode> e1types = elementTypes.iterator();
			Iterator<TypeNode> e2types = tnode.elementTypes().iterator();
			while(e1types.hasNext()) {
				if(!e1types.next().isSubtypeOf(e2types.next()))
					return false;
			}
			return true;
			
		} else if(node instanceof ArrayType) {
			/* A tuple is a subtype of an array if all elements of
			 * a tuple are subtypes of the array's element type.
			 */
			Iterator<TypeNode> etypes = elementTypes.iterator();
			ArrayType atype = (ArrayType)node;
			while(etypes.hasNext()) {
				if(!etypes.next().isSubtypeOf(atype.elementType()))
					return false;
			}
			return true;
		} else if(elementTypes.size() == 1) {
			/* A one element tuple is a subtype of the type of it's element */
			
			return elementTypes.get(0).isSubtypeOf(node);
		} else if (node instanceof UnitType)
			return true;
			
		return false;
	}
	
	public boolean indexRequiresConversion() {
		/* true only if tuple is homogeneous */
		Iterator<TypeNode> etypes = elementTypes.iterator();
		TypeNode atype = etypes.next();
		while(etypes.hasNext()) {
			if(!etypes.next().isIdenticalType(atype))
				return true;
		}
		return false;
	}
	

	public boolean isIdenticalType(TypeNode node) {
		if (!(node instanceof TupleType))
			return false;
		
		TupleType other = (TupleType) node;
		if (elementTypes.size() != other.elementTypes.size())
			return false;

		for (int i=0; i<elementTypes.size(); ++i)
			if (!elementTypes.get(i).isIdenticalType(other.elementTypes.get(i)))
				return false;

		return true;
	}


	public Copyable copy() {
		return new TupleType (position, new ArrayList<TypeNode> (elementTypes));
	}

	public String toString() {
		StringBuffer typeString = new StringBuffer();
		if (elementTypes.size() == 1)
			return elementTypes.get(0).toString();
		typeString.append('(');

		for (Iterator<TypeNode> it = elementTypes.iterator(); it.hasNext(); ) {
			typeString.append(it.next().toString());
			if (it.hasNext())
				typeString.append(", ");
		}
		typeString.append(')');
		return typeString.toString();
	}

	public int size() {
		Iterator<TypeNode> elems = elementTypes.iterator();
		int size = 0;
		while (elems.hasNext())
 			size += elems.next().size();
		return size;
	}
	
	public void match(TypeNode node, ErrorSink sink) {
		if (node instanceof TupleType) {
			TupleType tt = (TupleType) node;
			int size = this.elementTypes.size();
			if (size!=tt.elementTypes.size())
				sink.addError(new CompilationException("Unmatched tuple size, excepted: "
						+size+", got: "+tt.elementTypes.size(), node.position()));
			else {
				for (int i=0; i<size; i++) {
					this.elementTypes.get(i).match(tt.elementTypes.get(i),sink);
				}
			}
		}
		else
			sink.addError(new CompilationException("Unmatched type, excepted: "
					+this.label()+", got: "+node.label(), node.position()));
	}
}