package xic.codegen;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;
import java.util.TreeMap;

import xic.Configuration;
import xic.InternalCompilerException;
import xic.MathTools;
import xic.Statistics;
import xic.analysis.CFGGenerator;
import xic.analysis.CFGNode;
import xic.analysis.GraphColoringUtils;
import xic.analysis.LoopStructure;
import xic.analysis.WebGenerator;
import xic.analysis.WebGenerator.CoalesceCriteria;
import xic.analysis.WebGenerator.Web;
import xic.codegen.Instr.Op;
import xic.codegen.Instr.Register;
import xic.ir.IRFuncDecl;
import xic.ir.NameGenerator;


/**
 * More sophisticated register allocator using a graph coloring approach
 */
public class GraphColoringRegAlloc extends RegAlloc {
	public static boolean DEBUG_SPILLS = false;
	public static boolean VERBOSE = false;
	private Register sp;
	private int nColors, bitMask;
	private Statistics stats;
	private boolean dumpWebs;
	private CoalesceCriteria coalesceCriteria;
	private LoopStructure loopStructure;
	private NameGenerator nameGenerator;

	public GraphColoringRegAlloc(Statistics stats, boolean dumpWebs,
			CoalesceCriteria coalesceCriteria, NameGenerator nameGenerator) {
		this.nColors = Configuration.AVAILABLE_REGISTERS.length;
		this.coalesceCriteria = coalesceCriteria;
		this.stats = stats;
		this.dumpWebs = dumpWebs;
		this.nameGenerator = nameGenerator;

		/* Build a bitmask of available register colors */
		bitMask = 0;
		for (int i=0; i<nColors; ++i)
			bitMask |= 1 << i;

		sp = new Register(Register.Type.SP);
	}

	public InstrList regAlloc(String funcName, IRFuncDecl irDecl, InstrList il, boolean isFunction) {
		return regAlloc(funcName, irDecl, il, isFunction, 0);
	}

	public InstrList regAlloc(String funcName, IRFuncDecl irDecl, InstrList il, boolean isFunction, int spilledCount) {
		il = cleanupPass(il);
		
		/* Build the control flow graph */
		CFGNode cfgRoot = CFGGenerator.createCFG(funcName, il);
		Set<CFGNode> cfgNodes = cfgRoot.flatten();

		/* Generate an interference graph of webs */
		Set<Web> webs = WebGenerator.run(funcName, stats, spilledCount+1,
				cfgNodes, coalesceCriteria);

		/* Put all nodes into a priority queue */
		stats.startTimer("regAlloc/color");
		PriorityQueue<Web> pq = new PriorityQueue<Web>();
		for (Web web : webs) {
			if (!web.isColored())
				pq.add(web);
		}

		if (GraphColoringUtils.DEBUG)
			System.out.println("======================================");

		boolean colored = GraphColoringUtils.findOptimisticColoring(nColors, bitMask, pq);

		stats.stopTimer("regAlloc/color");

		/* Generate GraphViz files if requested */
		if (dumpWebs) {
			new LoopStructure(cfgRoot).annotate(webs);
			try {
				PrintWriter pw = new PrintWriter(new FileWriter("webs_" + funcName + "_"
						+ spilledCount + ".dot"));
				pw.print(GraphColoringUtils.dumpDot(webs));
				pw.close();
			} catch (IOException e) {
				throw new InternalCompilerException(
						"Could not generate debug output", e);
			}
		}


		if (!colored) {
			stats.startTimer("regAlloc/spill");
			if (spilledCount == 0) {
				stats.startTimer("regAlloc/spill/loop");
				loopStructure = new LoopStructure(cfgRoot);
				stats.stopTimer("regAlloc/spill/loop");
			}

			loopStructure.annotate(webs);
			
			Web spilledWeb = selectSpill(pq);

			InstrList rl = new InstrList();
			stats.log("regAlloc/spill", "'" + funcName + "': spilled web "
					+ spilledWeb.varNameString());

			if (Configuration.REG_NAMES.contains(spilledWeb.varNameString()))
				throw new InternalCompilerException("Tried to spill a web, which is also a register: " + spilledWeb);

			/* Oops - a web spilled - rewrite the assembly and retry */
			String tempName = "spilled_" + spilledCount + "_("
					+ spilledWeb.varNameString() + ")";
			Set<AbstractInstr> affectedUse = new HashSet<AbstractInstr>();
			Set<AbstractInstr> affectedDef = new HashSet<AbstractInstr>();

			for (CFGNode use : spilledWeb.uses())
				affectedUse.add((AbstractInstr) use.value());
			for (CFGNode def : spilledWeb.defs())
				affectedDef.add((AbstractInstr) def.value());

			for (AbstractInstr ai : il) {
				boolean useWeb = affectedUse.contains(ai);
				boolean defWeb = affectedDef.contains(ai);

				if (useWeb || defWeb)
					rewriteForSpill(useWeb, defWeb,
							rl, (Instr) ai, spilledCount, spilledWeb, tempName);
				else
					rl.add(ai);					
			}

			stats.stopTimer("regAlloc/spill");

			return regAlloc(funcName, irDecl, rl, isFunction, spilledCount + 1);
		}

		stats.log("regAlloc/webs", "'" + funcName + "': total # of webs = "
				+ webs.size() + ", # of spills = " + spilledCount);

		/* Everything worked - now generate the final assembly code */
		return searchAndReplace(funcName, irDecl, il, isFunction, webs, spilledCount);
	}

	private Web selectSpill(Collection<Web> uncoloredWebs) {
		int bestCost = Integer.MAX_VALUE;
		Web bestWeb = null;

		if (DEBUG_SPILLS)
			System.out.println("Choosing web to spill amongst "
					+ uncoloredWebs.size() + " choices:");

		for (Web web : uncoloredWebs) {
			int cost = web.spillCost();
			if (DEBUG_SPILLS)
				System.out.println("Web " + web.varNameString() + " has cost " + cost);
			if (cost < bestCost) {
				bestCost = cost;
				bestWeb = web;
			}
		}

		if (DEBUG_SPILLS)
			System.out.println("Chose web " + bestWeb.varNameString());

		if (bestWeb == null)
			throw new InternalCompilerException(
					"Unable to choose a web to spill - all choices are infinitely bad!");
		return bestWeb;
	}

	public InstrList searchAndReplace(String funcName, IRFuncDecl irDecl,
				InstrList il, boolean isFunction, Set<Web> webs, int spilledCount) {
		Map<AbstractInstr, List<Web>> instrToWebDef = 
				new HashMap<AbstractInstr, List<Web>>();
		Map<AbstractInstr, List<Web>> instrToWebUse = 
				new HashMap<AbstractInstr, List<Web>>();
		Map<String, Set<Integer>> varToReg = new TreeMap<String, Set<Integer>>();
		Set<String> usedRegisters = new HashSet<String>();
		InstrList rl = new InstrList();

		String regSP = new Register(Register.Type.SP).toString();
		String regA = new Register(Register.Type.A).toString();
		String regD = new Register(Register.Type.D).toString();

		stats.startTimer("regAlloc/translate");

		/* Create some mappings for later use */
		for (Web w: webs) {
			if (w.varNames().size() == 1 &&
					Configuration.REG_NAMES.contains(w.varNames().iterator().next()))
				continue;

			for (String name : w.varNames()) {
				if (!varToReg.containsKey(name))
					varToReg.put(name, new HashSet<Integer>());
				varToReg.get(name).add(w.color());
			}

			for (CFGNode def : w.defs()) {
				AbstractInstr ai = (AbstractInstr) def.value();
				if (!instrToWebDef.containsKey(ai))
					instrToWebDef.put(ai, new ArrayList<Web>());
				instrToWebDef.get(ai).add(w);
			}

			for (CFGNode def : w.uses()) {
				AbstractInstr ai = (AbstractInstr) def.value();
				if (!instrToWebUse.containsKey(ai))
					instrToWebUse.put(ai, new ArrayList<Web>());
				instrToWebUse.get(ai).add(w);
			}
			
			usedRegisters.add(Configuration.AVAILABLE_REGISTERS[w.color()].toString());
		}
		
		int savedRegisters = 0;
		for (String reg : Configuration.CALLEE_SAVE_REGS) {
			if (usedRegisters.contains(reg))
				++savedRegisters;
		}

		boolean hasCalls = irDecl != null ? irDecl.containsCalls() : false;
		int argCount = hasCalls ? irDecl.computeMaximumCallArguments() : 0;
		int rvCount = hasCalls ? irDecl.computeMaximumCallResults() : 0;
		if (rvCount == 1)
			rvCount = 0;
		
		/* Stack layout is:
		 * 
		 * 1. Return address
		 * 2. Any pushed callee-save registers
		 * 3. Any spilled temporaries
		 * 4. An unused stack alignment region, if necessary
		 * 5. Scratch space for calls that return tuples
		 * 6. Scratch space for arguments delivered on the stack
		 * 7. Shadow space, if necessary (->Win64)
		 */
		int baseStackSize = 1 + savedRegisters; // (1 for return address)
		int stackAlignment = Configuration.STACK_ALIGNMENT / Configuration.WORD_SIZE;
		int shadowSize = hasCalls ? Configuration.SHADOW_ZONE_SIZE / Configuration.WORD_SIZE : 0;
		int stackArgCount = Math.max(0, argCount-Configuration.PARAMETER_REGISTERS.length);

		int totalStackSize = baseStackSize + stackArgCount + rvCount + shadowSize + spilledCount;
		int alignmentShift = hasCalls ? MathTools.modulo(stackAlignment - totalStackSize, stackAlignment) : 0;
		totalStackSize += alignmentShift;

		// how much must the stack pointer manually be decreased after entering a function?
		int requiredSPOffset = totalStackSize-savedRegisters-1;
		
		// offset to the beginning of the frame
		int fpOffset = totalStackSize - 1;
		
		// offset to the region that is used to pass parameters
		int argScratchOffset = shadowSize;
		
		// offset to the region that is used to temporarily hold returned tuples
		int rvScratchOffset = argScratchOffset + stackArgCount;
		
		// offset to the region holding spilled webs
		int spillOffset = rvScratchOffset + alignmentShift + spilledCount + rvCount - 1;

		/* Step through the instruction list, search & replace */
		for (AbstractInstr ai : il) {
			/* Get the CFG node corresponding to an instruction */
			if (ai instanceof LabelInstr
					&& ((LabelInstr) ai).label().equals(funcName)) {
				rl.add(ai);
				if (VERBOSE) {
					rl.add(new AsmDirective("\t# Base stack size    = " + baseStackSize));
					rl.add(new AsmDirective("\t# Has calls?         = " + hasCalls));
					rl.add(new AsmDirective("\t# Max. argument ct   = " + argCount));
					rl.add(new AsmDirective("\t# Max. return ct     = " + rvCount));
					rl.add(new AsmDirective("\t# Spilled            = " + spilledCount));
					rl.add(new AsmDirective("\t# Shadow zone        = " + shadowSize));
					rl.add(new AsmDirective("\t# Stack alignment    = " + stackAlignment));
					rl.add(new AsmDirective("\t# Alignment shift    = " + alignmentShift));
					rl.add(new AsmDirective("\t# Total stack size   = " + totalStackSize));
					rl.add(new AsmDirective("\t# Allocated registers:"));
					for (String tempName : varToReg.keySet()) {
						String list = "";
						for (int i : varToReg.get(tempName))
							list += Configuration.AVAILABLE_REGISTERS[i].toString()
									+ ", ";
	
						rl.add(new AsmDirective("\t#   " + tempName + ": "
								+ list.substring(0, list.length() - 2)));
					}
				}

				/* Add the preamble */
				if (isFunction) {
					for (String reg : Configuration.CALLEE_SAVE_REGS) {
						if (usedRegisters.contains(reg))
							rl.add(new Instr("push", new Op(Op.Type.REG, reg)));
					}
				}

				if (requiredSPOffset > 0) {
					rl.add(new Instr("sub", new Op(Op.Type.REG, regSP), new Op(
							Configuration.WORD_SIZE * requiredSPOffset)));
				}

			} else if (ai instanceof Instr) {
				if (VERBOSE)
					rl.add(new AsmDirective("\t# Original: " + ai.toString()));
				if (((Instr) ai).isDeclaration())
					continue;

				Instr i = (Instr) ai.copy();
				i.replaceMemTempAddDisplacement(RegAlloc.FP_NAME, sp,
						(fpOffset-1) * Configuration.WORD_SIZE);
				i.replaceMemTempAddDisplacement(RegAlloc.CALL_ARG_SCRATCH, sp,
						argScratchOffset * Configuration.WORD_SIZE);
				i.replaceMemTempAddDisplacement(RegAlloc.CALL_RV_SCRATCH, sp,
						rvScratchOffset * Configuration.WORD_SIZE);
				i.replaceMemTempAddDisplacement(RegAlloc.SPILL_SCRATCH, sp,
						spillOffset * Configuration.WORD_SIZE);
				
				List<Web> relatedWebsUse = instrToWebUse.get(ai);
				List<Web> relatedWebsDef = instrToWebDef.get(ai);

				if (relatedWebsUse != null) {
					for (Web w : relatedWebsUse)
						for (String name : w.varNames())
							i.replaceUse(name, 
								Configuration.AVAILABLE_REGISTERS[w.color()]);
				}

				if (relatedWebsDef != null) {
					for (Web w : relatedWebsDef)
						for (String name : w.varNames())
							i.replaceDef(name, 
								Configuration.AVAILABLE_REGISTERS[w.color()]);
				}


				if (i.isTrivialMove())
					continue;

				if (!i.isDivOrMod()) {
					rl.add(i);
				} else {
					/* Special case for handling divisions / modulo operations */
					Instr instr = new Instr("mov",
							new Instr.Op(Instr.Op.Type.REG, regA),
							i.getOp1());

					if (!instr.isTrivialMove())
						rl.add(instr);

					rl.add(new Instr("mov", 
							new Instr.Op(Instr.Op.Type.REG, regD),
							new Instr.Op(Instr.Op.Type.REG, regA)
						));

					/* Clear D register before the division, but ensure that the sign is correct */
					rl.add(new Instr("sar", 
							new Instr.Op(Instr.Op.Type.REG, regD),
							new Instr.Op(Configuration.WORD_SIZE * 8-1)));

					rl.add(new Instr("idiv", i.getOp2()));

					instr = new Instr("mov", i.getOp1(), 
							new Instr.Op(Instr.Op.Type.REG, 
								i.getMnemonic().equals("idiv") ? regA: regD));

					if (!instr.isTrivialMove())
						rl.add(instr);
				}

				if (VERBOSE)
					rl.add(new AsmDirective(""));
			} else {
				rl.add(ai);
			}
		}

		if (requiredSPOffset > 0)
			rl.add(new Instr("add", new Op(Op.Type.REG, regSP),
					new Op(Configuration.WORD_SIZE * requiredSPOffset)));

		if (isFunction) {
			for (int i=Configuration.CALLEE_SAVE_REGS.length-1; i>=0; --i) {
				String reg = Configuration.CALLEE_SAVE_REGS[i];
				if (usedRegisters.contains(reg))
					rl.add(new Instr("pop", new Op(Op.Type.REG, reg)));
			}
		}

		if (isFunction)
			rl.add(new Instr("ret"));

		stats.stopTimer("regAlloc/translate");
		return rl;
	}

	/**
	 * Rewrite an instruction involving a spilled web
	 */
	public void rewriteForSpill(boolean rewriteUse, boolean rewriteDef, 
				InstrList rl, AbstractInstr ai, int spillsSoFar,
				Web spilledWeb, String tempName) {
		Instr.Op memOp = new Op(Op.Type.MEM, RegAlloc.SPILL_SCRATCH);
		memOp.setImmediate(-Configuration.WORD_SIZE*spillsSoFar);
		Set<String> use = new HashSet<String>(), def = new HashSet<String>();
		
		Instr i = (Instr) ai.copy();
		if (rewriteUse) {
			for (String name : spilledWeb.varNames())
				i.replaceUseByMem(name, memOp);

			i.analyze(use, def);
			for (String name : spilledWeb.varNames()) {
				if (use.contains(name)) {
					rl.add(new Instr("mov", 
							new Instr.Op(Instr.Op.Type.REG, tempName),
							memOp
					));
					i.replaceUse(name, tempName);
				}
			}
		}

		rl.add(i);

		if (rewriteDef) {
			for (String name : spilledWeb.varNames())
				i.replaceDefByMem(name, memOp);

			i.analyze(use, def);
			for (String name : spilledWeb.varNames()) {
				if (def.contains(name)) {
					rl.add(new Instr("mov", 
							memOp,
							new Instr.Op(Instr.Op.Type.REG, tempName)
					));
					i.replaceDef(name, tempName);
				}
			}
		}
	}
	
	private InstrList cleanupPass(InstrList il) {
		InstrList rl = new InstrList();
		for (AbstractInstr ai : il) {
			if (ai instanceof Instr) {
				Instr i = (Instr) ai;
				i = (Instr) i.copy();
				i.replaceAbstractRegisters();
					
				if (i.needsSplitting()) {
					splitInstr(i, rl);
				} else if (i.needsRewriting()) {
					rewriteInstr(i, rl);
				} else {
					rl.add(i);
				}
			} else {
				rl.add(ai);
			}
		}
		
		return rl;
	}

	private void splitInstr(Instr i, InstrList rl) {
		if (!i.isMove())
			throw new InternalCompilerException("splitInstr: only moves are currently supported!");
	
		String tempName = nameGenerator.nextTemp();
		Instr.Op op1 = i.getOp1(), op2 = i.getOp2();
		String mnemonic = "mov";
		if (op2.getType() == Instr.Op.Type.REG) {
			mnemonic = "lea";
			op2.setType(Instr.Op.Type.MEM);
		}
		rl.add(new Instr(mnemonic, new Instr.Op(Instr.Op.Type.REG, tempName), op2));
		rl.add(new Instr("mov", op1, new Instr.Op(Instr.Op.Type.REG, tempName)));			
	}

	private void rewriteInstr(Instr i, InstrList rl) {
		if (!i.isMove())
			throw new InternalCompilerException("rewriteInstr: only moves are currently supported!");

		i.setMnemonic("lea");
		i.getOp2().setType(Instr.Op.Type.MEM);
		rl.add(i);
	}
}