package xic.codegen;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xic.Configuration;
import xic.Configuration.Architecture;
import xic.Configuration.AssemblySyntax;
import xic.InternalCompilerException;
import xic.codegen.Instr.Op.Type;


public class Instr implements AbstractInstr {
	public static class Op {
		public enum Type {
			IMMEDIATE,  // immediate
			REG,        // reg
			MEM         // [base + index*scale + offset]
		};

		private Type type;
		private String base, index;
		private int scale;
		private long immediate;
		private String immediateName;
		private int size;

		public Op(Type type) {
			this.type = type;
			this.size = Configuration.WORD_SIZE * 8;
		}

		public Op(long immediate) {
			this.type = Type.IMMEDIATE;
			this.immediate = immediate;
			this.size = Configuration.WORD_SIZE * 8;
		}

		public Op(String immediateName) {
			this.type = Type.IMMEDIATE;
			this.immediateName = immediateName;
			this.size = Configuration.WORD_SIZE * 8;
		}

		public Op(long immediate, int size) {
			this.type = Type.IMMEDIATE;
			this.size = size;
			this.immediate = immediate;
		}

		public Op(Type type, String base) {
			this.type = type;
			this.base = base;
			this.size = Configuration.WORD_SIZE * 8;
		}

		public Op(String base, String index, int scale, long displacement) {
			this.type = Type.MEM;
			this.base = base;
			this.index = index;
			this.scale = scale;
			this.immediate = displacement;
			this.size = Configuration.WORD_SIZE * 8;
		}
		
		public void setType(Type type) {
			this.type = type;
		}

		public Type getType() {
			return type;
		}
		
		public void setImmediateName(String immediateName) {
			this.immediateName = immediateName;
		}

		public String getImmediateName() {
			return immediateName;
		}

		public boolean hasImmediateName() {
			return immediateName != null;
		}

		public void setSize(int size) {
			this.size = size;
		}
		
		public int getSize() {
			return size;
		}
		
		public void setScale(int scale) {
			this.scale = scale;
		}

		public int getScale() {
			return scale;
		}
		
		public boolean hasBase() {
			return base != null;
		}

		public void setBase(String base) {
			this.base = base;
		}

		public String getBase() {
			return base;
		}
		
		public void setImmediate(long immediate) {
			this.immediate = immediate;
		}
		
		public long getImmediate() {
			return immediate;
		}

		public boolean hasIndex() {
			return index != null;
		}
		
		public void setIndex(String index) {
			this.index = index;
		}
		
		public String getIndex() {
			return index;
		}

		public boolean isStackPointerRegOp() {
			return type == Type.REG &&
				RegAlloc.isAStackPointer(base);
		}

		public boolean isStackPointerMemOp() {
			return type == Type.MEM &&
				RegAlloc.isAStackPointer(base);
		}

		public void replaceTemp(String name, Register reg) {
			if (type == Type.REG) {
				if (name.equals(base))
					base = reg.toString(size);
			} else {
				if (name.equals(base))
					base = reg.toString();
				if (name.equals(index))
					index = reg.toString();
			}
		}

		public void replaceAbstractRegisters() {
			if (type == Type.REG) {
				if (RegAlloc.isAbstractRegisterName(base))
					base = Configuration.PARAMETER_REGISTERS[RegAlloc.getAbstractRegisterIndex(base)];
			} else if (type == Type.MEM) {
				if (base != null && RegAlloc.isAbstractRegisterName(base))
					base = Configuration.PARAMETER_REGISTERS[RegAlloc.getAbstractRegisterIndex(base)];
				if (index != null && RegAlloc.isAbstractRegisterName(index))
					index = Configuration.PARAMETER_REGISTERS[RegAlloc.getAbstractRegisterIndex(index)];
			}
		}

		public void replaceTemp(String name, String newName) {
			if (type == Type.REG) {
				if (name.equals(base))
					base = newName;
			} else if (type == Type.MEM) {
				if (name.equals(base))
					base = newName;
				if (name.equals(index))
					index = newName;
			}
		}
		

		public void replaceMemTempAddDisplacement(String name, Register reg,
				int displacement) {
			if (type == Type.MEM && name.equals(base)) {
				base = reg.toString();
				this.immediate += displacement;
			}	
		}

		public String toString() {
			if (Configuration.SYNTAX == AssemblySyntax.INTEL_SYNTAX) {
				switch (type) {
					case IMMEDIATE:
						if (immediateName == null)
							return String.valueOf(immediate);
						else if (immediate == 0)
							return immediateName;
						else
							throw new InternalCompilerException(
									"Named immediate operand with offsets are not supported! (" + immediateName + "+" + immediate + ")");
					case REG:
						return base;
					case MEM: {
							String result;
							if (size == 64) 
								result = "qword ptr ";
							else if (size == 32) 
								result = "dword ptr ";
							else if (size == 16)
								result = "word ptr ";
							else if (size == 8)
								result = "byte ptr ";
							else
								throw new InternalCompilerException("Unknown operand size " + size);
							result += "[";
							boolean empty = true;
	
							if (base != null) {
								result += base;
								empty = false;
							}
	
							if (index != null && scale != 0) {
								if (!empty)
									result += "+";
								if (scale == 1)
									result += index;
								else
									result += index + "*" + scale;
								empty = false;
							}
	
							if (immediateName != null) {
								if (!empty)
									result += "+";
								result += immediateName;
								empty = false;
							}
	
							if (immediate != 0) {
								if (immediate > 0 && !empty) 
									result += "+";
								result += immediate;
							}
	
							return result + "]";
						}
					default:
						throw new InternalCompilerException("Unknown operand type");
				}
			} else {
				switch (type) {
					case IMMEDIATE:
						if (immediateName == null)
							return "$" + String.valueOf(immediate);
						else if (immediate == 0)
							return immediateName;
						else
							throw new InternalCompilerException(
									"Named immediate operand with offsets are not supported! (" + immediateName + "+" + immediate + ")");
					case REG:
						return "%" + base;
					case MEM: {
							String result = "";

							if (immediateName != null)
								result += immediateName;

							if (immediate != 0) {
								if (immediate > 0 && immediateName != null)
									result += "+";

								result += immediate;
							}
	
							if (base != null || (index != null && scale != 0)) {
								result += "(";
								if (base != null)
									result += "%" + base;
								
								if (index != null && scale != 0)
									result += ",%" + index + "," + scale;
								result += ")";
							}
							return result;
						}
					default:
						throw new InternalCompilerException("Unknown operand type");
				}
			}
		}
		
		public Op copy() {
			Op r = new Op(this.type);
			r.setBase(this.base);
			r.setImmediate(this.immediate);
			r.setImmediateName(this.immediateName);
			r.setIndex(this.index);
			r.setScale(this.scale);
			r.setSize(this.size);
			return r;
		}

		public boolean equals(Op op) {
			if (op.base != null && base == null ||
			    op.base == null && base != null)
					return false;
			if (op.index != null && index == null ||
				op.index == null && index != null)
					return false;
			if (op.immediateName != null && immediateName == null ||
				op.immediateName == null && immediateName != null)
					return false;
			if (base  != null && !base.equals(op.base) ||
				index != null && !index.equals(op.index) ||
				immediateName != null && !immediateName.equals(op.immediateName))
				return false;

			return type == op.type
				&& immediate == op.immediate 
				&& scale == op.scale
				&& size == op.size;
		}

		
		public int hashCode() {
			int hashCode = 37;
			if (hasBase())
				hashCode += 17 * hashCode + base.hashCode();
			if (hasIndex())
				hashCode += 17 * hashCode + index.hashCode();
			if (hasImmediateName())
				hashCode += 17 * hashCode + immediateName.hashCode();
			hashCode += 17 * hashCode + type.ordinal();
			hashCode += 17 * hashCode + immediate;
			hashCode += 17 * hashCode + scale;
			hashCode += 17 * hashCode + size;
			return hashCode;
		}

		public boolean isRegister(String name) {
			return type == Type.REG && base.equals(name);
		}

		public boolean isRegister(Register reg) {
			return isRegister(reg.toString(size));
		}
	
		public boolean isImmediate(int value) {
			return type == Type.IMMEDIATE && immediate == value;
		}

		public boolean isNamedImmediate() {
			return type == Type.IMMEDIATE && immediateName != null;
		}
	};

	public static class Register {
		private Type type;

		public enum Type {
			A, B, C, D, SI, DI, SP, BP, R8, R9, R10, R11, R12, R13, R14, R15
		}

		public Register(Type t) {
			this.type = t;
		}

		public String toString() {
			return toString(Configuration.WORD_SIZE * 8);
		}

		public String toString(int size) {
			switch (type) {
			case A:
				if (size == 8)
					return "al";
				else if (size == 16)
					return "ax";
				else if (size == 32)
					return "eax";
				else
					return "rax";
			case B:
				if (size == 8)
					return "bl";
				else if (size == 16)
					return "bx";
				else if (size == 32)
					return "ebx";
				else
					return "rbx";
			case C:
				if (size == 8)
					return "cl";
				else if (size == 16)
					return "cx";
				else if (size == 32)
					return "ecx";
				else if (size == 64)
					return "rcx";
			case D:
				if (size == 8)
					return "dl";
				else if (size == 16)
					return "dx";
				else if (size == 32)
					return "edx";
				else
					return "rdx";
			case SI:
				if (size == 8) {
					if (Configuration.ARCH == Architecture.I386)
						throw new InternalCompilerException("32-bit mode: SI contains no 8-bit register");
					else
						return "sil";
				}
				else if (size == 16)
					return "si";
				else if (size == 32)
					return "esi";
				else
					return "rsi";
			case DI:
				if (size == 8) {
					if (Configuration.ARCH == Architecture.I386)
						throw new InternalCompilerException("32-bit mode: DI contains no 8-bit register");
					else
						return "dil";
				}
				else if (size == 16)
					return "di";
				else if (size == 32)
					return "edi";
				else
					return "rdi";
			case SP:
				if (size == 8) {
					if (Configuration.ARCH == Architecture.I386)
						throw new InternalCompilerException("32-bit mode: SP contains no 8-bit register");
					else
						return "spl";
				}
				else if (size == 16)
					return "sp";
				else if (size == 32)
					return "esp";
				else
					return "rsp";
			case BP:
				if (size == 8) {
					if (Configuration.ARCH == Architecture.I386)
						throw new InternalCompilerException("32-bit mode: BP contains no 8-bit register");
					else
						return "bpl";
				}
				else if (size == 16)
					return "bp";
				else if (size == 32)
					return "ebp";
				else 
					return "rbp";
			case R8:
				if (size == 8)
					return "r8b";
				else if (size == 16)
					return "r8w";
				else if (size == 32)
					return "r8d";
				else
					return "r8";
			case R9:
				if (size == 8)
					return "r9b";
				else if (size == 16)
					return "r9w";
				else if (size == 32)
					return "r9d";
				else
					return "r9";
			case R10:
				if (size == 8)
					return "r10b";
				else if (size == 16)
					return "r10w";
				else if (size == 32)
					return "r10d";
				else
					return "r10";
			case R11:
				if (size == 8)
					return "r11b";
				else if (size == 16)
					return "r11w";
				else if (size == 32)
					return "r11d";
				else
					return "r11";
			case R12:
				if (size == 8)
					return "r12b";
				else if (size == 16)
					return "r12w";
				else if (size == 32)
					return "r12d";
				else
					return "r12";
			case R13:
				if (size == 8)
					return "r13b";
				else if (size == 16)
					return "r13w";
				else if (size == 32)
					return "r13d";
				else
					return "r13";
			case R14:
				if (size == 8)
					return "r14b";
				else if (size == 16)
					return "r14w";
				else if (size == 32)
					return "r14d";
				else
					return "r14";
			case R15:
				if (size == 8)
					return "r15b";
				else if (size == 16)
					return "r15w";
				else if (size == 32)
					return "r15d";
				else
					return "r15";
			default:
				throw new InternalCompilerException("Unknown register!");
			}
		}
	}

	private static final Set<String> readWriteOps = new HashSet<String>();
	private static final Set<String> readReadOps = new HashSet<String>();
	private static final Set<String> readOps = new HashSet<String>();
	private static final Set<String> partialWriteOps = new HashSet<String>();
	private static final Set<String> noneOps = new HashSet<String>();
	
	static {
		/* Operations of the form
		 * a <- a OP b
		 */
		//Set<String> destSrcBinOps = new HashSet<String>();
		readWriteOps.add("mov"); // trivial op
		readWriteOps.add("add");
		readWriteOps.add("sub");
		readWriteOps.add("imul");
		readWriteOps.add("idiv");
		readWriteOps.add("imod");
		readWriteOps.add("and");
		readWriteOps.add("or");
		readWriteOps.add("xor");
		readWriteOps.add("shl");
		readWriteOps.add("shr");
		readWriteOps.add("sal");
		readWriteOps.add("sar");
		readWriteOps.add("lea");
		readWriteOps.add("declare");

		/* Operations, which read from both operands */
		readReadOps.add("cmp");
		readReadOps.add("test");

		/* Operations, which write to their parameter, but don't completely
		   erase what it previously contained */
		partialWriteOps.add("sete");
		partialWriteOps.add("setne");
		partialWriteOps.add("setl");
		partialWriteOps.add("setle");
		partialWriteOps.add("setg");
		partialWriteOps.add("setge");
		partialWriteOps.add("setb");
		partialWriteOps.add("seta");
		partialWriteOps.add("setbe");
		partialWriteOps.add("setae");
		partialWriteOps.add("inc");
		partialWriteOps.add("dec");
		partialWriteOps.add("neg");
		partialWriteOps.add("not");

		/* Operations, which only read from their parameter */
		readOps.add("push");
		readOps.add("jmp");
		readOps.add("je");
		readOps.add("jne");
		readOps.add("jl");
		readOps.add("jle");
		readOps.add("jg");
		readOps.add("jge");
		readOps.add("jb");
		readOps.add("ja");
		readOps.add("jbe");
		readOps.add("jae");
		readOps.add("jnz");
		readOps.add("jz");
		readOps.add("call");
		
		noneOps.add("nop");
	}

	private String mnemonic;
	private Op op1, op2;
	private int registerArgumentCount = 0; // Number of register arguments of a CALL instr.

	public enum InstrType {
		READ_WRITE,  	//	a <- a OP b : reads a and b, stores in a
		READ_READ,   	//	reads from two params
		PARTIAL_WRITE,	//	writes to param, but doesn't destroy its content
		READ,			//  has one op that it reads from
		NONE,			//	doesn't read or write operands
		OTHER			//  other unclassified instructions
	};

	private InstrType type;

	public Instr(String mnemonic) {
		this.mnemonic = mnemonic;
		this.type = findInstrType(mnemonic);
		this.op1 = null;
		this.op2 = null;
	}

	public Instr(String mnemonic, Op op1) {
		this.mnemonic = mnemonic;
		this.type = findInstrType(mnemonic);
		this.op1 = op1;
		this.op2 = null;
	}

	public Instr(String mnemonic, Op op1, Op op2) {
		this.mnemonic = mnemonic;
		this.type = findInstrType(mnemonic);
		this.op1 = op1;
		this.op2 = op2;
	}

	public void setMnemonic(String mnemonic) {
		this.mnemonic = mnemonic;
	}

	public String getMnemonic() {
		return mnemonic;
	}	
	
	public Op getOp1() {
		return op1;
	}

	public boolean hasOp1() {
		return op1 != null;
	}
	
	public void setOp1(Op op) {
		this.op1 = op;
	}
	
	public Op getOp2() {
		return op2;
	}
	
	public boolean hasOp2() {
		return op2 != null;
	}

	public void setOp2(Op op) {
		this.op2 = op;
	}
	
	public int getRegisterArgumentCount() {
		return registerArgumentCount;
	}

	public void setRegisterArgumentCount(int argumentCount) {
		this.registerArgumentCount = argumentCount;
	}
			

	private static InstrType findInstrType(String mnemonic){
		if(readWriteOps.contains(mnemonic))
			return InstrType.READ_WRITE;
		else if (readReadOps.contains(mnemonic))
			return InstrType.READ_READ;
		else if (partialWriteOps.contains(mnemonic))
			return InstrType.PARTIAL_WRITE;
		else if (readOps.contains(mnemonic))
			return InstrType.READ;
		else if (noneOps.contains(mnemonic))
			return InstrType.NONE;
		else 
			return InstrType.OTHER;
			//throw new InternalCompilerException("Unknown instruction " + mnemonic);		
	}

	public InstrType type() {
		return this.type;
	}

	/**
	 * When this is a conditional or unconditional jump, returns
	 * the jump target label.
	 */
	public String getJumpTarget() {
		return op1.getImmediateName();
	}

	/**
	 * Generate 'use' and 'def' information for an instruction
	 */
	public void analyze(Set<String> use, Set<String> def) {
		use.clear(); def.clear();
		if (type == InstrType.READ_WRITE 
				|| type == InstrType.READ_READ) {

			if (isDeclaration()) {
				def.add(op1.getBase());
				return;
			} else if (op1.getType() == Instr.Op.Type.REG) {
				if (type != InstrType.READ_READ)
					def.add(op1.getBase());
				if (!(mnemonic.equals("mov") || mnemonic.equals("lea")))
					use.add(op1.getBase());
			} else if (op1.getType() == Instr.Op.Type.MEM) {
				if (op1.hasBase())
					use.add(op1.getBase());
				if (op1.hasIndex())
					use.add(op1.getIndex());
			}

			if (op2 == null) {
				/* Do nothing */
			} else if (op2.getType() == Instr.Op.Type.REG) {
				use.add(op2.getBase());
			} else if (op2.getType() == Instr.Op.Type.MEM) {
				if (op2.hasBase())
					use.add(op2.getBase());
				if (op2.hasIndex())
					use.add(op2.getIndex());
			}
		} else if (type == InstrType.READ) {
			if (op1.getType() == Instr.Op.Type.REG) {
				use.add(op1.getBase());
			} else if (op1.getType() == Instr.Op.Type.MEM) {
				if (op1.hasBase())
					use.add(op1.getBase());
				if (op1.hasIndex())
					use.add(op1.getIndex());
			}
		} else if (type == InstrType.PARTIAL_WRITE) {
			if (op1.getType() == Instr.Op.Type.REG) {
				use.add(op1.getBase());
				def.add(op1.getBase());
			} else if (op1.getType() == Instr.Op.Type.MEM) {
				if (op1.hasBase())
					use.add(op1.getBase());
				if (op1.hasIndex())
					use.add(op1.getIndex());
			}
		} else if (type == InstrType.NONE || type == InstrType.OTHER) {
			// Nothing
		} else {
			throw new InternalCompilerException("Unknown instruction type " + type + "(" + toString() + ")");
		}

		// Don't report special stack pointer-related things
		for (Iterator<String> i = use.iterator(); i.hasNext(); )
			if (RegAlloc.isAStackPointer(i.next()))
				i.remove();
		for (Iterator<String> i = def.iterator(); i.hasNext(); ) 
			if (RegAlloc.isAStackPointer(i.next()))
				i.remove();

		if (isCall()) {
			for (int i=0; i<registerArgumentCount; ++i)
				use.add(Configuration.PARAMETER_REGISTERS[i].toString());
			def.add(Configuration.AVAILABLE_REGISTERS[0].toString());
		}
	}

	// replace a TEMP with a register
	public void replaceTemp(String name, Register reg) {
		if (op1 != null)
			op1.replaceTemp(name, reg);
		if (op2 != null)
			op2.replaceTemp(name, reg);
	}
	
	// replace abstract registers with concrete ones
	public void replaceAbstractRegisters() {
		if (op1 != null)
			op1.replaceAbstractRegisters();
		if (op2 != null)
			op2.replaceAbstractRegisters();
	}
	
	// replace a used TEMP with a register
	public void replaceUse(String name, Register reg) {
		switch (type) {
			case PARTIAL_WRITE:
				op1.replaceTemp(name, reg);
				break;
			case READ_WRITE:
				if (op1.getType() == Op.Type.MEM)
					op1.replaceTemp(name, reg);
				op2.replaceTemp(name, reg);
				break;
			case READ_READ:
				op1.replaceTemp(name, reg);
				op2.replaceTemp(name, reg);
				break;
			case READ:
				op1.replaceTemp(name, reg);
				break;
		}
	}

	public void replaceUse(String name, String newName) {
		switch (type) {
			case PARTIAL_WRITE:
				op1.replaceTemp(name, newName);
				break;
			case READ_WRITE:
				if (op1.getType() == Op.Type.MEM)
					op1.replaceTemp(name, newName);
				op2.replaceTemp(name, newName);
				break;
			case READ_READ:
				op1.replaceTemp(name, newName);
				op2.replaceTemp(name, newName);
				break;
			case READ:
				op1.replaceTemp(name, newName);
				break;
		}
	}


	// replace a defined TEMP with a register
	public void replaceDef(String name, Register reg) {
		switch (type) {
			case PARTIAL_WRITE:
				op1.replaceTemp(name, reg);
				break;
			case READ_WRITE:
				op1.replaceTemp(name, reg);
				break;
		}
	}

	public void replaceDef(String name, String newName) {
		switch (type) {
			case PARTIAL_WRITE:
				op1.replaceTemp(name, newName);
				break;
			case READ_WRITE:
				op1.replaceTemp(name, newName);
				break;
		}
	}
	// replace a TEMP with a register (use or def)
	public void replaceTemp(String name, String newName) {
		if (op1 != null)
			op1.replaceTemp(name, newName);
		if (op2 != null)
			op2.replaceTemp(name, newName);
	}

	// replace a TEMP with a given operand
	public void replaceTemp(String name, Op newOp) {
		if (op1 != null && op1.isRegister(name)) {
			newOp.setSize(op1.getSize());
			op1 = newOp;
		} else if (op2 != null && op2.isRegister(name)) {
			newOp.setSize(op2.getSize());
			op2 = newOp;
		}
	}
	
	/**
	 *  Replace a TEMP used in a memory operand by a register. Also
	 *  adds the specified displacement to the existing value
	 */
	public void replaceMemTempAddDisplacement(String name, Register reg, int displacement) {
		if (op1 != null)
			op1.replaceMemTempAddDisplacement(name, reg, displacement);
		if (op2 != null)
			op2.replaceMemTempAddDisplacement(name, reg, displacement);
	}

	/**
	 * Check if it is possible to replace a "use" TEMP with a 
	 * MEM (needed by the register allocator)
	 */
	public void replaceUseByMem(String name, Op newOp) {
		InstrType type = type();

		if (isLea())
			return;

		switch (type) {
			case READ_READ:
				if (op1.isRegister(name) && op2.getType() != Op.Type.MEM) {
					newOp.setSize(op1.getSize());
					op1 = newOp;
				} else if (op2.isRegister(name) && op1.getType() != Op.Type.MEM) {
					newOp.setSize(op2.getSize());
					op2 = newOp;
				}
				break;
			case READ_WRITE:
				if (op2.isRegister(name) && op1.getType() != Op.Type.MEM) {
					newOp.setSize(op2.getSize());
					op2 = newOp;
				}
				break;
			case READ:
				if (op1.isRegister(name)) {
					newOp.setSize(op1.getSize());
					op1 = newOp;
				}
				break;
			case PARTIAL_WRITE:
				if (op1.isRegister(name)) {
					newOp.setSize(op1.getSize());
					op1 = newOp;
				}
				break;
			case NONE:
			case OTHER:
			default:
				break;
		}
	}

	/**
	 * Check if it is possible to replace a "def" TEMP with a 
	 * MEM (needed by the register allocator)
	 */
	public void replaceDefByMem(String name, Op newOp) {
		InstrType type = type();

		if (isLea())
			return;

		switch (type) {
			case READ_READ:
				break;
			case READ_WRITE:
				if (op1.isRegister(name) && op2.getType() != Op.Type.MEM
						&& !mnemonic.equals("imul")) { // Gotta love the x86 instruction set
					newOp.setSize(op1.getSize());
					op1 = newOp;
				}
				break;
			case READ:
				break;
			case PARTIAL_WRITE:
				if (op1.isRegister(name)) {
					newOp.setSize(op1.getSize());
					op1 = newOp;
				}
				break;
			case NONE:
			case OTHER:
			default:
				break;
		}
	}

	public String toString() {
		String result = mnemonic;

		if (op1 != null) {
			if (Configuration.SYNTAX == AssemblySyntax.INTEL_SYNTAX) {
				result += " " + op1.toString();
				
				if (op2 != null)
					result += ", " + op2.toString();
			} else {
				if (op2 != null || !op1.isNamedImmediate()) {
					switch (op1.getSize()) {
						case 8: result += "b"; break;
						case 16: result += "w"; break;
						case 32: result += "l"; break;
						case 64: result += "q"; break;
						default:
					}
				}
				result += " ";

				if (isCall())
					if (!op1.isNamedImmediate())
						result += "*";

				if (op2 != null)
					result += op2.toString() + ", ";

				result += op1.toString();
			}
		}
		
		return result;
	}
	
	public AbstractInstr copy() {
		return new Instr(mnemonic, 
				(op1 == null) ? null : op1.copy(),
				(op2 == null) ? null : op2.copy()
		);
	}
	
	
	public boolean isUnconditionalJump() {
		return mnemonic.equals("jmp");
	}

	public boolean isConditionalJump() {
		return mnemonic.equals("je") || mnemonic.equals("jne") ||
			mnemonic.equals("jl") || mnemonic.equals("jle") ||
			mnemonic.equals("jg") || mnemonic.equals("jge") ||
			mnemonic.equals("ja") || mnemonic.equals("jae") ||
			mnemonic.equals("jb") || mnemonic.equals("jbe") ||
			mnemonic.equals("jz") || mnemonic.equals("jnz");
	}

	public boolean isBinaryOp() {
		return readWriteOps.contains(mnemonic);
	}

	public boolean isDeclaration() {
		return mnemonic.equals("declare");
	}

	public boolean isCall() {
		return mnemonic.equals("call");
	}

	public boolean isMove() {
		return mnemonic.equals("mov");
	}

	public boolean isLea() {
		return mnemonic.equals("lea");
	}

	public boolean isRet() {
		return mnemonic.equals("ret");
	}

	public boolean isXORSelf() {
		return mnemonic.equals("xor") && op1.equals(op2);
	}

	public boolean isAddSub() {
		return mnemonic.equals("add") || mnemonic.equals("sub");
	}

	public boolean isPushPop() {
		return mnemonic.equals("push") || mnemonic.equals("pop");
	}

	public boolean isTrivialMove() {
		return isMove() && op1.equals(op2);
	}

	public boolean isDivOrMod() {
		return mnemonic.equals("idiv") || mnemonic.equals("imod");
	}

	public boolean isMoveTempToTemp() {
		return isMove() 
			&& op1.getType() == Op.Type.REG
			&& op2.getType() == Op.Type.REG
			&& !Configuration.REG_NAMES.contains(op1.getBase())
			&& !Configuration.REG_NAMES.contains(op2.getBase());
	}

	public boolean isMoveRegToReg() {
		return isMove() 
			&& op1.getType() == Op.Type.REG
			&& op2.getType() == Op.Type.REG
			&& Configuration.REG_NAMES.contains(op1.getBase())
			&& Configuration.REG_NAMES.contains(op2.getBase());
	}

	public boolean isMoveTempToReg() {
		return isMove() 
			&& op1.getType() == Op.Type.REG
			&& op2.getType() == Op.Type.REG
			&& Configuration.REG_NAMES.contains(op1.getBase())
			&& !Configuration.REG_NAMES.contains(op2.getBase());
	}

	public boolean isMoveRegToTemp() {
		return isMove() 
			&& op1.getType() == Op.Type.REG
			&& op2.getType() == Op.Type.REG
			&& !Configuration.REG_NAMES.contains(op1.getBase())
			&& Configuration.REG_NAMES.contains(op2.getBase());
	}

	public boolean is8BitOperation() {
		return (hasOp1() && op1.getSize() == 8)
			|| (hasOp2() && op2.getSize() == 8);
	}
	
	public boolean needsRewriting() {
		/// Check if the instruction must be rewritten to have a LEA mnemonic
		return isMove() && op2.getType() == Type.REG
				&& RegAlloc.isAStackPointer(op2.getBase());
	}

	public boolean needsSplitting() {
		if (!hasOp1() || !hasOp2())
			return false;

		/// Determine whether this instruction can't be expressed
		/// using the x86 instruction set since has too many memory refs
		boolean op1mem = op1.getType() == Type.MEM;
		boolean op1sp  = op1.isStackPointerRegOp();
		boolean op2mem = op2.getType() == Type.MEM;
		boolean op2sp  = op2.isStackPointerRegOp();

		return (op1mem || op1sp) && (op2mem || op2sp);
	}

	public boolean hasStackPointerRegOp() {
		return (hasOp1() && op1.isStackPointerRegOp()) ||
			   (hasOp2() && op2.isStackPointerRegOp());
	}
}
