package xic.codegen;

import java.util.ArrayList;
import java.util.List;

import xic.Statistics;


/**
 * Peephole optimizer - moves a search&replace window over an
 * instruction sequence and performs simplifications according to
 * specified patterns.
 */
public class PeepholeOptimizer {
	private static final int WINDOW_SIZE = 2;
	
	public static InstrList transform(Statistics stats, 
			String funcName, InstrList il) {
		int numReplacements = 0;
		InstrList rl = new InstrList();
		
		List<Instr> window = new ArrayList<Instr>();

		for (int index=0; index<il.size(); ) {
			if (!(il.get(index) instanceof Instr)) {
				rl.add(il.get(index++));
				continue;
			}
			window.clear();

			for (int index2=index; index2<il.size(); ++index2) {
				if (!(il.get(index2) instanceof Instr))
					continue;
				if (il.get(index2) instanceof LabelInstr)
					break;
				window.add((Instr) il.get(index2));
				if (window.size() == WINDOW_SIZE)
					break;
			}

			int currentPos = rl.size();

			int advanceBy = optimize(rl, window);
			if (advanceBy > 1)
				numReplacements++;
			while (advanceBy > 0) {
				if (il.get(index) instanceof Instr)
					advanceBy--;
				else
					rl.add(currentPos, il.get(index));
				++index;
			}
		}

		if (numReplacements > 0)
			stats.log("peep", funcName + ": performed " + numReplacements
					+ " replacements");

		return rl;
	}

	private static int optimize(InstrList rl, List<Instr> window) {
		if (window.size() >= 2) {
			/**
			 * Matches sequences of the form
			 * 
			 * mov a,b
			 * add a,c
			 * 
			 * and replaces them with
			 * 
			 * lea a, [b+c]
			 */
			Instr move = window.get(0);
			Instr add = window.get(1);
			if (move.getMnemonic().equals("mov")
				&& add.getMnemonic().equals("add")
				&& move.getOp2().getType()!=Instr.Op.Type.MEM
				&& add.getOp2().getType()!=Instr.Op.Type.MEM
				&& move.getOp1().getType()==Instr.Op.Type.REG
				&& add.getOp1().getType()==Instr.Op.Type.REG
				&& !move.getOp2().hasImmediateName()
				&& !add.getOp2().hasImmediateName()
				&& add.getOp1().isRegister(move.getOp1().getBase())
				) {
				Instr.Op memOp = new Instr.Op(Instr.Op.Type.MEM);
				
				memOp.setBase(move.getOp2().getBase());
				memOp.setIndex(add.getOp2().getBase());
				memOp.setScale(1);

				memOp.setImmediate(move.getOp2().getImmediate()
						+ add.getOp2().getImmediate());

				rl.add(new Instr("lea",
						move.getOp1(),
						memOp));
				return 2;
			}
		}
		
		if (window.size() >= 1) {
			/**
			 * Matches instructions of the form
			 * 
			 * lea a, [register]
			 * 
			 * and replaces them with
			 * 
			 * mov a, register
			 */
			Instr i = window.get(0);
			Instr.Op op2 = i.getOp2();
			if (i.isLea() && op2.getImmediate() == 0 && op2.getImmediateName() == null && (op2.getIndex() == null || op2.getScale() == 0)) {
				i = (Instr) i.copy();
				i.getOp2().setType(Instr.Op.Type.REG);
				i.setMnemonic("mov");
				rl.add(i);
				return 1;
			}
		}

		if (window.size() >= 1) {
			/**
			 * Matches sequences of the form
			 * 
			 * mov reg, 0
			 * 
			 * and replaces them with
			 * 
			 * xor reg, reg
			 * 
			 * which occurrs when a function is called
			 * directly with the return value of another.
			 */
			Instr instr1 = window.get(0);

			if (instr1.getMnemonic().equals("mov")
				&& instr1.getOp1().getType() == Instr.Op.Type.REG
				&& instr1.getOp2().isImmediate(0)
				&& instr1.getOp2().getImmediateName() == null) {
				rl.add(new Instr("xor",
						new Instr.Op(Instr.Op.Type.REG, instr1.getOp1().getBase()),
						new Instr.Op(Instr.Op.Type.REG, instr1.getOp1().getBase())
				));
				return 1;
			}
		}

		if (window.size() > 0) {
			rl.add(window.get(0));
			return 1;
		} else {
			return 0;
		}
	}
}
