package xic.codegen.matchers;

import java.util.List;

import xic.InternalCompilerException;
import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.Instr.Op.Type;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRBinOp;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;
import xic.ir.IRBinOp.OpType;


/**
 * Handles binary operations and certain moves of the form:
 * MEM(x)  <- MEM(x)  op TILE
 * TEMP(x) <- TEMP(x) op TILE
 * MEM(x)  <- TILE op MEM(x)  (where op is commutative)
 * TEMP(x) <- TILE op TEMP(x) (where op is commutative)
 */
public class BinOpMatcher implements Matcher {
	protected static class BinOpTile extends AbstractTile {
		public enum TileType {
			NORMAL,      ///      a =  b  OP  c
			MEM_LEFT,    ///      a = [b] OP  c
			MEM_RIGHT,   ///      a =  b  OP [c]
		};

		private TileType tileType;  /// See above
		private OpType opType;      /// Operation type from IRBinOp
		private AddrMode addrMode;  /// Addressing mode in case there is a memory access
		private boolean overwrite;  /// Should the left-hand side be overwritten by the operation?
		private boolean jumpOnResult = false;  /// Should the result of a compare be stored or used in a jcc?
		private String jumpTarget = null;

		public void setJumpOnResult(boolean jumpOnResult) {
			this.jumpOnResult = jumpOnResult;
		}
		
		public boolean jumpOnResult(){
			return jumpOnResult;
		}
		public void setJumpTarget(String target) {
			this.jumpTarget = target;
		}
		
		public String jumpTarget(){
			return jumpTarget;
		}

		/**
		 * Construct a binary operation tile without memory access
		 */
		public BinOpTile(OpType opType, Tile leftTile, Tile rightTile, boolean overwrite) {
			this.tileType = TileType.NORMAL;
			this.opType = opType;
			this.overwrite = overwrite;
			children.add(leftTile);
			children.add(rightTile);
		}

		/**
		 * Construct a binary operation tile with memory access
		 */
		public BinOpTile(TileType tileType, OpType opType, AddrMode addrMode, Tile exprTile, boolean overwrite) {
			this.tileType = tileType;
			this.opType = opType;
			this.addrMode = addrMode;
			this.overwrite = overwrite;
			children.add(exprTile);
		}

		/**
		 * Constants can sometimes be integrated directly into instruction,
		 * and are thus "free" (which is why this function is overridden).
		 */
		@Override
		public int estimateCost() {
			if (cost != -1)
				return cost;

			boolean division = (opType == OpType.DIV || opType == OpType.MOD);

			cost = 0;
			if (tileType == TileType.MEM_LEFT) {
				cost += addrMode.estimateCost();
				/// Division doesn't handle constants on the right-hand side
				if (division || !(children.get(0) instanceof ConstMatcher.ConstTile))
					cost += children.get(0).estimateCost();
			} else if (tileType == TileType.MEM_RIGHT) {
				cost += addrMode.estimateCost();
				if (!(children.get(0) instanceof ConstMatcher.ConstTile))
					cost += children.get(0).estimateCost();
			} else {
				if (!(children.get(0) instanceof ConstMatcher.ConstTile))
					cost += children.get(0).estimateCost();
				if (division || !(children.get(1) instanceof ConstMatcher.ConstTile))
					cost += children.get(1).estimateCost();
			}
			cost += localCost();


			return cost;
		}

		protected int localCost() {
			int cost;
			switch (opType) {
				case MOD:
				case DIV: cost = 26; break;
				case MUL: cost = 3; break;
				default: cost = 1;
			}
			if (!jumpOnResult) {
				if (!overwrite)
					cost += 1;
			}
			return cost;
		}

		public String translate(InstrList il, NameGenerator ng) {
			String mnemonic = null;
			boolean comparison = false, division = false;

			switch (opType) {
				case ADD: mnemonic = "add"; break;
				case SUB: mnemonic = "sub"; break;
				case MUL: mnemonic = "imul"; break;
				case DIV: mnemonic = "idiv"; division = true; break;
				case MOD: mnemonic = "imod"; division = true; break;
				 // Fake imod instruction -- defer to register allocator
				case AND: mnemonic = "and"; break;
				case OR:  mnemonic = "or"; break;
				case LOGAND: mnemonic = "and"; break;
				case LOGOR: mnemonic = "or"; break;
				case XOR: mnemonic = "xor"; break;
				case LSHIFT: mnemonic = "shl"; break;
				case RSHIFT: mnemonic = "shr"; break;
				case AR_LSHIFT: mnemonic = "sal"; break;
				case AR_RSHIFT: mnemonic = "sar"; break;
				case EQ:  mnemonic = jumpOnResult ? "je" : "sete"; comparison = true; break;
				case NEQ: mnemonic = jumpOnResult ? "jne" : "setne"; comparison = true; break;
				case LT:  mnemonic = jumpOnResult ? "jl" : "setl"; comparison = true; break;
				case GT:  mnemonic = jumpOnResult ? "jg" : "setg"; comparison = true; break;
				case LE:  mnemonic = jumpOnResult ? "jle" : "setle"; comparison = true; break;
				case GE:  mnemonic = jumpOnResult ? "jge" : "setge"; comparison = true; break;
				case ULT: mnemonic = jumpOnResult ? "jb" : "setb"; comparison = true; break;
				case UGT: mnemonic = jumpOnResult ? "ja" : "seta"; comparison = true; break;
				case ULE: mnemonic = jumpOnResult ? "jbe" : "setbe"; comparison = true; break;
				case UGE: mnemonic = jumpOnResult ? "jae" : "setae"; comparison = true; break;
				default:
					throw new InternalCompilerException("Unknown operation");
			}

			Instr.Op leftOp, rightOp;

			if (addrMode != null) {
				if (tileType == TileType.MEM_LEFT) {
					leftOp = addrMode.toOp(il, ng);
					/// Division doesn't handle constants on the right-hand side
					if (!division && children.get(0) instanceof ConstMatcher.ConstTile
							&& ((ConstMatcher.ConstTile) children.get(0)).is32BitOrLess())
						rightOp = new Instr.Op(((ConstMatcher.ConstTile) children.get(0)).value());
					else
						rightOp = new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng));
				} else {
					if (children.get(0) instanceof ConstMatcher.ConstTile
							&& ((ConstMatcher.ConstTile) children.get(0)).is32BitOrLess())
						leftOp = new Instr.Op(((ConstMatcher.ConstTile) children.get(0)).value());
					else
						leftOp = new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng));
					rightOp = addrMode.toOp(il, ng);
				}
			} else {
				if (children.get(0) instanceof ConstMatcher.ConstTile
						&& ((ConstMatcher.ConstTile) children.get(0)).is32BitOrLess())
					leftOp = new Instr.Op(((ConstMatcher.ConstTile) children.get(0)).value());
				else
					leftOp = new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng));

				/// Division doesn't handle constants on the right-hand side
				if (!division && children.get(1) instanceof ConstMatcher.ConstTile 
						&& ((ConstMatcher.ConstTile) children.get(1)).is32BitOrLess())
					rightOp = new Instr.Op(((ConstMatcher.ConstTile) children.get(1)).value());
				else
					rightOp = new Instr.Op(Instr.Op.Type.REG, children.get(1).translate(il, ng));
			}

			if (!comparison) {
				if (!overwrite) {
					String tempName = ng.nextTemp();
					il.add(new Instr("mov",
							new Instr.Op(Instr.Op.Type.REG, tempName),
							leftOp));
					il.add(new Instr(mnemonic,
							new Instr.Op(Instr.Op.Type.REG, tempName),
							rightOp));

					if (jumpOnResult) {
						il.add(new Instr("test", new Instr.Op(
								Instr.Op.Type.REG, tempName), new Instr.Op(
								Instr.Op.Type.REG, tempName)));
						Instr.Op op = new Instr.Op(Instr.Op.Type.IMMEDIATE);
						op.setImmediateName(jumpTarget);
						il.add(new Instr("jnz", op));
						return null;
					}
					return tempName;
				} else {
					il.add(new Instr(mnemonic, leftOp, rightOp));
					return null; // This is a statement
				}
					
			} else {
				if (leftOp.getType() == Type.IMMEDIATE) {
					String tempName = ng.nextTemp();
					/// Can't have a constant on the left-hand side of a CMP
					il.add(new Instr("mov",
							new Instr.Op(Instr.Op.Type.REG, tempName),
							leftOp
						));
					leftOp = new Instr.Op(Instr.Op.Type.REG, tempName); 
				}
				String tempName = null;

				if (!jumpOnResult) {
					/* XOR clears some flags - do this before the CMP */
					tempName = ng.nextTemp();
					il.add(new Instr("xor", 
							new Instr.Op(Instr.Op.Type.REG, tempName),
							new Instr.Op(Instr.Op.Type.REG, tempName)));
				}

				il.add(new Instr("cmp", leftOp, rightOp));

				if (!jumpOnResult) {
					Instr.Op target = new Instr.Op(Instr.Op.Type.REG, tempName);
					target.setSize(8);
					il.add(new Instr(mnemonic, target));
					return tempName;
				} else {
					Instr.Op op = new Instr.Op(Instr.Op.Type.IMMEDIATE);
					op.setImmediateName(jumpTarget);
					il.add(new Instr(mnemonic,op));
					return null;
				}
			}
		}
	}

	public Tile match(IRNode node, Tiler tiler) {
		Tile bestTile = null;
		int bestCost = Integer.MAX_VALUE;
		
		if (node instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) node;
			
			if (binOp.left() instanceof IRConst && !(binOp.right() instanceof IRConst) && binOp.swapArgs() != null) {
				/* Can't have constants on the left side of x86 BinOp-style mnemonics -- just flip
				   the operands if possible -- this reduces register pressure */
				binOp = binOp.swapArgs();
			}
				
			Tile leftTile = tiler.tile(binOp.left());
			Tile rightTile = tiler.tile(binOp.right());
			
			
			// Handle operations of the form
			//    TILE op TILE
			bestTile = new BinOpTile(binOp.opType(), leftTile, rightTile, false);
			bestCost = bestTile.estimateCost();

			if (binOp.left() instanceof IRMem) {
				// Handle operations of the form
				//    MEM(x) op TILE
				IRMem mem = (IRMem) binOp.left();
				List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);

				for (AddrMode addrMode : modes) {
					Tile tile = new BinOpTile(BinOpTile.TileType.MEM_LEFT, binOp.opType(), addrMode, rightTile, false);
					int cost = tile.estimateCost();
					if (cost < bestCost) {
						bestTile = tile;
						bestCost = cost;
					}
				}
			}
	
			if (binOp.right() instanceof IRMem) {
				// Handle operations of the form
				//    TILE op MEM(x)
				IRMem mem = (IRMem) binOp.right();
				List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);

				for (AddrMode addrMode : modes) {
					Tile tile = new BinOpTile(BinOpTile.TileType.MEM_RIGHT, binOp.opType(), addrMode, leftTile, false);
					int cost = tile.estimateCost();
					if (cost < bestCost) {
						bestTile = tile;
						bestCost = cost;
					}
				}
			}
		} else if (node instanceof IRMove) {
			// Handle operations of the form
			//    MEM(x)  <- MEM(x)  op TILE
			//    TEMP(x) <- TEMP(x) op TILE
			//    MEM(x)  <- TILE op MEM(x)  (where OP is commutative)
			//    TEMP(x) <- TILE op TEMP(x) (where OP is commutative)

			IRMove move = (IRMove) node;

			if (move.expr() instanceof IRBinOp
					/* The intel instruction set does not support 'imul [addr], reg' */
					&& ((IRBinOp) move.expr()).opType() != IRBinOp.OpType.MUL 
				&& (move.target() instanceof IRMem || move.target() instanceof IRTemp)) {
				IRExpr target = ((IRMove) node).target();
				IRBinOp binOp = (IRBinOp) ((IRMove) node).expr();
				Tile leftTile = tiler.tile(binOp.left());
				Tile rightTile = tiler.tile(binOp.right());

				boolean commutative;
				switch (binOp.opType()) {
					case ADD:
					case MUL:
					case AND:
					case OR:
					case LOGAND:
					case LOGOR:
					case XOR:
						commutative = true;
						break;
					case EQ:  
					case NEQ:
					case LT:
					case GT:
					case LE:
					case GE: 
					case ULT:
					case UGT:
					case ULE:
					case UGE:
						return bestTile; // We don't handle comparisons here
					default:
						commutative = false;
				}

				if (target.equalsTree(binOp.left())) {
					if (binOp.left() instanceof IRMem) {
						List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) binOp.left());
						for (AddrMode addrMode : modes) {
							Tile tile = new BinOpTile(BinOpTile.TileType.MEM_LEFT, binOp.opType(), addrMode, rightTile, true);
							int cost = tile.estimateCost();
							if (cost < bestCost) {
								bestTile = tile;
								bestCost = cost;
							}
						}
					} else if (binOp.left() instanceof IRTemp) {
						bestTile = new BinOpTile(binOp.opType(), leftTile, rightTile, true);

						if (binOp.right() instanceof IRMem) {
							List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) binOp.right());
							for (AddrMode addrMode : modes) {
								Tile tile = new BinOpTile(BinOpTile.TileType.MEM_RIGHT, binOp.opType(), addrMode, leftTile, true);
								int cost = tile.estimateCost();
								if (cost < bestCost) {
									bestTile = tile;
									bestCost = cost;
								}
							}
						}
					}
				} else if (commutative && target.equalsTree(binOp.right())) {
					if (binOp.right() instanceof IRMem) {
						List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) binOp.right());
						for (AddrMode addrMode : modes) {
							Tile tile = new BinOpTile(BinOpTile.TileType.MEM_LEFT, binOp.opType(), addrMode, leftTile, true);
							int cost = tile.estimateCost();
							if (cost < bestCost) {
								bestTile = tile;
								bestCost = cost;
							}
						}
					} else if (binOp.right() instanceof IRTemp) {
						bestTile = new BinOpTile(binOp.opType(), rightTile, leftTile, true);

						if (binOp.left() instanceof IRMem) {
							List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) binOp.left());
							for (AddrMode addrMode : modes) {
								Tile tile = new BinOpTile(BinOpTile.TileType.MEM_RIGHT, binOp.opType(), addrMode, rightTile, true);
								int cost = tile.estimateCost();
								if (cost < bestCost) {
									bestTile = tile;
									bestCost = cost;
								}
							}
						}
					}
				}
			}
		}

		return bestTile;
	}
}
