package xic.codegen.matchers;

import java.util.List;

import xic.InternalCompilerException;
import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.Instr.Op;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRCJump;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;


public class CJumpMatcher extends BinOpMatcher {

	protected static class CJumpTile extends AbstractTile {
		public enum TileType {
			REG,      	///     CJUMP(TEMP(X),    t, f)
			MEM,		///		CJUMP(MEM(X),     t, f)
			BINOP,		/// 	CJUMP(BINOP(X,Y), t, f)
			OTHER		/// 	CJUMP(tile, t f)
		};
		private TileType tileType;
		private String tempName;
		private String trueLabel;
		private String falseLabel;
		private AddrMode addrMode;
		
		public CJumpTile(String tempName, String trueLabel, String falseLabel) {
			this.trueLabel = trueLabel;
			this.falseLabel = falseLabel;
			this.tileType = TileType.REG;
			this.tempName = tempName;
		}

		public CJumpTile(Tile condTile, String trueLabel, String falseLabel) {
			this.trueLabel = trueLabel;
			this.falseLabel = falseLabel;
			children.add(condTile);
			
			if (condTile instanceof BinOpTile) {
				/* for BinOps we can optimize */
				this.tileType = TileType.BINOP;
				BinOpTile binOpTile = (BinOpTile)condTile;
				binOpTile.setJumpOnResult(true);
				binOpTile.setJumpTarget(trueLabel);
			} else {
				this.tileType = TileType.OTHER;				
			}
		}

		public CJumpTile(AddrMode addrMode, String trueLabel,
				String falseLabel) {
			this.addrMode = addrMode;
			this.trueLabel = trueLabel;
			this.falseLabel = falseLabel;
			this.tileType = TileType.MEM;
		}

		protected int localCost() {

			switch(this.tileType){
			case BINOP:
				return 1;
			case MEM:
				return addrMode.estimateCost() + 3;
			default:
				return 3;
			}
		}

		public String translate(InstrList il, NameGenerator ng) {
			/* CJUMP(TEMP, t, f);
			 */
			Op op1, op2, trueTarget, falseTarget;
			trueTarget = new Instr.Op(Instr.Op.Type.IMMEDIATE);
			trueTarget.setImmediateName(trueLabel);	
			
			falseTarget = new Instr.Op(Instr.Op.Type.IMMEDIATE);
			falseTarget.setImmediateName(falseLabel);	

			switch(tileType) {
			case OTHER:
				tempName = children.get(0).translate(il, ng);
				/* fall-through */
			case REG:
				op1 = new Instr.Op(Instr.Op.Type.REG, tempName);
				op2 = new Instr.Op(Instr.Op.Type.REG, tempName);
				il.add(new Instr("test", op1, op2));
				il.add(new Instr("jnz", trueTarget));
				break;
			case MEM:
				op1 = addrMode.toOp(il, ng);
				op2 = new Instr.Op(0x1,8);
				il.add(new Instr("cmp", op1, op2));
				il.add(new Instr("je", trueTarget));
				break;
			case BINOP:
				/* generates a cmp and jcc */
				children.get(0).translate(il, ng);
				break;
			default:
				throw new InternalCompilerException("unknown tile type");
			}
			if (falseLabel != null)
				il.add(new Instr("jmp", falseTarget));
			return null;
		}
		
	}
	public Tile match(IRNode node, Tiler tiler) {
		if (!(node instanceof IRCJump))
			return null;
		
		IRCJump cjump = (IRCJump) node;
		IRExpr expr = cjump.expr();
		String trueLabel = cjump.trueLabel();
		String falseLabel = cjump.falseLabel();
		
		Tile bestTile = null;
		int bestCost = Integer.MAX_VALUE;
		
		if (trueLabel == null) {
			trueLabel = falseLabel;
			falseLabel = null;
			expr = expr.negate();
		}

		// cjump Tile, trueLabel, falseLabel
		Tile condTile = tiler.tile(expr);


		bestTile = new CJumpTile(condTile,trueLabel,falseLabel);
		if (expr instanceof IRTemp) {
			IRTemp temp = (IRTemp) expr;
			Tile tempTile = new CJumpTile(temp.name(), trueLabel, falseLabel);
			int cost = tempTile.estimateCost();
			if (cost < bestCost) {
				bestTile = tempTile;
				bestCost = cost;
			}
		} else if (expr instanceof IRMem) {
			IRMem mem = (IRMem) expr;
			List<AddrMode> modes = AddrModeHelper.findMemAddressingModes(tiler, mem);

			for (AddrMode addrMode : modes) {
				Tile tile = new CJumpTile(addrMode, trueLabel, falseLabel);
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
		}

		return bestTile;
	}
}