package xic.codegen.matchers;

import java.util.List;

import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRBinOp;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;


/**
 * Generates LEA instructions for suitable IR subtrees
 */
public class LeaMatcher implements Matcher {
	protected static class LeaTile extends AbstractTile {
		private AddrMode addrMode;

		public LeaTile(AddrMode addrMode) {
			this.addrMode = addrMode;
		}

		public LeaTile(AddrMode addrMode, Tile dest) {
			this.addrMode = addrMode;
			children.add(dest);
		}

		public int localCost() {
			return 1 + addrMode.estimateCost();
		}

		public String translate(InstrList il, NameGenerator ng) {
			String tempName;
	
			if (children.size() > 0)
				tempName = children.get(0).translate(il, ng);
			else
				tempName = ng.nextTemp();

			il.add(new Instr("lea",
					new Instr.Op(Instr.Op.Type.REG, tempName),
					addrMode.toOp(il, ng)
				));
			return tempName;
		}
	};

	public Tile match(IRNode node, Tiler tiler) {
		Tile bestTile = null;
		int bestCost = Integer.MAX_VALUE;

		if (node instanceof IRBinOp) {
			IRBinOp binOp = (IRBinOp) node;
			List<AddrMode> modes = AddrModeHelper.findLeaAddressingModes(tiler, binOp);

			for (AddrMode addrMode : modes) {
				if (addrMode.hasBase() && !addrMode.hasImmediateName() && 
						(!addrMode.hasIndex() || addrMode.getScale() == 1))
					continue;
				Tile tile = new LeaTile(addrMode);
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
		}
		
		if (node instanceof IRMove
				&& ((IRMove) node).target() instanceof IRTemp
				&& ((IRMove) node).expr() instanceof IRBinOp) {
			IRMove move = (IRMove) node;
			IRBinOp binOp = (IRBinOp) move.expr();
			IRTemp temp = (IRTemp) move.target();
			List<AddrMode> modes = AddrModeHelper.findLeaAddressingModes(tiler, binOp);

			for (AddrMode addrMode : modes) {
				if (addrMode.hasBase() && !addrMode.hasImmediateName() && 
						(!addrMode.hasIndex() || addrMode.getScale() == 1))
					continue;

				Tile tile = new LeaTile(addrMode, tiler.tile(temp));
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
		}

		return bestTile;
	}
}
