package xic.codegen.matchers;

import java.util.List;

import xic.codegen.AbstractTile;
import xic.codegen.Instr;
import xic.codegen.InstrList;
import xic.codegen.Matcher;
import xic.codegen.Tile;
import xic.codegen.Tiler;
import xic.codegen.matchers.AddrModeHelper.AddrMode;
import xic.ir.IRExpr;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRTemp;
import xic.ir.NameGenerator;


public class MoveMatcher implements Matcher {
	protected static class MoveToTempTile extends AbstractTile {
		public enum TileType {
			NORMAL,      ///       a  =  b
			MEM_LEFT,    ///      [a] =  b
			MEM_RIGHT,   ///       a  = [b]
		};

		private TileType tileType;
		private AddrMode addrMode;

		public MoveToTempTile(Tile leftTile, Tile rightTile) {
			this.tileType = TileType.NORMAL;
			children.add(leftTile);
			children.add(rightTile);
		}

		public MoveToTempTile(TileType tileType, AddrMode addrMode, Tile otherTile) {
			this.tileType = tileType;
			this.addrMode = addrMode;
			children.add(otherTile);
		}

		/**
		 * Constants are integrated directly into instruction,
		 * and are thus "free" (which is why this function is overridden).
		 */
		@Override
		public int estimateCost() {
			if (cost != -1)
				return cost;

			cost = 0;
			for (Tile t : children) {
				if (!(t instanceof ConstMatcher.ConstTile) && !(t instanceof NameMatcher.NameTile))
					cost += t.estimateCost();
			}
			
			cost += localCost();
			return cost;
		}

		public int localCost() {
			return 1 + (addrMode != null ? addrMode.estimateCost() : 0);
		}

		public String translate(InstrList il, NameGenerator ng) {
			Instr.Op op0, op1 = null;
			
			if (children.get(0) instanceof ConstMatcher.ConstTile)
				op0 = new Instr.Op(((ConstMatcher.ConstTile) children.get(0)).value());
			else
				op0 = new Instr.Op(Instr.Op.Type.REG, children.get(0).translate(il, ng));

			if (children.size() > 1) {
				if (children.get(1) instanceof ConstMatcher.ConstTile) {
					op1 = new Instr.Op(((ConstMatcher.ConstTile) children.get(1)).value());
				} else {
					op1 = new Instr.Op(Instr.Op.Type.REG, children.get(1).translate(il, ng));
					if (!op1.hasBase())
						return null; /// Unnecessary move (-> CallMatcher) - don't generate any code
				}
			}

			if (tileType == TileType.NORMAL) {
				il.add(new Instr("mov", op0, op1));
			} else if (tileType == TileType.MEM_LEFT) {
				il.add(new Instr("mov",  addrMode.toOp(il, ng), op0));
			} else if (tileType == TileType.MEM_RIGHT) {
				il.add(new Instr("mov", op0, addrMode.toOp(il, ng)));
			}
			return null;
		}
	}

	public Tile match(IRNode node, Tiler tiler) {
		if (!(node instanceof IRMove))
			return null;

		IRMove move = (IRMove) node;
		IRNode target = move.target();
		IRExpr expr = move.expr();
		Tile leftTile = tiler.tile(target);
		Tile rightTile = tiler.tile(expr);
		Tile bestTile = null;
		int bestCost = Integer.MAX_VALUE;

		if (target instanceof IRTemp) {
			bestTile = new MoveToTempTile(leftTile, rightTile);
			bestCost = bestTile.estimateCost();

			if (expr instanceof IRMem) {
				List<AddrMode> addrModes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) expr);
				
				for (AddrMode addrMode : addrModes) {
					Tile tile = new MoveToTempTile(MoveToTempTile.TileType.MEM_RIGHT, addrMode, leftTile);
					int cost = tile.estimateCost();
					if (cost < bestCost) {
						bestTile = tile;
						bestCost = cost;
					}
				}
			}
		} else if (target instanceof IRMem) {
			List<AddrMode> addrModes = AddrModeHelper.findMemAddressingModes(tiler, (IRMem) target);

			for (AddrMode addrMode : addrModes) {
				Tile tile = new MoveToTempTile(MoveToTempTile.TileType.MEM_LEFT, addrMode, rightTile);
				int cost = tile.estimateCost();
				if (cost < bestCost) {
					bestTile = tile;
					bestCost = cost;
				}
			}
		}
		return bestTile;
	}
}
