package xic.ir;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import xic.ast.ASTNode;
import xic.ast.Copyable;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

public class IRCall extends IRExpr {
	private IRExpr target;
	private List<IRExpr> params;
	int returnSize;

	public IRCall(IRExpr target, List<IRExpr> params, int returnSize) {
		this.target = target;
		this.params = params;
		this.returnSize = returnSize;
	}

	public IRCall(IRExpr target, int returnSize) {
		this.target = target;
		this.params = new ArrayList<IRExpr>();
		this.returnSize = returnSize;
	}

	public int returnSize() {
		return returnSize;
	}
	
	public boolean isToProcedure() {
		return returnSize == 0;
	}

	public IRExpr target() {
		return target;
	}
	
	public List<IRExpr> params() {
		return params;
	}


	public String label() {
		return "CALL";
	}

	public void appendParam(IRExpr irExpr) {
		params.add(irExpr);
	}
	public void insertParam(int index, IRExpr irExpr) {
		params.add(index,irExpr);
	}
    public ASTNode visitChildren(NodeVisitor v) {
    	boolean modified = false;
    	IRExpr target = (IRExpr) v.visit(this, this.target);
    	
    	if (target != this.target)
    		modified = true;

    	ArrayList<IRExpr> results = new ArrayList<IRExpr>();
    	for (IRExpr param : params) {
    		IRExpr newExpr = (IRExpr) v.visit(this, param);
    		if (newExpr != param)
    			modified = true;
    		results.add(newExpr);
    	}

		if (modified)
			return new IRCall(target, results, returnSize);

		return this;
	}
    
    public Copyable copy() {
		return new IRCall(target, new ArrayList<IRExpr>(params), returnSize);
    }
    
    public Copyable deepCopy() {
    	List<IRExpr> exprs = new ArrayList<IRExpr>();
    	for (IRExpr e : this.params)
    		exprs.add((IRExpr) e.deepCopy());
		return new IRCall((IRExpr) target.deepCopy(), exprs, returnSize);
    }


	public Iterable<VisualizableTreeNode> children() {
		ArrayList<VisualizableTreeNode> result 
			= new ArrayList<VisualizableTreeNode>(params.size());
		result.add(target);
		result.addAll(params);
		return result;
	}
	
	public boolean containsCalls() {
		return true;
	}
	
	public boolean containsMemReads() {
		boolean results = target.containsMemReads();
		for (IRExpr expr : params)
			results |= expr.containsMemReads();
		return results;
	}

	public boolean containsMemWrites() {
		boolean results = target.containsMemWrites();
		for (IRExpr expr : params)
			results |= expr.containsMemWrites();
		return results;
	}

	public void findReadTemps(Set<String> tempNames) {
		target.findReadTemps(tempNames);
		for (IRExpr expr : params)
			expr.findReadTemps(tempNames);
	}

	public void findWrittenTemps(Set<String> tempNames) {
		target.findWrittenTemps(tempNames);
		for (IRExpr expr : params)
			expr.findWrittenTemps(tempNames);
	}

	public int nodeCount() {
		int nodeCount = 0;
		for (IRExpr e : params)
			nodeCount += e.nodeCount();
		return 1 + target.nodeCount() + nodeCount;
	}

	public int computeMaximumCallArguments() {
		return Math.max(target.computeMaximumCallArguments(), params.size());
	}

	public int computeMaximumCallResults() {
		return Math.max(target.computeMaximumCallResults(), returnSize);
	}
	
	public boolean equalsTree(Object object) {
		if (!(object instanceof IRCall))
			return false;
		IRCall other = (IRCall) object;
		if (!other.target.equalsTree(target))
			return false;
		if (other.params.size() != params.size())
			return false;
		for (int i=0; i<params.size(); ++i)
			if (!other.params.get(i).equalsTree(params.get(i)))
				return false;
		return true;
	}
	
	public int treeHashCode() {
		int code=17;
		code= 37*code + target.hashCode();
		for(IRExpr expr : params)
			code= 37*code + expr.treeHashCode();
		return code;
	}
}