package xic.ir;
import java.util.List;

import xic.ast.ArrayType;
import xic.ast.ClassDecl;
import xic.ast.ClassType;
import xic.ast.FuncDecl;
import xic.ast.PrimitiveType;
import xic.ast.TupleDecl;
import xic.ast.TupleType;
import xic.ast.TypeIdentifier;
import xic.ast.TypeNode;



public class NameMangler {
	/**
	 * Encoding a function declaration into a symbol name for 
	 * better separate compilation support.
	 * The encoding of a procedure or function name is the sequence of the following:
	 * 		The string _I
	 * 		The name of the function, encoded as described below
	 * 			The underscore character,
	 * 			The encoding of the return type, or p if this is a procedure
	 *			Encodings of types of each of the arguments, if there are any
	 * @param cd 
	 */
	public static String getEncoding(FuncDecl decl, ClassDecl cd) {
		String result = "_I";
		if (cd != null)
			result = result + "_" + cd.name() + "_";
		result += translateName(decl.name());
		result += "_";
		if (decl.isProcedure())
			result += "p";
		else
			result += typeEncoding(decl.retType());
		List<TupleDecl> params = decl.param().subTupleDecls();
		for (int i=0; i<params.size(); i++) {
			TupleDecl par = params.get(i);
			result += typeEncoding(par.getType());
		}
		return result;
	}

	public static String translateGlobalVarName(String name, TypeNode type) {
		return "_I_g_" + translateName(name) + "_" + typeEncoding(type);
	}
	
	public static String translateGlobalVarInit(String name, TypeNode type) {
		return "_I_init_" + translateName(name) + "_" + typeEncoding(type);		
	}
	
	/**
	 * encode function names by replacing a single underscore character with two of them
	 * (and 2 will get replaced with 4, and so on).
	 */
	public static String translateName(String name) {
		if (name == null)
			return null;
		return name.replaceAll("_", "__");
	}
	
	private static String typeEncoding (TypeNode type) {
		if (type instanceof PrimitiveType
				&& ((PrimitiveType) type).type() == PrimitiveType.Type.INTEGER) {
			return "i";
		} else if (type instanceof PrimitiveType
				&& ((PrimitiveType) type).type() == PrimitiveType.Type.BOOLEAN) {
			return "b";
		} else if (type instanceof ArrayType) {
			ArrayType aType = (ArrayType) type;
			TypeNode subType = aType.elementType();
			return "a"+typeEncoding(subType);
		} else if (type instanceof ClassType) {
			String orig = ((ClassType) type).name();
			String name = translateName(orig);
			return "o"+orig.length() + name;
		} else if (type instanceof TypeIdentifier) {
			String orig = ((TypeIdentifier) type).name();
			String name = translateName(orig);
			return "o"+orig.length() + name;
		} else if (type instanceof TupleType) {
			TupleType tType = (TupleType) type;
			List<TypeNode> eleTypes = tType.elementTypes();
			String tEncoding = "t"+eleTypes.size();
			if (eleTypes.size() == 1)
				return typeEncoding(eleTypes.get(0));

			for (int i=0; i<eleTypes.size(); i++) {
				TypeNode eType = eleTypes.get(i);
				tEncoding += typeEncoding(eType);
			}
			return tEncoding;
		}
		else {
			return "$";
		}
	}
}
