package xic.visitors;

import xic.InternalCompilerException;
import xic.ast.ASTNode;

/**
 * Abstract node vistor based on the visitor pattern in Polyglot. Allows
 * AST->AST transformations using lazy referencing of unmodified nodes.
 */
public abstract class NodeVisitor {
	/**
	 * Allows to hijack the traversal of a subtree. This function is called by
	 * <code>NodeVisitor.visit()</code> upon entering node <code>n</code> --
	 * when it returns a non-null node <code>n0</code>, the traversal is stopped
	 * and the resulting AST has <code>n0</code> in place of <code>n</code>.
	 * 
	 * By default, overriding is inactive.
	 */
	protected ASTNode override(ASTNode parent, ASTNode n) {
		return null;
	}

	/**
	 * Called upon entering <code>n</code> during the AST traversal. This allows
	 * to perform certain actions, including returning a new Node visitor to be
	 * used in the subtree.
	 */
	protected NodeVisitor enter(ASTNode parent, ASTNode n) {
		return this;
	}

	/**
	 * Called after finishing traversal of the subtree rooted at <code>n</code>.
	 * When <code>enter()</code> creates a new visitor to be used on the
	 * subtree, the old visitor still receives the call to <code>leave()</code>
	 * -- that is, <code>leave()</code> always executed the same number of times
	 * as <code>enter()</code>. This node provides the final opportunity of placing
	 * an updated node in the output AST.
	 * 
	 * @param parent
	 *            The parent AST node of <code>n</code> or <code>null</code>
	 *            when it is the root.
	 * @param n
	 *            The original node in the input AST
	 * @param nNew
	 *            The node returned by <code>visitChildren</code>
	 * @param vNew
	 *            The new node visitor created by <code>enter()</code>, or <code>this</code>.
	 */

	protected ASTNode leave(ASTNode parent, ASTNode n, ASTNode nNew,
			NodeVisitor vNew) {
		return nNew;
	}

	/**
	 * Recursively traverse the AST subtree rooted at <code>n</code>
	 */
	public ASTNode visit(ASTNode parent, ASTNode n) {
		if (n == null)
			return null;

		/* Allow the visitor implementation to hijack traversal of n */
		ASTNode overrideValue = override(parent, n);
		if (overrideValue != null)
			return overrideValue;

		NodeVisitor v0 = enter(parent, n);
		if (v0 == null)
			throw new InternalCompilerException(
					"NodeVisitor.enter() returned null!");

		ASTNode nNew = n.visitChildren(v0);
		if (nNew == null)
			throw new InternalCompilerException(
					"NodeVisitor.visitChildren() returned null!");

		nNew = leave(parent, n, nNew, v0);
		if (nNew == null)
			throw new InternalCompilerException(
					"NodeVisitor.leave() returned null!");

		return nNew;
	}

	/**
	 * Recursively traverse the AST subtree rooted at <code>n</code>
	 */
	public ASTNode visit(ASTNode node) {
		return visit(null, node);
	}
}
