package xic.visitors;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import xic.Configuration;
import xic.ErrorSink;
import xic.InternalCompilerException;
import xic.PositionImpl;
import xic.Trackable;
import xic.ast.ASTNode;
import xic.ast.ArrayType;
import xic.ast.Assignment;
import xic.ast.BinaryExpr;
import xic.ast.Block;
import xic.ast.BoolLiteral;
import xic.ast.Break;
import xic.ast.Call;
import xic.ast.CharLiteral;
import xic.ast.ClassDecl;
import xic.ast.ClassType;
import xic.ast.CompUnit;
import xic.ast.DataSource;
import xic.ast.Decl;
import xic.ast.DeclBlock;
import xic.ast.DotExpr;
import xic.ast.ErrorStmt;
import xic.ast.Expr;
import xic.ast.ExprStmt;
import xic.ast.FuncDecl;
import xic.ast.FuncType;
import xic.ast.Identifier;
import xic.ast.If;
import xic.ast.IntLiteral;
import xic.ast.InvalidType;
import xic.ast.Length;
import xic.ast.Literal;
import xic.ast.LocalVarDecl;
import xic.ast.New;
import xic.ast.Null;
import xic.ast.PrimitiveType;
import xic.ast.Return;
import xic.ast.Stmt;
import xic.ast.StringLiteral;
import xic.ast.This;
import xic.ast.TupleDecl;
import xic.ast.TupleExpr;
import xic.ast.TupleType;
import xic.ast.TypeIdentifier;
import xic.ast.TypeNode;
import xic.ast.Typed;
import xic.ast.UnaryExpr;
import xic.ast.UnitType;
import xic.ast.UseBlock;
import xic.ast.UseStmt;
import xic.ast.VarDecl;
import xic.ast.While;
import xic.ast.Call.CallType;
import xic.ir.NameGenerator;
import xic.parser.IotaParser;
import xic.parser.ParserImpl.Language;
import xic.parser.TokenType;
import xic.parser.XiParser;
import edu.cornell.cs.cs4120.xi.CompilationException;
import edu.cornell.cs.cs4120.xi.Position;

/**
 * Type checker -- implemented using the Polyglot visitor pattern
 * (see <code>NodeVisitor</code> for more detail). When encountering
 * type errors, a special 'invalid' type is propagated, which prevents
 * further errors from being generated in the affected expression.
 */
public class TypeChecker extends ErrorHandlingVisitor {
	private TypeChecker parent;
	private Map<String, TypeNode>   symbolTable;
	private Map<String, DataSource> symbolDataSource;
	private boolean inWhile;
	private TypeNode returnType;
	private boolean firstPass;
	private NameGenerator nameGen;
	private int paramOffset = -1; // Records the offset of the last function parameter in words
	private List<Decl> imports;
	private List<String> searchPaths;
	private ClassType enclosingClass;
	private boolean inUninitializedArrayType;
	// typerChecker for all the used files. Either no conflicts between this implementation and interface
	// or this implementation totally matches the interface.
	private List<CompUnit> usedFiles;
	private Language language; 

	/**
	 * Create a new type checking note visitor. Expects
	 * an error sink in which to put error messages encountered
	 * during the verification. 
	 */
	public TypeChecker(Language language, ErrorSink errorSink, List<String> searchPaths) {
		super(errorSink);
		this.language = language;
		this.searchPaths = searchPaths;
		inWhile = false;
		returnType = null;
		symbolTable = new HashMap<String, TypeNode>();
		symbolDataSource = new HashMap<String, DataSource>();
		firstPass = true;
		nameGen = new NameGenerator();
		paramOffset = -1;
		imports = new ArrayList<Decl>();
		inUninitializedArrayType = false;
		enclosingClass = null;
		usedFiles = new ArrayList<CompUnit>();
	}

	/**
	 * Create a nested type checker with its own local
	 * symbol table
	 */
	protected TypeChecker(TypeChecker parent, boolean inWhile, TypeNode returnType, ClassType enclosingClass) {
		super(parent.getErrorSink(), parent.getIgnoredTypes());
		this.parent = parent;
		this.inWhile = inWhile;
		this.returnType = returnType;
		this.nameGen = parent.nameGen;
		this.enclosingClass = enclosingClass;
		inUninitializedArrayType = parent.inUninitializedArrayType;
		symbolTable      = new HashMap<String, TypeNode>();
		symbolDataSource = new HashMap<String, DataSource>();
		imports = new ArrayList<Decl>();
		firstPass = parent.firstPass;
		paramOffset = -1;
	}

	public List<Decl> getImports() {
		return imports;
	}

	public NameGenerator getNameGenerator() {
		return nameGen;
	}

	/**
	 * Recursively walk up a chain of type checker instances
	 * in search for a symbol with the given name.
	 */
	protected TypeNode lookup(String name) {
		if (symbolTable.containsKey(name))
			return symbolTable.get(name);
		else if (parent != null)
			return parent.lookup(name);
		else
			return null;
	}

	/**
	 * Recursively walk up a chain of type checker instances
	 * in search for a data source with the given name.
	 */
	protected DataSource lookupDataSource(String name) {
		if (symbolTable.containsKey(name))
			return symbolDataSource.get(name);
		else if (parent != null)
			return parent.lookupDataSource(name);
		else
			return null;
	}	
	
	/**
	 * Runs the two type-checking passes in sequence
	 */
	public CompUnit run(CompUnit compUnit) {
		addIgnore(Stmt.class);
		firstPass = true;
		CompUnit result = (CompUnit) visit(compUnit);
		firstPass = false;
		removeIgnore(Stmt.class);
		
		Map<String, TypeNode>   importSymbolTable = new HashMap<String, TypeNode>();
		List<Decl> importDecls = new ArrayList<Decl>();
		
		// type check used files
		for (CompUnit unit:usedFiles) {
			checkUsedFile(unit, errorSink, importSymbolTable, importDecls);
		}
		
		if (errorSink.hasErrors())
			return result;
		else {
			symbolTable.putAll(importSymbolTable);
			imports.addAll(importDecls);
		}

		for (TypeNode n: new HashSet<TypeNode>(symbolTable.values())) {
			if (!(n instanceof ClassType))
				continue;
			ClassType ct = (ClassType) n;
			if (ct.parentName() != null) {
				TypeNode parentType = symbolTable.get(ct.parentName());
				if (parentType == null) {
					errorSink.addError(new CompilationException("Class '"
							+ ct.name() + "' extends from unknown class '"
							+ ct.parentName() + "'", ct.position()));
				} else if (!(parentType instanceof ClassType)) {
					errorSink.addError(new CompilationException("Class '"
							+ ct.name() + "' extends from a non-class entity '"
							+ ct.parentName() + "'", ct.position()));
				} else if (parentType.isSubtypeOf(ct)) {
					errorSink.addError(new CompilationException("Classes '"
							+ ct.name() + "' and '"
							+ ct.parentName() + "' form a cycle!", ct.position()));
			    } else {
					ct = ct.setParent((ClassType) parentType);
					symbolTable.put(ct.name(), ct);
				}
		

			}
		}
		
		if (!errorSink.hasErrors()) {
			for (String name : new HashSet<String>(symbolTable.keySet())) {
				TypeNode t = finalizeType(symbolTable.get(name));
				if (!(t instanceof ClassType) && !(t instanceof FuncType)
						&& !symbolDataSource.containsKey(name)) {
					DataSource ds = new DataSource();
					/* This is an imported global - associate it with a data source */
					for (int i=0; i<t.size(); ++i)
						ds.addGlobalName(name, t,  i);
					symbolDataSource.put(name, ds);
				}
				symbolTable.put(name, t);
			}
	
			addIgnore(UseBlock.class);
			if (!errorSink.hasErrors())
				result = (CompUnit) visit(compUnit);
			removeIgnore(UseBlock.class);
		}
		return result;
	}
	
	private void checkUsedFile(CompUnit compUnit, ErrorSink sink, Map<String, TypeNode> importSymbolTable,
						List<Decl> importDecls) {
		ErrorSink checkSink = new ErrorSink();
		TypeChecker subVisitor = new TypeChecker(language, checkSink, searchPaths);
		
		boolean needAddSymbolTable = true;
		compUnit = (CompUnit) subVisitor.visit(compUnit);

		for (String key : subVisitor.symbolTable.keySet()) {
			ErrorSink fieldSink = new ErrorSink();
			if (lookup(key) != null) {
				needAddSymbolTable = false;
				// need to check if this implementation matches interface
				checkInterface(subVisitor.symbolTable, symbolTable, fieldSink,
						compUnit.name(), lookup(key).position());
				if (fieldSink.hasErrors()) {
					checkSink.addErrors(fieldSink);
					break;
				}
			}
		}

		if (needAddSymbolTable) {
			for (Decl decl : compUnit.globalStmtBlock().decls()) {
				importDecls.add(decl);
			}
			for (String symbol:subVisitor.symbolTable.keySet()) {
				if (importSymbolTable.containsKey(symbol)) {
					sink.addError( new CompilationException("Duplicated symbol <" + symbol + "> in import files"
						, importSymbolTable.get(symbol).position()));
					return;
				}
			}
			importSymbolTable.putAll(subVisitor.symbolTable);
		}
		if (checkSink.hasErrors())
			sink.addErrors(checkSink);
	}
	
	private void checkInterface (Map<String, TypeNode> intf, Map<String, TypeNode> impl, ErrorSink sink, String name, Position pos) {
		if (intf.size() != impl.size())
			sink.addError(new CompilationException("Implementation does not match interface " 
					+ name+": Symbol table size mismatch!", pos ));
		for (String key:intf.keySet()) {
			TypeNode n1 = intf.get(key);
			if (!impl.containsKey(key))
				sink.addError(new CompilationException("Implementation does not match interface " 
						+ name+": Implemetation missed "+key, pos ));
			else {
				TypeNode n2 = impl.get(key);
				n1.match(n2, sink);
			}
		}
	}

	private TypeNode finalizeType(TypeNode n) {
		if (n instanceof TypeIdentifier) {
			return (TypeNode) handleTypeIdentifier((TypeIdentifier) n);
		} if (n instanceof ArrayType) {
			ArrayType t = (ArrayType) n;
			ArrayType newT = new ArrayType(finalizeType(t.elementType()));
			return newT;
		} if (n instanceof TupleType) {
			TupleType t = (TupleType) n;
			TupleType newT = new TupleType();
			for (TypeNode st : t.elementTypes())
				newT.appendType(finalizeType(st));
			return newT;
		} if (n instanceof ClassType) {
			return finalizeClassType((ClassType) n);
		} if (n instanceof FuncType) {
			FuncType f = (FuncType) n;
			f = f.setRetType(finalizeType(f.retType()));
			f = f.setParamType(finalizeType(f.paramType()));
			return f;
		} else {
			return n;
		}
	}

	private ClassType finalizeClassType(ClassType ct) {
		/* Build dispatch tables and check for consistency */
		ct = (ClassType) symbolTable.get(ct.name());
		if (ct.isFinalized())
			return ct;

		ClassType parent = ct.parent();
		if (parent != null) {
			finalizeClassType(parent);
			parent = (ClassType) symbolTable.get(parent.name());
			ct.updateParent(parent);
		}
		
		Map<String, Integer> dispatchTable = new HashMap<String, Integer>();
		Map<String, List<Integer>> fieldTable = new HashMap<String, List<Integer>>();
		Map<String, TypeNode> symbols = new HashMap<String, TypeNode>();

		if (parent != null)
			dispatchTable.putAll(parent.getDispatchTable());

		int fieldIdx = 0;
		for (String name : ct.getOrder()) {
			TypeNode type = ct.getSymbolTable().get(name);
			symbols.put(name, finalizeType(type));
		}

		for (String name : ct.getOrder()) {
			TypeNode type = symbols.get(name);
			if (type instanceof FuncType) {
				if (!dispatchTable.containsKey(name)) {
					dispatchTable.put(name, dispatchTable.size()+1+ct.inheritanceDepth());
				} else {
					if (!parent.lookupMethod(name).isIdenticalType(symbols.get(name))) {
						errorSink.addError(new CompilationException(
								"Tried to override '" + name + "' with a different type " +
								" ('"+type + "' vs '"+parent.lookupMethod(name)+"')", type.position()));
					}
				}
			} else {
				if (dispatchTable.containsKey(name)) {
					errorSink.addError(new CompilationException(
						"Type conflict between '" + name + "' and a method" +
						" defined in a parent class.", type.position()));
				}
				List<Integer> indices = new ArrayList<Integer>();
				for (int i=0; i<type.size(); ++i)
					indices.add(++fieldIdx);
				fieldTable.put(name, indices);
			}
		}

		ct.getDispatchTable().clear();
		ct.getDispatchTable().putAll(dispatchTable);
		ct.getFieldTable().clear();
		ct.getFieldTable().putAll(fieldTable);
		ct.getSymbolTable().clear();
		ct.getSymbolTable().putAll(symbols);
		ct.setFinalized(true);
		return ct;
	}

	protected ASTNode override(ASTNode parent, ASTNode n) {
		/* Do not enter length expressions during the first pass */
		if (firstPass && parent instanceof ArrayType && !(n instanceof TypeNode))
			return n;
		return super.override(parent, n);
	}
	
	/**
	 * Upon entering a new block, create a nested node visitor
	 * with its own local symbol table
	 */
	protected NodeVisitor enterNode(ASTNode parent, ASTNode n)
			throws CompilationException {
		
		if (n instanceof Block) {
			return new TypeChecker(this, inWhile, returnType, enclosingClass);
		} else if (n instanceof While) {
			return new TypeChecker(this, true, returnType, enclosingClass);
		} else if (n instanceof ArrayType) {
			TypeChecker tc = new TypeChecker(this, true, returnType, enclosingClass);
			tc.inUninitializedArrayType = !((ArrayType) n).hasLengthExpr();
			return tc;
		} else if (n instanceof FuncDecl) {
			TypeNode retType = ((FuncDecl) n).retType();
			retType = (TypeNode) visit(retType);
			TypeChecker checker = new TypeChecker(this, inWhile,
					retType == null ? UnitType.UNIT_TYPE : retType, enclosingClass);
			if (retType != null && retType.size() > 1)
				checker.paramOffset = Configuration.WORD_SIZE;
			else
				checker.paramOffset = 0;
			if (enclosingClass != null)
				checker.paramOffset += Configuration.WORD_SIZE;
			return checker;
		} else if (n instanceof ClassDecl) {
			ClassType ct = ((ClassType) lookup(((ClassDecl) n).name()));
			if (ct == null)
				ct = (ClassType) ((ClassDecl) n).getType();
			TypeChecker checker = new TypeChecker(this, inWhile, returnType, ct);
			if (!firstPass) {
				checker.symbolTable = ct.getSymbolTable();
				for (String name : ct.getFieldTable().keySet()) {
					DataSource ds = new DataSource();
					ds.addAttribute(name, ct.getFieldTable().get(name));
					checker.symbolDataSource.put(name, ds);
				}
			}
			return checker;
		}

		return this;
	}

	/**
	 * Call the implementation leaveNode() method and possibly
	 * collect errors -- also sets the node type to 
	 * <code>InvalidType</code> in case of errors
	 */
	protected ASTNode leave(ASTNode parent, ASTNode n, ASTNode nNew,
			NodeVisitor vNew) {
		try {
			return leaveNode(parent, n, nNew, vNew);
		} catch (CompilationException e) {
			errorSink.addError(e);
			if (nNew instanceof Typed)
				return (ASTNode) (((Typed) nNew).setType(InvalidType.INVALID_TYPE));
			return nNew;
		}
	}

	/**
	 * Identifies the node type and delegates work to specialized handler methods
	 */
	protected ASTNode leaveNode(ASTNode parent, ASTNode n, 
			ASTNode nNew, NodeVisitor vNew) throws CompilationException {

		/* Expressions */
		if (nNew instanceof Expr) {
			if (nNew instanceof Literal)
				return handleLiteral((Literal) nNew);
			else if (nNew instanceof TupleExpr)
				return handleTupleExpr((TupleExpr) nNew);
			else if (nNew instanceof BinaryExpr) 
				return handleBinaryExpr((BinaryExpr) nNew);			
			else if (nNew instanceof Identifier)
				return handleIdentifier((Identifier) nNew);
			else if (nNew instanceof Call)
				return handleCall((Call) nNew);
			else if (nNew instanceof UnaryExpr)
				return handleUnaryExpr ((UnaryExpr) nNew);
			else if (nNew instanceof Length)
				return handleLength((Length) nNew);
			else if (nNew instanceof New)
				return handleNew((New) nNew);
			else if (nNew instanceof This)
				return handleThis ((This) nNew);
			else if (nNew instanceof DotExpr)
				return handleDotExpr((DotExpr) nNew);
			else if (nNew instanceof Null)
				return nNew;
			else
				throw new InternalCompilerException("Encountered an unknown expression node - handle" 
						+ nNew.getClass().getSimpleName() + "() still needs to be implemented!");
		}

		/* Statements */
		if (nNew instanceof Stmt) {
			if (nNew instanceof LocalVarDecl)
				return handleLocalVarDecl((LocalVarDecl) nNew);
			else if (nNew instanceof If)
				return handleIf((If) nNew);
			else if (nNew instanceof Block)
				return ((TypeChecker) vNew).handleBlock((Block) nNew);
			else if (nNew instanceof While) 
				return handleWhile((While) nNew);
			else if (nNew instanceof Break) 
				return handleBreak((Break) nNew);
			else if (nNew instanceof Return) 
				return handleReturn((Return) nNew);
			else if (nNew instanceof Assignment) 
				return handleAssignment((Assignment) nNew);
			else if (nNew instanceof ExprStmt)
				return handleExprStmt((ExprStmt) nNew);
			else if (nNew instanceof ErrorStmt)
				throw new InternalCompilerException("Attempted to type-check an AST with syntax errors!");
			else
				throw new InternalCompilerException("Encountered an unknown statement node - handle" 
						+ nNew.getClass().getSimpleName() + "() still needs to be implemented!");
		}

		/* Other AST nodes */
		if (nNew instanceof TupleDecl)
			return handleTupleDecl((TupleDecl) nNew);
		if (nNew instanceof ArrayType)
			return handleArrayType((ArrayType) nNew);
		else if (nNew instanceof FuncDecl)
			return handleFuncDecl((FuncDecl) nNew, (TypeChecker) vNew);
		else if (nNew instanceof UseStmt)
			return handleUseStmt((UseStmt) nNew);
		else if (nNew instanceof DeclBlock)
			return handleGlobalDeclBlock((DeclBlock) nNew);
		else if (nNew instanceof TupleType)
			return handleTupleType((TupleType) nNew);
		else if (nNew instanceof ClassDecl)
			return handleClassDecl((ClassDecl) nNew, (TypeChecker) vNew);
		else if (nNew instanceof TypeIdentifier)
			return handleTypeIdentifier((TypeIdentifier) nNew);
		else if (nNew instanceof VarDecl)
			return handleVarDecl((VarDecl) nNew);

		return nNew;
	}

	// ***************************************************************
	// ***                       Expressions                       ***
	// ***************************************************************
	
	/**
	 * Local variable, parameter and function references
	 */
	protected ASTNode handleIdentifier(Identifier i) {
		TypeNode typeNode = lookup(i.name());
		if (typeNode == null) {
			if (enclosingClass != null && enclosingClass.lookupMethod(i.name()) != null)
				return (ASTNode) i.setType(enclosingClass.lookupMethod(i.name()));

			throw new CompilationException("Unbound identifier '"
					+ i.name() + "'", i.position());
		}

		DataSource dataSource = lookupDataSource(i.name());
		return (ASTNode) i.setDataSource(dataSource).setType(typeNode);
	}


	private ASTNode handleDotExpr(DotExpr nNew) {
		if (nNew.target().getType() instanceof InvalidType)
			return (ASTNode) nNew.setType(InvalidType.INVALID_TYPE);
		if (nNew.target().getType() instanceof ClassType) {
			ClassType ct = (ClassType) nNew.target().getType();
			FuncType tMethod = ct.lookupMethod(nNew.member());
			TypeNode tAttribute = ct.lookupAttribute(nNew.member());
			
			if (tMethod == null && tAttribute == null)
				throw new CompilationException("Class member '" 
						+nNew.member()+"' is unbound", nNew.position());

			if (tMethod != null)
				return (ASTNode) nNew.setType(tMethod);
			else
				return (ASTNode) nNew.setType(tAttribute);
		} else {
			throw new CompilationException("Expected object in dot expression, got '" 
									+nNew.target().getType()+"'", nNew.position());
		}
	}

	private ASTNode handleThis(This nNew) {
		if (enclosingClass != null) {
			return (ASTNode) nNew.setType(enclosingClass);
			
		} else 
			throw new CompilationException("'this' keyword is only valid within classes", 
												nNew.position());
	}

	private ASTNode handleNew(New n) {
		TypeNode t = lookup(n.name());

		if (t == null)
			throw new CompilationException(
					"Unable to find type identifier '" + n.name() + "'", n.position());
		else if (!(t instanceof ClassType))
			throw new CompilationException(
					"Type '" + n.name() + "' is not a class", n.position());
		else if (!((ClassType) t).name().equals(n.name())) {
			throw new CompilationException(
					"The type identifier '" + n.name() + "' does not refer to a proper type", n.position());
		}

		return (ASTNode) n.setType(t);
	}

	/**
	 * BinaryExpr node handler:
	 * Identify the particular type and delegate type-checking to specialized handlers.
	 */
	private ASTNode handleBinaryExpr(BinaryExpr b) {
		BinaryExpr.OpType op = b.opType();
		TypeNode l = b.left().getType(), r = b.right().getType();

		if (op == BinaryExpr.OpType.PLUS && 
				((l instanceof ArrayType || l instanceof TupleType) ||
				(r instanceof ArrayType || r instanceof TupleType)))
			return handleArrayAddition(b, l, r);
		
		if (op == BinaryExpr.OpType.PLUS || op == BinaryExpr.OpType.MINUS
				|| op == BinaryExpr.OpType.TIMES || op == BinaryExpr.OpType.DIVIDE
				|| op == BinaryExpr.OpType.MODULO)
			return handleArithmeticBinaryExpr(b, l, r);
		
		else if (op == BinaryExpr.OpType.LT || op == BinaryExpr.OpType.LEQ
				|| op == BinaryExpr.OpType.GT || op == BinaryExpr.OpType.GEQ)
			return handleCompareIntBinaryExpr(b, l, r);
		
		else if (op == BinaryExpr.OpType.AND || op == BinaryExpr.OpType.OR)
			return handleBoolBinaryExpr(b, l, r);
		
		else if (op == BinaryExpr.OpType.EQUAL || op == BinaryExpr.OpType.NOT_EQUAL)
				return handleComparisonBinaryExpr(b, l, r);
		else
			throw new CompilationException("Illegal operator for types '"
				+ l + "', '" + r + "'", b.position());
	}

	protected ASTNode handleArithmeticBinaryExpr(BinaryExpr b, TypeNode l, TypeNode r) {
		if (l instanceof InvalidType || r instanceof InvalidType)
			return (ASTNode) b.setType(InvalidType.INVALID_TYPE);

		else if (l.isSubtypeOf(PrimitiveType.INTEGER_TYPE)
				&& r.isSubtypeOf(PrimitiveType.INTEGER_TYPE))
			return (ASTNode) b.setType(PrimitiveType.INTEGER_TYPE);

		throw new CompilationException("Illegal operator for types '"
				+ l + "', '" + r + "'", b.position());
	}

	protected ASTNode handleArrayAddition(BinaryExpr b, TypeNode l, TypeNode r) {
		if (l instanceof InvalidType || r instanceof InvalidType)
			return (ASTNode) b.setType(InvalidType.INVALID_TYPE);
		
		if (l instanceof ArrayType)
			l = ((ArrayType) l).elementType();
		else if (l instanceof TupleType)
			l = findArrayIndexingType((TupleType) l);

		if (r instanceof ArrayType)
			r = ((ArrayType) r).elementType();
		else if (r instanceof TupleType)
			r = findArrayIndexingType((TupleType) r);

		if (l != null && r != null && (l.isIdenticalType(r)))
			return (ASTNode) b.setType(new ArrayType(l));
		else
			throw new CompilationException("Array addition: illegal operator for types '"
					+ l + "', '" + r + "'", b.position());
	}

	
	protected ASTNode handleBoolBinaryExpr(BinaryExpr b, TypeNode l, TypeNode r) {
		if (l instanceof InvalidType || r instanceof InvalidType)
				return (ASTNode) b.setType(InvalidType.INVALID_TYPE);

		else if (l.isSubtypeOf(PrimitiveType.BOOLEAN_TYPE) && 
					r.isSubtypeOf(PrimitiveType.BOOLEAN_TYPE))
			return (ASTNode) b.setType(PrimitiveType.BOOLEAN_TYPE);

		 throw new CompilationException("Illegal operator for types '"
				+ l + "', '" + r + "'", b.position());
	}

	protected ASTNode handleComparisonBinaryExpr(BinaryExpr b, TypeNode l, TypeNode r) {		 
		if (l instanceof InvalidType || r instanceof InvalidType)	
			return (ASTNode) b.setType(InvalidType.INVALID_TYPE);

		else if (l.isSubtypeOf(r) || r.isSubtypeOf(l)) {
			return (ASTNode) b.setType(PrimitiveType.BOOLEAN_TYPE);
		}

		throw new CompilationException("Illegal operator for types '"
					+ l + "', '" + r + "'", b.position());
	}

	protected ASTNode handleCompareIntBinaryExpr(BinaryExpr b, TypeNode l, TypeNode r) {
		BinaryExpr.OpType op = b.opType();

		if (l instanceof InvalidType || r instanceof InvalidType)
			return (ASTNode) b.setType(InvalidType.INVALID_TYPE);
		else if (l.isSubtypeOf(PrimitiveType.INTEGER_TYPE) && 
				r.isSubtypeOf(PrimitiveType.INTEGER_TYPE)
				&& (op == BinaryExpr.OpType.EQUAL || op == BinaryExpr.OpType.NOT_EQUAL
					|| b.opType() == BinaryExpr.OpType.LT || op == BinaryExpr.OpType.LEQ
					|| op == BinaryExpr.OpType.GT || op == BinaryExpr.OpType.GEQ))
			return (ASTNode) b.setType(PrimitiveType.BOOLEAN_TYPE);

		throw new CompilationException("Illegal operator for types '"
				+ l + "', '" + r + "'", b.position());
	}

	protected ASTNode handleUnaryExpr(UnaryExpr u) {
		UnaryExpr.OpType op = u.opType();
		TypeNode eType = u.expr().getType();

		if (eType instanceof InvalidType)
			return (ASTNode) u.setType(InvalidType.INVALID_TYPE);
		if (op == UnaryExpr.OpType.MINUS) {
			if (!eType.isSubtypeOf(PrimitiveType.INTEGER_TYPE))
				throw new CompilationException(
						"Incompatible type in an unary expression  -- expected 'int', got '"
								+ eType + "'", u.position());
			return (ASTNode) u.setType(PrimitiveType.INTEGER_TYPE);
		} else if (op == UnaryExpr.OpType.NOT) {
			if (!eType.isSubtypeOf(PrimitiveType.BOOLEAN_TYPE))
				throw new CompilationException(
						"Incompatible type in an unary expression  -- expected 'bool', got '"
								+ eType + "'", u.position());
			return (ASTNode) u.setType(PrimitiveType.BOOLEAN_TYPE);
		} else
			throw new InternalCompilerException(
					"Encountered an unknown unary operator");
	}

	/**
	 * Literals
	 */
	protected ASTNode handleLiteral(Literal l) {
		if (l instanceof IntLiteral)
			return (ASTNode) l.setType(PrimitiveType.INTEGER_TYPE);
		else if (l instanceof BoolLiteral)
			return (ASTNode) l.setType(PrimitiveType.BOOLEAN_TYPE);
		else if (l instanceof StringLiteral)
			return (ASTNode) l.setType(ArrayType.INT_ARRAY_TYPE);
		else if (l instanceof CharLiteral)
			return (ASTNode) l.setType(PrimitiveType.INTEGER_TYPE);
		else
			throw new InternalCompilerException("Encountered a literal of unknown type");
	}

	/**
	 * Tuple expressions
	 */
	protected ASTNode handleTupleExpr(TupleExpr t) {
		TupleType type = new TupleType();

		DataSource ds = new DataSource();
		if (t.subExprs().size() == 1) {
			Expr subExpr = t.subExprs().get(0);
			if (subExpr.getDataSource() == null || subExpr.getDataSource().size() == 0) {
				ds.addTempName(nameGen.nextTemp());
			} else {
				ds.add(subExpr.getDataSource());
			}
			return (ASTNode) t.setDataSource(ds).setType(subExpr.getType());
		}

		for (Expr e: t.subExprs()) {
			TypeNode exprType = e.getType();

			if (e.getType() instanceof InvalidType)
				return (ASTNode) t.setType(InvalidType.INVALID_TYPE);

			if (e instanceof TupleExpr) {
				if (e.getDataSource().size() != exprType.size())
					throw new InternalCompilerException(
							"Size mismatch in tuple expression! (" + e.getDataSource().size()
									+ " vs " + exprType.size() + ")");
				ds.add(e.getDataSource());
			} else {
				/* Allocate new temps */
				for (int i=0; i<exprType.size(); ++i)
					ds.addTempName(nameGen.nextTemp());
			}
	
			type.appendType(exprType);
		}

		return (ASTNode) t.setDataSource(ds).setType(type);
	}

	/**
	 * Calls, array indexing, tuple indexing
	 */
	protected ASTNode handleCall(Call c) {
		TypeNode targetType = c.target().getType();
		DataSource ds = new DataSource();

		if (targetType instanceof InvalidType || c.param().getType() instanceof InvalidType)
			return (ASTNode) c.setType(InvalidType.INVALID_TYPE);

		if (targetType instanceof FuncType) {
			FuncType fType = (FuncType) targetType;
			
			if (c.callType() == CallType.INDEXING_OPERATION)
				throw new CompilationException(
						"Incompatible target of an array indexing operation (got a function)!", c.position());
				

			/* Rewrite if this is a call to a method */
			if (enclosingClass != null && c.target() instanceof Identifier && 
					enclosingClass.lookupMethod(((Identifier) c.target()).name()) != null) {
				return visit(new Call(c.position(), new DotExpr(c.position(),
						new This(c.position()), ((Identifier) c.target()).name()), c.param()));
			}
			
			if (!c.param().getType().isSubtypeOf(fType.paramType())) {
				throw new CompilationException(
						"Incompatible parameter in a function call -- expected '"
						+ fType.paramType() + "', got '"
						+ c.param().getType() + "'", c.param().position());
			}

			if (!(fType.retType() instanceof UnitType) && fType.retType().size() > 1) {
				/* Allocate new temps */
				for (int i=0; i<fType.retType().size(); ++i)
					ds.addTempName(nameGen.nextTemp());
			}

			return (ASTNode) c.setDataSource(ds).setType(fType.retType());
		} else if (targetType.isSubtypeOf(ArrayType.ANYTYPE_ARRAY_TYPE)) {
			if (c.callType() == CallType.FUNCTION_CALL)
				throw new CompilationException(
						"Incompatible target of a function call (got an array or tuple)!", c.position());

			if (targetType instanceof ArrayType) {
				ArrayType aType = (ArrayType) targetType;
				if (aType.elementType().size() > 1) {
					/* Allocate new temps */
					for (int i=0; i<aType.elementType().size(); ++i)
						ds.addTempName(nameGen.nextTemp());
				}
				if (!c.param().getType().isSubtypeOf(PrimitiveType.INTEGER_TYPE))
					throw new CompilationException(
						"Incompatible index for array access -- expected 'int', got '"
						+ c.param().getType() + "'", c.param().position());
				return (ASTNode) c.setDataSource(ds).setType(aType.elementType());
			} else if (targetType instanceof TupleType) {
				TupleType tType = (TupleType) targetType;
				TypeNode elementType = findArrayIndexingType(tType);

				if (elementType == null)
					throw new CompilationException(
						"Tuple of type "
						+ tType + " cannot be used as an array", c.target().position());

				if (elementType.size() > 1) {
					/* Allocate new temps */
					for (int i=0; i<elementType.size(); ++i)
						ds.addTempName(nameGen.nextTemp());
				}
				if (!c.param().getType().isSubtypeOf(PrimitiveType.INTEGER_TYPE))
					throw new CompilationException(
						"Incompatible index for array access -- expected 'int', got '"
						+ c.param().getType() + "'", c.param().position());
				return (ASTNode) c.setDataSource(ds).setType(elementType);
			} else {
				throw new CompilationException(
						"Encountered an indexing operation into the invalid type '"
								+ targetType + "'", c.position());
			}
		} else {
			throw new CompilationException(
					"Encountered an invalid function call/array indexing operation",
					c.position());
		}
	}

	private TypeNode findArrayIndexingType(TupleType tType) {
		for (TypeNode elType : tType.elementTypes()) {
			boolean fail = false;
			for (TypeNode elType2 : tType.elementTypes()) {
				if (!elType2.isSubtypeOf(elType)) {
					fail = true;
					break;
				}
			}
			if (!fail)
				return elType;
		}

		for (TypeNode elType : tType.elementTypes()) {
			if (!(elType instanceof TupleType))
				continue;
			TypeNode type = new ArrayType(findArrayIndexingType((TupleType) elType));
			boolean fail = false;
			for (TypeNode elType2 : tType.elementTypes()) {
				if (!elType2.isSubtypeOf(type)) {
					fail = true;
					break;
				}
			}
			if (!fail)
				return type;
		}

		return null;
	}

	/**
	 * Length expressions
	 */
	protected ASTNode handleLength(Length nNew) {
		TypeNode eType = nNew.expr().getType();
		if (eType instanceof InvalidType)
			return (ASTNode) nNew.setType(InvalidType.INVALID_TYPE);
		if (!(eType.isSubtypeOf(ArrayType.ANYTYPE_ARRAY_TYPE)))
			throw new CompilationException(
					"Incompatible type in a 'length' expression -- expected an array, got '"
							+ eType + "'", nNew.expr().position());
		return (ASTNode) nNew.setType(PrimitiveType.INTEGER_TYPE);
	}

	// ***************************************************************
	// ***                        Statements                       ***
	// ***************************************************************

	protected ASTNode handleBlock(Block bl) {
		/* Save the local symbol table */
		return bl.setSymbolTable(symbolTable);
	}

	protected ASTNode handleIf(If n) {
		// Check that the guard is of type boolean
		TypeNode guardType = n.guard().getType();
		if (guardType instanceof InvalidType)
			return n;
		if (!guardType.isSubtypeOf(PrimitiveType.BOOLEAN_TYPE))
			throw new CompilationException(
					"Incompatible expression in an 'if' statement -- expected 'bool', got '"
							+ guardType + "'", n.guard().position());
		return n;
	}

	/**
	 * Local variable declarations
	 */
	protected ASTNode handleLocalVarDecl(LocalVarDecl l) {
		if (!l.hasInit())
			return l; // No initialization -> nothing to do

		TypeNode lType = l.tupleDecl().getType();
		TypeNode rType = l.init().getType();

		if (lType instanceof InvalidType || rType instanceof InvalidType)
			return l;

		if (!rType.isSubtypeOf(lType))
			throw new CompilationException(
					"Incompatible types in initialization -- expected '"
					+ lType.toString() + "', got '"
					+ rType.toString()+ "'", l.position());
		return l;
	}

	/**
	 * While loops
	 */
	protected ASTNode handleWhile(While nNew) {
		TypeNode condType = nNew.cond().getType();
		if (condType instanceof InvalidType)
			return nNew;
		if (!(condType.isSubtypeOf(PrimitiveType.BOOLEAN_TYPE)))
			throw new CompilationException(
					"Incompatible type in 'while' statement -- expected 'bool', got '"
					+ condType + "'", nNew.cond().position());
		return nNew;
	}

	/**
	 * Break statements
	 */
	protected ASTNode handleBreak(Break nNew) {
		if (inWhile)
			return nNew;
		throw new CompilationException(
				"Encountered a break statement outside of a while loop!",
				nNew.position());
	}

	/**
	 * Return statements
	 */
	protected ASTNode handleReturn(Return nNew) {
		if (returnType  == null) // The parser should handle this case
			throw new InternalCompilerException(
					"Internal error: encountered a return statement outside of a function body!",
					nNew.position());
		
		if (nNew.hasExpr()) {
			TypeNode rType = nNew.expr().getType();
			if (rType instanceof InvalidType)
				return nNew;
			if (returnType instanceof UnitType)
				throw new CompilationException("Cannot return values from a procedure", nNew.position());
			if (!rType.isSubtypeOf(returnType))
				throw new CompilationException(
						"Incompatible type in 'return' statement -- expected '"+returnType+"', got '"
						+ rType + "'", nNew.expr().position());
			else
				return nNew;
		} else {
			if (!(returnType instanceof UnitType)) 
				throw new CompilationException(
					"Expected a return value of type '" + returnType.toString() + "'",
							nNew.position());
			else
				return nNew;
		}
	}

	/**
	 * Assignment statements 
	 */
	private ASTNode handleAssignment(Assignment nNew) {
		Expr rvalue = nNew.rvalue();
		Expr lvalue = nNew.lvalue();
		TypeNode ltype = lvalue.getType();
		TypeNode rtype = rvalue.getType();
		if (rtype instanceof InvalidType || ltype instanceof InvalidType)
			return nNew;
		if (!rtype.isSubtypeOf(ltype))
			throw new CompilationException(
					"Incompatible types in assignment: '"
							+ ltype
							+ "' vs '"
							+ rtype + "'", nNew.position());
		else 
			recursiveAssignmentCheck(lvalue);

		return nNew;
	}

	/* lvalues may only be tuples, identifiers, array accesses or tuple accesses*/
	private void recursiveAssignmentCheck(Expr lvalue) {
		if (lvalue instanceof TupleExpr) {
			TupleExpr expr = (TupleExpr) lvalue;
			for (int i=0; i<expr.subExprs().size(); i++)
				recursiveAssignmentCheck(expr.subExprs().get(i));
		} else if (lvalue instanceof DotExpr) {
			return; //verified.
		} else if (lvalue instanceof Identifier) {
			return; // verified.
		} else if (lvalue instanceof Call) {
			Call cvalue = (Call)lvalue;
			TypeNode targetType = cvalue.target().getType();
			if(targetType instanceof ArrayType)
				return; // verified.
			else if (targetType instanceof TupleType)
				if(((TupleType)targetType).indexRequiresConversion())
					throw new CompilationException(
							"Cannot assign to non-homogeneous tuple: '"
									+ targetType + "'", lvalue.position());
				else
					return; // verified
			else 
				throw new CompilationException(
						"Invalid assignment", cvalue.position());
		} else 
			throw new CompilationException("Invalid assignment ",lvalue.position());

	}

	// ***************************************************************
	// ***                   Other types of nodes                  ***
	// ***************************************************************

	protected ASTNode handleGlobalDeclBlock(DeclBlock bl) {
		/* Save the local symbol table */
		return bl.setSymbolTable(symbolTable);
	}

	protected ASTNode handleTupleDecl(TupleDecl n) throws CompilationException {
		DataSource ds = new DataSource();
		if (n.isUnderscore()) {
			return (ASTNode) n.setDataSource(ds);
		} else if (!n.isLeaf()) {
			TupleType type = new TupleType();
			if (n.subTupleDecls().size() == 1) {
				TupleDecl subTupleDecl = n.subTupleDecls().get(0);
				return (ASTNode) n.setDataSource(subTupleDecl.getDataSource())
					.setType(subTupleDecl.getType());
			}

			for (TupleDecl td : n.subTupleDecls()) {
				if (td.getType() instanceof InvalidType)
					return (ASTNode) n.setType(InvalidType.INVALID_TYPE);
				ds.add(td.getDataSource());
				type.appendType(td.getType());
			}

			return (ASTNode) n.setDataSource(ds).setType(type);
		} else {
			boolean addToSymbolTable = true;
			if (returnType == null && !firstPass)
				addToSymbolTable = false;

			if (addToSymbolTable && lookup(n.name()) != null) {
				throw new CompilationException(
					"Identifier '" + n.name() + "' is already bound.",
					n.position());
			}

			TypeNode type = n.getType();
			
			if (type instanceof InvalidType)
				throw new CompilationException("Encountered an invalid type", type.position());

			if (type instanceof TypeIdentifier) {
				symbolTable.put(n.name(), type);
				return n;
			}

			if (returnType == null) {
				/* This is a global variable / class attribute */
				if (enclosingClass == null) {
					for (int i=0; i<type.size(); ++i)
						ds.addGlobalName(n.name(), type, i);
				}
			} else if (paramOffset == -1) {
				/* This is a local variable reference */
				if (type.size() !=  1) {
					for (int i=0; i<type.size(); ++i)
						ds.addTempName("_" + n.name() + "$" + i);
				} else {
					ds.addTempName("_" + n.name());
				}
			} else {
				/* This is a function parameter */
				if (type.size() == 1) {
					ds.addParamOffset(paramOffset, "_" + n.name());
				} else {
					for (int i=0; i<type.size(); ++i)
						ds.addParamOffset(paramOffset + i*Configuration.WORD_SIZE, "_" + n.name() + "$"+i);
				}
				paramOffset += type.size()*Configuration.WORD_SIZE;
			}
			if (addToSymbolTable)
				symbolTable.put(n.name(), type);

			if (ds.size() > 0)
				symbolDataSource.put(n.name(), ds);

			return (ASTNode) n.setDataSource(ds);
		}
	}

	protected ASTNode handleClassDecl(ClassDecl c, TypeChecker subTypeChecker)
			throws CompilationException {
		if (firstPass) {
			ClassType cType = new ClassType(c.position(), c.name(), c.parentName(), 
					c.isImported(), null, subTypeChecker.symbolTable, 
					new HashMap<String, Integer>(),
					new HashMap<String, List<Integer>>(), 
					new ArrayList<String>(),false
			);
			if (lookup(c.name()) != null) {
				throw new CompilationException(
					"Identifier '" + c.name() + "' is already bound.",
					c.position());
			}
			List<String> order = new ArrayList<String>();
			for (Decl d : c.declBlock().decls())
				order.addAll(d.getNames());
			cType = cType.setOrder(order);
			symbolTable.put(c.name(), cType);
		}

		return (ASTNode) c.setSymbolTable(subTypeChecker.symbolTable).setType(
				symbolTable.get(c.name()));
	}

	protected ASTNode handleFuncDecl(FuncDecl f, TypeChecker subTypeChecker)
			throws CompilationException {
		FuncType fType = new FuncType(p(f.param(), f.retType()), 
					f.param().getType(), f.retType() != null 
				? f.retType() : UnitType.UNIT_TYPE);
		if (firstPass) {
			if (lookup(f.name()) != null) {
				throw new CompilationException(
					"Identifier '" + f.name() + "' is already bound.",
					f.position());
			}

			symbolTable.put(f.name(), fType);
		}
		return (ASTNode) f.setSymbolTable(subTypeChecker.symbolTable).setType(fType);
	}

	protected ASTNode handleUseStmt(UseStmt useStmt) throws CompilationException {
		File f = new File(useStmt.name() + ".ixi");

		for (String path : searchPaths) {
			f = new File(path + "/" + useStmt.name() + ".ixi");
			if (f.exists() && f.canRead())
				break;
			if (language == Language.IOTA_09) {
				f = new File(path + "/" + useStmt.name() + ".i9");
				if (f.exists() && f.canRead())
					break;
			}
		}
		
		if (!f.exists() || !f.canRead())
			throw new CompilationException("Unable to read interface file '" + f.getName() + "'", useStmt.position());

		try {
			ErrorSink subErrorSink = new ErrorSink();

			CompUnit compUnit = null;
			if (language == Language.XI)
				compUnit = new XiParser(f, TokenType.HEADER_IDENTIFIER).doParse(subErrorSink);
			else if (language == Language.IOTA_09)
				compUnit = new IotaParser(f, TokenType.HEADER_IDENTIFIER).doParse(subErrorSink);
			else
				throw new RuntimeException("Unknown language!");

			if (!subErrorSink.hasErrors()) {
				// record the comp unit, and postpone the check after first pass
				usedFiles.add(compUnit);
			}
			
			errorSink.addErrors(subErrorSink);
		} catch (IOException e) {
			throw new CompilationException("I/O error while parsing interface file '"
					+ f.getName() + "'", useStmt.position());
		}
		return useStmt;
	}

	private ASTNode handleExprStmt(ExprStmt n) {
		Expr expr = n.expr();
		if (expr.getType() instanceof InvalidType)
			return n;
		if (!(expr.getType() instanceof UnitType))
			throw new CompilationException("Calling a function as if it were a procedure (return value is of type '"
					+ expr.getType() + "')", expr.position());
		return n;
	}
	
	private ASTNode handleTypeIdentifier(TypeIdentifier ti) {
		if (firstPass)
			return ti;
		TypeNode t = lookup(ti.name());

		if (t == null) {
			throw new CompilationException(
					"Unable to find type identifier '" + ti.name() + "'", ti.position());
		} else if (!(t instanceof ClassType)) {
			throw new CompilationException(
					"Type '" + ti.name() + "' is not a class", ti.position());
		} else if (t instanceof TypeIdentifier) {
			throw new CompilationException(
					"Type '" + ti.name() + "' resolves to a type identifier", ti.position());
		} else if (!((ClassType) t).name().equals(ti.name())) {
			throw new CompilationException(
					"The type identifier '" + ti.name() + "' does not refer to a proper type", ti.position());
		}

		return t;
	}

	private ASTNode handleVarDecl(VarDecl vd) {
		if (vd.hasInit()) {
			if (vd.type() instanceof PrimitiveType) {
				if (!(vd.init() instanceof Literal))
					throw new CompilationException(
							"Can only perform static initializations using literals", vd.position());
				if (!vd.type().isIdenticalType(vd.init().getType()))
					throw new CompilationException(
							"Type mismatch in static initialization - got '"+vd.init().getType()+"', expected '"+vd.type()+"'", vd.position());
			} else {
				throw new CompilationException(
						"Only primitive types (int, bool) may be statically initialized", vd.position());
			}
		}
		return vd;
	}
	
	protected ASTNode handleTupleType(TupleType t) {
		if (t.elementTypes().size() == 1)
			return t.elementTypes().get(0);
		else
			return t;
	}
	
	protected ASTNode handleArrayType(ArrayType a) {
		if (a.hasLengthExpr() && !firstPass) {
			if (a.lengthExpr().getType() instanceof InvalidType)
				return InvalidType.INVALID_TYPE;

			if (paramOffset != -1)
				throw new CompilationException(
						"Array length expressions are not permitted in function parameters",
						a.lengthExpr().position());
			if(!a.lengthExpr().getType().isSubtypeOf(PrimitiveType.INTEGER_TYPE))
				throw new CompilationException(
						"Array length expression has type '" + 
						a.lengthExpr().getType() + "', expected type '" +
						PrimitiveType.INTEGER_TYPE + "'.",
						a.lengthExpr().position());

			if (returnType == null && enclosingClass == null) {
				/* This is a global variable */
				if (!(a.lengthExpr() instanceof Literal) &&
					!(a.lengthExpr() instanceof Identifier)) {
					throw new CompilationException(
							"Array length expression in global variable must be given by "+
							"an integer literal or an integer-valued global variable!", a.lengthExpr().position());
				}
			}
			
			if (inUninitializedArrayType) {
				throw new CompilationException(
						"Cannot nest initialized arrays into uninitialized ones", a.position());
			}
		}
		return a;
	}

	private Position p(Trackable l, Trackable r) {
		if (r == null && l != null)
			return l.position();
		else if (l == null && r != null)
			return r.position();
		else if (l == null && r == null)
			return null;
		else if (l.position() == null && r.position() == null)
			return null;
		else if (r.position() == null)
			return l.position();
		else if (l.position() == null)
			return r.position();

		Position lPos = l.position(), rPos = r.position();
		return new PositionImpl(lPos.unit(), lPos.lineStart(), rPos.lineEnd(),
			lPos.columnStart(), rPos.columnEnd());
	}
}
