package xic;

import static org.junit.Assert.assertTrue;

import java.util.ArrayList;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TestName;

import xic.Configuration.Architecture;
import xic.ast.ASTNode;
import xic.ast.CompUnit;
import xic.parser.IotaParser;
import xic.parser.ParserImpl.Language;
import xic.visitors.TypeChecker;
import edu.cornell.cs.cs4120.xi.CompilationException;


public class TypeCheckerTest {
	@Rule public TestName name = new TestName();

	@Before
	public void initialize() {
		Configuration.setArchitecture(Architecture.I386);
	}

	private void typeCheck(String input)
			throws CompilationException {
		ASTNode parsedAST = IotaParser.getAST(input);
		ErrorSink errorSink = new ErrorSink();

		ArrayList<String> searchPath = new ArrayList<String>();
		searchPath.add(ExecutableTest.getIotaIncludeDirectory());
		new TypeChecker(Language.IOTA_09, errorSink, searchPath).run((CompUnit) parsedAST);
		if (errorSink.hasErrors()) {
			for (CompilationException e : errorSink.getErrors()) {
				System.err.println("Problem in test case \""
						+ name.getMethodName() + "\": \"" + e.toString() + "\"");
				e.printStackTrace();
			}
		}
		assertTrue(!errorSink.hasErrors());
	}

	@Test
	public void test1() {
		// Tests a basic tuple declaration + initialization
		typeCheck(
			"test() = {\n" 
			+ "  (a:bool, b:int) = (false, 1);"
			+ "}"
		);
	}

	@Test
	public void test2() {
		// Tests a more sophisticated tuple declaration + initialization
		typeCheck(
			"test() = {\n" 
			+ "  (a:bool, (b:bool, _), c:int) = (false, (true, 3), 1);"
			+ "}"
		);
	}

	@Test
	public void test3() {
		// Tuples can be used in place of arrays.
		// (b:int) and b:int have the same type, therefore this is ok
		typeCheck(
			"test() = {\n" 
			+ "  b:int = 3;"
			+ "  a:int = length b;"
			+ "  c:int = length 1;"
			+ "}"
		);
	}

	@Test
	public void test4() {
		// Returning from a procedure
		typeCheck(
			"test() = {\n" 
			+ "  return;"
			+ "}"
		);
	}

	@Test
	public void test5() {
		// Simple function call
		typeCheck(
			"test(a:int) : int = {\n" 
			+ "  return a;"
			+ "}"
			+ ""
			+ "test2(a:int) : int = {"
			+ "  return test a;"
			+ "}"
		);
	}

	@Test
	public void test6() {
		// Mutually recursive function call
		typeCheck(
			"test(a:int) : int = {\n" 
			+ "  return test2 a;"
			+ "}"
			+ ""
			+ "test2(a:int) : int = {"
			+ "  return test a;"
			+ "}"
		);
	}

	@Test
	public void test9() {
		// break in while statement
		typeCheck(
			"test(a:int):int = {\n"
			+ "while(true) {\n"
			+ "  a=a+1;\n"
			+ "  break;"
			+ "}"
			+"}"
		);
	}
	
	@Test
	public void test10() {
		// gcd example
		typeCheck(
				"gcd(a: int, b: int): int = {\n" 
				+ "  while (a != 0) {\n"
				+ "    if (a<b) b = b-a;\n"
				+ "    else a = a - b;\n" 
				+ "  }\n"
				+ "  return b;\n" 
				+ "}\n"
				+ "\n"
		);
	}
	
	@Test
	public void test11() {
		// sort example
		typeCheck(
				"sort(a: int[]) = {"
				+	"  i:int = 0;"
				+	"  n:int = length a;"
				+	"  while (i < n) {"
				+	"    j:int = i;"
				+	"    while (j > 0) {"
				+	"      if (a(j-1) > a(j)) { // parens are optional on a(j)\n"
				+	"         swap:int = a(j);"
				+	"         a(j) = a(j-1);"
				+	"         a(j-1) = swap;"
				+	"	   }"
				+	"      j = j-1;"
				+	"    }"
				+	"    i = i+1;"
				+	"  }"
				+	"}"
		);
	}
	
	@Test
	public void test12() {
		//subtype assignments, nested size-1 tuples
		typeCheck(
				"test() : int = {\n" +
					"(x:int, y:int[], z:(bool,bool)) =" +
					" ('a', \"AAAA\", ((true),(((false)))));\n" +
					"b:(int) = y 3;\n" +
					"c:int[] = ('b', 3, y 0, x);\n" +
					"return y 2;" +
				"}"
		);
	}
	
	@Test
	public void test13() {
		//allow assignments to arrays in tuple lvalue
		typeCheck(
				"test() : int = {\n" +
					"(x:int, y:int[], z:(bool,bool)) =" +
					"(1, (2,3,4), (true,true));" +
					"(x , y 1, z) = (5, 6, (false,false));\n"+
				"}"
		);
	}
	@Test
	public void test14() {
		// Assignment statements of the following type:
		// x:int = 3
		// (x:int, y:int) = (1,2)
		// (array(i), x:int) = function(y), where "function" returns a tuple
		typeCheck(
				 "g(x:int):(int,int) = {"
				+"    return (x,x);"
				+"}"

				+"f(a:int[]) = {"
				+"    i:int = 3;"
				+"    a(i) = 2;"
				+"    (j:int, k:int) = (4, 5);"
				+"    (a(i-1), k) = g(j);"
				+"}"
			);
	}
	

	
	@Test
	public void test16() {
		// Assignment statements of type: 
		// arrayOfArrays(i) = array
		// array = "Error!"
		typeCheck(
				 "arrayOfArrays(x:int[][], i:int):int[] = {"
				+"    x(i) = arrayOfArrays(x, i);"
				+"    v:int[] = x(i);"
				+"    rez:int[] = v;"
				+"    fallback:int[] = \"Error!\";"
				+"    if(v!=x(i))"
				+"        return fallback;"
				+"    else"
				+"        return rez;"
				+"}"
				);
		}
	
	@Test
	public void test17() {
		// Assignment statements of type:
		// (x:int, _) = (2, ((3)));
		// (array(i)) = 1;
		typeCheck(
				 "test_underscores(x:int):int = {"
				+"    v:int[];"
				+"    v(1) = x;"	
				+"    (y:int, _) = (x, ((x+1)));"
				+"    v(2) = y;"
				+"    (v(3)) = y+3;"
				+"}"
				);
	}

	@Test
	public void test18() {
		// should be able to compare two arrays 
		typeCheck(
			"test() = {\n"
			+ "x:int[];\n"
			+ "y:int[];\n"
			+ "z:bool=(x==y);\n"
			+ "}"
		);
	}
	
	@Test
	// unary minus with array access
	public void test19() {
		typeCheck(
				"test() : int = {\n" +
					"(x:int, y:int[], z:(bool,bool)) =" +
					" ('a', \"AAAA\", ((true),(((false)))));\n" +
					"return -y 2;" +
				"}"
		);
	}
	
	@Test
	public void test20() {
		// Mutual recursion
		typeCheck(
				 "abs(x:int):int = {"
				+"    if(x<0)"
				+"        return (-x);"
				+"    else"
				+"        return x;"
				+"}"
				+"isEven(number:int):bool = {"
				+"    if(number==0)"
				+"        return true;"
				+"    else"
				+"        return isOdd(abs(number)-1);"
				+"}"
				+"isOdd(number:int):bool = {"
				+"    if(number==0)"
				+"        return false;"
				+"    else"
				+"        return isEven(abs(number)-1);" 
				+"}"
				+"testingMyStuff():bool = {"
				+"    x:bool=true;"
				+"    return !(x!=false);" 
				+"}"
				);
	}
	@Test
	public void test21() {
		// interface usage
		typeCheck(
				"use io;\n" +
				 "foo(a:int[]) = {"
				+"	  print(\"foobar\");"
				+"	  print(a);"
				+"}"
				);
	}
	@Test
	public void test22() {
		// interface usage
		typeCheck(
				"use io;\n" +
				"foo() = {" +
					"while(!eof()) {\n" +
					"	line:int[] = readln();\n" +
					"	print(line);\n" +
					"	a:int = getchar();\n" +
					"}\n" +
				"}"
		);
	}
	@Test
	public void test23() {
		// ok to define same name in different blocks
		typeCheck(
				"foo(b:bool):int = {" +
					"if(b) {" +
					"	a:int = 5;\n" +
					"	return a;\n" +
					"} else {\n" +
					"	a:int = 3;\n" +
					"	return a;\n" +
					"}\n" +
				"}"
				);
	}
	@Test
	public void test24() {
		// Break inside nested if
		typeCheck(
				"foo(b:bool):int = {" +
					"if (b) {" +
						"while(true) {\n" +
						"	if(b) {" +
						"		a:int = 5;\n" +
						"		return a;\n" +
						"	} else {\n" +
						"		a:int = 3;\n" +
						"		if (false) break;\n" +
						"	}\n" +
						"}\n" +
					"}\n" +
				"}"
				);
	}

	@Test
	public void test25() {
		// complicated expression in function param
		typeCheck(
				"foo(b:bool):int = {\n" +
				"	l:bool[] = (true,false,true);\n" +
				"	return foo(bar 4 | bar 5 | l 2);" +
				"}\n" +
				"bar(c:int):bool = {" +
				"	return true;\n" +
				"}"
				);
	}	
	
	@Test
	public void test28() {
		// nasty piece of code which tries to make use of 
		// all expression typing rules
		typeCheck(
				"f1(a:int, b:bool, (c:bool, d:int)) : (int, bool, (bool, int)) = {" +
				"  return (d, c, (b, a));" +
				"}" +
				"" +
				"f2() : bool = {" +
				"  (i:int, b:bool, ia:int[], ib:bool[], " +
				"         t:(int, bool, (bool, int)));" +
				"  return (true & false " +
				"   | 3 == 4 & ia == \"abc\" " +
				"   & -i+i-i*i/i%i!=i" +
				"   | !b | i>ia 1 | ia i < i & ia i <= ia (i/2) " +
				"   & (ia (ia i) >= ia (i) " +
				"      | ib == (true, true)) " +
				"   | (ib (ia i), ib 2) == ib" +
				"   | !(b==b) " +
				"   | (1,b,(ib i, i)) == t" +
				"   | t != f1 (1,ib (length t)," +
				"        (ia i != ia (ia i), length ia))" +
				");" +
				"}"
		);
	}


	@Test
	public void test29() {
		// Explicitly calling a function as a procedure
		typeCheck(
				"a():int[] = {" +
				"    return \"test\";" +
				"}" +
				"b() = {" +
				"    _ = a();" +
				"}"
		);
	}

	@Test
	// Testing two-pass type checking 
	public void test30() {
		typeCheck(
				"euclid(x:int, y:int):int = {" +
				"    if(x<y)" +
				"        return euclid(x,y-x);" +
				"    else if(x>y)" +
				"        return euclid(x-y,y);" +
				"    else return x;" +
				"}"
		);
	}
	
	@Test
	// Testing use statements (Iota9 code with header files)
	public void test31() {
		typeCheck(
				 "use io;"
				+"gcd(a:int, b:int):int = {" 
				+"    if(a<b)" 
				+"        return gcd(a,b-a);" 
				+"    else if(a>b)" 
				+"        return gcd(a-b,b);" 
				+"    else " 
				+"        return a;" 
				+"}"				
				+"callFunctionsInHeader() = {"
				+"    x:int = gcd(2,3);"
				+"    y:int = 1;"
				+"    haveGCD: bool;"
				+"    if(x==y)"
				+"        haveGCD = true;"
				+"    else "
				+"        haveGCD = false;"
				+"    stringHaveGCD:int[];"
				+"    if(haveGCD)"
				+"        stringHaveGCD = \"true\";"
				+"    else"
				+"        stringHaveGCD = \"false\";"
				+"    print(stringHaveGCD);"
				+"}"
				);
	}

	@Test
	public void test32() {
		// Self recursion
		typeCheck(
				"factorial(n:int):int = {" +
				"  if (n == 0)" +
				"    return 1;" +
				"  else" +
				"    return n*factorial(n-1);" +
				"}"
		);
	}

	@Test
	public void test33() {
		// Function returning an array
		typeCheck(
				"a(i:int):int[] = {" +
				"  x1:int[] = \"test1\";" +
				"  x2:int[] = \"test2\";" +
				"  x:int[][] = (x1, x2);" +
				"}" + 
				"b():int = {" +
				"	return a 1 2;" +
				"}"
		);
	}
	@Test
	public void test34() {
		//  Assign primitive type to array via tuple subtype
		typeCheck(
			"test() = {\n"
			+ "  a:int[] = (1,2,3);"
			+ "  a = 3;\n"
			+ "}"
		);
	}
	@Test
	public void test35() {
		//  Array decl with length
		typeCheck(
			"test() = {\n"
			+ "  a:int[3] = (1,2,3);"
			+ "  a = 3;\n"
			+ "}"
		);
	}

	@Test
	public void test36() {
		typeCheck(
			"test() = {\n"
			+ "  a:(int,bool)[] = ((1,true),(2,false));"
			+ "  a 1 = (4,true);\n"
			+ "}"
		);
	}

	@Test
	public void test37() {
		typeCheck(
			  "a(i:int, _, j:int) : int = {"
			+ "   return i+j;"
			+ "}"
			+ "test() = {\n"
			+ "  _ = a(1, 2, 3);"
			+ "}"
		);
	}

	@Test
	public void test38() {
		// Spec update: can have global variables with an appropriate initialization
		typeCheck(
			  "i:int = 1;" +
			  "j:int = 'c';" +
			  "k:bool = false;"
		);
	}
	
	
	@Test
	public void test39() {
		// Spec update: can have global array declarations with simple length expressions 
		typeCheck(
			  "i:int = 1;" +
			  "j:int[15][12];" +
			  "k:int[15][];" +
			  "l:int[15];" +
			  "m:int[i];" +
			  "n:int[i][i];"
		);
	}
	
	@Test
	public void test40() {
		// The following is legal:
		typeCheck(
			"i, j: int = 0;"
		);
	}

	@Test
	public void test41() {
		/* We support neither covariance nor contravariance for array 
		   types, but invariant types are OK (see also test33 in TypeCheckerErrorTest) */
		typeCheck("class Point {};" +
		"class ColoredPoint extends Point { }"+
		"a() = {"+
		"    array1: ColoredPoint[5];"+
		"    array2 : ColoredPoint[];"+
		"    array2 = array1;"+
		"	}");
	}
}

	
